unit pr_Common;

interface

{$I PR.INC}

uses
  SysUtils, Windows, Classes, Graphics, WinSpool, DB,
  Dialogs, typinfo, Math, Forms, IniFiles, Messages, Controls,
  stdctrls, CommDlg, ComCtrls, ActnList, DsgnIntf,

  pr_Progress, pr_MultiLang, pr_Dataset, pr_Utils;

{$R prStrings.res}
{$R prDialogs.res}
{$R preport.res}

const
  AngleSize= 5;

  _sPrIniFileName     = 'pr.ini';
  sDefaultObjResName  = 'TprDefaultObjResName';

  MAX_PAGEHEADERSFOOTERS = 3;
  MAX_PAGEBANDSORDER = 8;
  MAX_PAGEBANDSORDERVERT = 8;

type

TprGenCell = class;
TprGenVector = class;
TprGenGrid = class;
TprBand = class;
TprBands = class;
TprObj = class;
TprValue = class;
TprObjClass = class of TprObj;
TprObjRecClass = class of TprObjRec;
TprGroup = class;
TprGroups = class;
TprBandClass = class of TprBand;
TprCustomEndPage = class;
TprCustomPage = class;
TprCustomReport = class;
TprDesigner = class;
TprDesignerClass = class of TprDesigner;
TprPreview = class;
TprPreviewClass = class of TprPreview;
TprCustomReportClass = class of TprCustomReport;

TprIntegerArray = array [0..1] of integer;
PprIntegerArray = ^TprIntegerArray;

TprHAlign = (prhLeft,prhCenter,prhRight);
TprVAlign = (prvTop,prvCenter,prvBottom);
TprCommonAlign = (praToMin,praCenter,praToMax);

TprColDirectionType = (prcdTopBottomLeftRight,prcdLeftRightTopBottom);

/////////////////////////////
//
//  ,   Parser
//
/////////////////////////////
TprVarValueType = (prvvtBoolean,prvvtInteger,prvvtDateTime,prvvtString,prvvtDouble,prvvtObject,prvvtNull,prvvtNotCalced);
TprVarValue = record
  vType   : TprVarValueType;

  vString : string;
  case TprVarValueType of
    prvvtBoolean  : (vBoolean  : boolean);
    prvvtInteger  : (vInteger  : integer);
    prvvtDateTime : (vDateTime : TDateTime);
    prvvtDouble   : (vDouble   : double);
    prvvtObject   : (vObject   : TObject);
end;
PprVarValue  = ^TprVarValue;
TprVarsArray = array of TprVarValue;

///////////////////////
//
// TprScrollBox
//
///////////////////////
TOnScroll = procedure (Sender : TObject; Msg : TWMScroll) of object;
TprScrollBox = class(TScrollBox)
private
  FOnVScroll : TOnScroll;
  FOnHScroll : TOnScroll;

  procedure mWM_VSCROLL(var Msg : TWMScroll); message WM_VSCROLL;
  procedure mWM_HSCROLL(var Msg : TWMScroll); message WM_HSCROLL;
public
  property OnVScroll : TOnScroll read FOnVScroll write FOnVScroll;
  property OnHScroll : TOnScroll read FOnHScroll write FOnHScroll;
end;

/////////////////////////////////////
//
//  :
//   -  
//   -    
//
/////////////////////////////////////
TprForm = class(TForm)
protected
  procedure prRestoreProperties(Ini : TIniFile; sn : string); virtual;
  procedure prSaveProperties(Ini : TIniFile; sn : string); virtual;
public
  procedure AfterConstruction; override;
  procedure BeforeDestruction; override;
end;

TprDesigner = class(TprForm)
private
protected
  procedure Loaded; override;
public
  Report : TprCustomReport;

  procedure CreateWnd; override;

  procedure UpdateCurrentPage; virtual; abstract;

  procedure ConvertToDesignerCoords(const rSource : TRect; var rDest : TRect); virtual;
  function  ConvertXToDesignerCoords(X : integer) : integer; virtual;
  function  ConvertYToDesignerCoords(Y : integer) : integer; virtual;
  procedure ConvertFromDesignerCoords(const rSource : TRect; var rDest : TRect); virtual;
  function  ConvertXFromDesignerCoords(X : integer) : integer; virtual;
  function  ConvertYFromDesignerCoords(Y : integer) : integer; virtual;

  procedure   BeforeDestruction; override;
  constructor CreateDesigner(AOwner : TComponent; _Report : TprCustomReport); virtual;
end;

TprPreview  = class(TprForm)
protected
  procedure Loaded; override;
public
  Report : TprCustomReport;

  procedure DoOnCustomAction;
  
  procedure CreateWnd; override;

  procedure   BeforeDestruction; override;
  constructor CreatePreview(AOwner : TComponent; _Report : TprCustomReport); virtual;
end;

TprResizeType    = (ppLeftTop,ppTop,ppRightTop,ppRight,ppRightBottom,ppBottom,ppLeftBottom,ppLeft);
TprResizeTypeSet = set of TprResizeType;
TprLinkType      = (ltLeft,ltTop,ltRight,ltBottom);
TprLinkTypeSet   = set of TprLinkType;

//////////////////////
//
// rDesignInfo
//
//////////////////////
rDesignInfo = record
  dPageRect         : TRect; //     ,
                             //     
                             //   ,    ,
                             //      :
                             //  x - 0 .. *
                             //  y - 0 .. *

  AllowResizeTypes  : TprResizeTypeSet; //    

  AllowDrag         : boolean;          //  

  AllowLinkTypes    : TprLinkTypeSet;   //       

  Parent            : TObject;          // ,   
end;
pDesignInfo = ^rDesignInfo;

/////////////////////
//
// TprObjRecVersion
//
/////////////////////
TprObjRecVersion = class(TCollectionItem)
private
  FFormula         : string;
  FVisible         : boolean;
  FCompiledFormula : string;

  FMayBeUse        : boolean;
  FSecondPassCalcCurVersionNeeded : boolean;
public
  procedure Assign(Source : TPersistent); override;
  constructor Create(Collection : TCollection); override; 
published
  property Formula : string read FFormula write FFormula;
  property Visible : boolean read FVisible write FVisible;
end;

/////////////////////
//
// TprObjRecVersions
//
/////////////////////
TprObjRecVersions = class(TCollection)
private
  function GetItm(i : integer) : TprObjRecVersion;
public
  property Items[i : integer] : TprObjRecVersion read GetItm; default;
end;

/////////////////////
//
// TprObjRec
//
/////////////////////
TprObjRec = class(TPersistent)
private
  FVersions   : TprObjRecVersions; //    ,   
  FDefVersion : integer;           //   

  FSecondPassCalcCurVersionNeeded : boolean;

  function GetLeft : integer;
  function GetTop : integer;
  function GetRight : integer;
  function GetBottom : integer;
  procedure SetLeft(Value : integer);
  procedure SetTop(Value : integer);
  procedure SetRight(Value : integer);
  procedure SetBottom(Value : integer);
protected
  function CreateVersions : TprObjRecVersions; virtual; abstract;
public
  Page      : TprCustomEndPage;
  Obj       : TprObj;              //   ,    
  X,Y,DX,DY : integer;             //   FirstPass

  pRect      : TRect;              //  
  CurVersion : integer;            // ,    

  procedure FirstPassCalcCurVersion; virtual;

  procedure SecondPass; virtual;

  procedure PlaceOnEndPage(Device : TObject; r : TRect); virtual; abstract;

  procedure Assign(Source : TPersistent); override;
  function  CreateCopy : TprObjRec; virtual; abstract;

  procedure Save(Stream : TStream); virtual;
  procedure Load(Stream : TStream); virtual;

  constructor Create(_Page : TprCustomEndPage; _Obj : TprObj); virtual;
  destructor Destroy; override;
published
  property DefVersion : integer read FDefVersion write FDefVersion;
  property Versions   : TprObjRecVersions read FVersions write FVersions;
  property Left       : integer read GetLeft write SetLeft;
  property Top        : integer read GetTop write SetTop;
  property Right      : integer read GetRight write SetRight;
  property Bottom     : integer read GetBottom write SetBottom;
end;

/////////////////////
//
// TprObjs
//
/////////////////////
TprObjs = class(TList)
private
  function GetItm(index : integer) : TprObj;
public
  property Items[index : integer] : TprObj read GetItm; default;
end;

/////////////////////
//
// TprObj
//
/////////////////////
TprVLinkMode    = (prlmMaxBottom,prlmMinBottom);
TprHLinkMode    = (prlmMaxRight,prlmMinRight);
TprVResizeMode  = (prrmMaxBottom, prrmMinBottom);
TprHResizeMode  = (prrmMaxRight, prrmMinRight);
TprLinkMode     = (prlLeft,prlTop,prlWidth,prlHeight);

TprObj = class(TComponent)
private
  FBand              : TprBand;

  FirstPassProcessed : boolean;

  FVisible           : boolean;

  FTopMode           : TprVLinkMode;
  FLeftMode          : TprHLinkMode;
  FWidthMode         : TprHResizeMode;
  FHeightMode        : TprVResizeMode;

  FTopObjsNames      : TStrings; //     (ReadLinks)    UpdateReadedLinks
  FLeftObjsNames     : TStrings;
  FWidthObjsNames    : TStrings;
  FHeightObjsNames   : TStrings;

  procedure SetBand(Value : TprBand);
  function  GetReport : TprCustomReport;
  function  IsLinkedTo(LinkMode : TprLinkMode; Obj : TprObj) : boolean;
protected
  FdRec : TprObjRec;

  procedure InitdRec; virtual; abstract;

  procedure ReadLeft(Reader : TReader);
  procedure WriteLeft(Writer : TWriter);
  procedure ReadTop(Reader : TReader);
  procedure WriteTop(Writer : TWriter);
  procedure ReadWidth(Reader : TReader);
  procedure WriteWidth(Writer : TWriter);
  procedure ReadHeight(Reader : TReader);
  procedure WriteHeight(Writer : TWriter);

  procedure DefineProperties(Filer : TFiler); override;

  procedure UpdatePDIdPageRect; virtual;

  procedure SetParentComponent(Value : TComponent); override;
  function  GetChildOwner : TComponent; override;

  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  aRec       : TprObjRec;     // ,     ,   FistPass
  PDI        : pDesignInfo;   //  

  TopObjs    : TprObjs;
  LeftObjs   : TprObjs;
  WidthObjs  : TprObjs;
  HeightObjs : TprObjs;

  property Band   : TprBand read FBand write SetBand;
  property Report : TprCustomReport read GetReport;
  //  ,  
  function  GetDesc : string; virtual;
  //     ,    r
  procedure DrawDesign(DC : HDC; ExData : pointer); virtual;
  //    
  procedure UpdatePDI(ExData : pointer); virtual;
  //     
  procedure FirstPass; virtual;

  //        Resize,
  //    Drag   ,
  //    ,     ,
  //    
  procedure DesignerResize(oTop,oLeft,oBottom,oRight : integer; ExData : pointer); virtual;
  procedure DesignerDrag(dx,dy : integer; ExData : pointer); virtual;
  procedure DesignerLink(Linked : pDesignInfo; LinkMode : TprLinkType; var LinkAccepted : boolean; ExData : pointer); virtual;
  procedure DesignerDelete; virtual;

  function  AllowInplaceEdit : boolean; virtual;
  procedure InplaceEdit(_Parent : TWinControl; var InplaceEditor : TWinControl); virtual;
  procedure SaveInplaceEdit(InplaceEditor : TWinControl); virtual;

  procedure AfterReportLoaded;

  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;

  function  HasParent : boolean; override;
  function  GetParentComponent : TComponent; override;
published
  property dRec       : TprObjRec read FdRec write FdRec;

  property TopMode    : TprVLinkMode read FTopMode write FTopMode;
  property LeftMode   : TprHLinkMode read FLeftMode write FLeftMode;
  property WidthMode  : TprHResizeMode read FWidthMode write FWidthMode;
  property HeightMode : TprVResizeMode read FHeightMode write FHeightMode;

  property Visible    : boolean read FVisible write FVisible;
end;





TprBandType = (bthTitle,
               bthSummary,
               bthPageHeader,
               bthPageFooter,
               bthDetail,
               bthDetailHeader,
               bthDetailFooter,
               bthGroupHeader,
               bthGroupFooter,
               btvTitle,
               btvSummary,
               btvPageHeader,
               btvPageFooter,
               btvDetail,
               btvDetailHeader,
               btvDetailFooter,
               btvGroupHeader,
               btvGroupFooter);
TprBandTypeArray = array of TprBandType;
TOnGetRegions = procedure (L,SelList : TList) of object;
TGenerateCellCallbackProc = procedure(Sender : TprBand; Cell : TprGenCell; Objects : TprObjs) of object;
////////////////////////
//
// TprBand
//
////////////////////////
TprBandResizeMode = (prbrmNone,prbrmMaxObj,prbrmMaxResizeObj,prbrmMinResizeObj);
TprBandLinkType = (prltAfter,prltBefore);

TprBand = class(TComponent)
private
  FPage            : TprCustomPage;

  FResizeObjsNames : TStrings;

  FResizeMode      : TprBandResizeMode;
  FVisible         : boolean;

  procedure SetPage(Value : TprCustomPage);
protected
  procedure GroupRemoved(Group : TprGroup); virtual;
  procedure GroupAdded(Group : TprGroup); virtual;

  procedure GetChildren(Proc : TGetChildProc; Root : TComponent); override;
  procedure SetParentComponent(Value : TComponent); override;
  procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  function  GetChildOwner : TComponent; override;

  procedure ReadResizeObjs(Reader : TReader);
  procedure WriteResizeObjs(Writer : TWriter);
  procedure DefineProperties(Filer : TFiler); override;
public
  FCalced      : boolean;

  BandType     : TprBandType;        //  Band

  ResizeObjs   : TprObjs;

  dPageRect    : TRect;              //  ,   
                                     //    , 
                                     //  ,     

  PDI          : pDesignInfo;        //  

  Objects      : TprObjs;

  //  ,   Band
  property  Page   : TprCustomPage read FPage write SetPage;

  // ,   Band,   Page
  function  Report : TprCustomReport;

  function  GetFullBandName : string;

  //   
  //     Cells  Page
  function  GenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc; fSimpleCreateCell : boolean) : TprGenCell; virtual; abstract;

  function  GetStartNewPage : boolean; virtual;
  function  GetLinkToBand : TprBand; virtual;
  function  GetLinkType : TprBandLinkType; virtual; 

  procedure InitDataSet; virtual;

  //  
  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); virtual; abstract;
  procedure UpdatePDI(ExData : pointer); virtual;
  procedure OnInsertIntoPage(p : TprCustomPage); virtual;

  procedure DesignerDelete; virtual;
  procedure DesignerResize(oTop,oLeft,oBottom,oRight : integer); virtual;
  procedure DesignerLink(Linked : pDesignInfo; LinkMode : TprLinkType; var LinkAccepted : boolean); virtual;

  procedure DrawDesign(DC : HDC; ExData : pointer); virtual; abstract;
  function  GetDrawDesignCaption : string; virtual;

  procedure AfterReportLoaded; virtual;

  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;

  function  HasParent : boolean; override;
  function  GetParentComponent : TComponent; override;
published
  property ResizeMode : TprBandResizeMode read FResizeMode write FResizeMode;
  property Visible    : boolean read FVisible write FVisible;
end;

////////////////////////
//
// TprCustomHBand
//
////////////////////////
TprCustomHBand = class(TprBand)
private
protected
  FHeight           : integer;
  FUseVerticalBands : boolean;
public
  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); override;

  //  
  function  GetUseColumns   : boolean; virtual;
  function  GetColCount     : integer; virtual;
  function  GetColDirection : TprColDirectionType; virtual;
  function  GenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc; fSimpleCreateCell : boolean) : TprGenCell; override;
  procedure GenerateCellCallback(Sender : TprBand; Cell : TprGenCell; Objects : TprObjs);

  //    
  function  PrintOnStartPage : boolean; virtual;

  //  
  procedure DesignerResize(oTop,oLeft,oBottom,oRight : integer); override;
  procedure DesignerLink(Linked : pDesignInfo; LinkMode : TprLinkType; var LinkAccepted : boolean); override;

  constructor Create(AOwner : TComponent); override;
published
  property Height           : integer read FHeight write FHeight;
  property UseVerticalBands : boolean read FUseVerticalBands write FUseVerticalBands;
end;

////////////////////////
//
// TprCustomHTitleBand
//
////////////////////////
TprCustomHTitleBand = class(TprCustomHBand)
public
  function PrintOnStartPage : boolean; override;
end;

//////////////////////////////
//
// TprCustomHSummaryBand
//
//////////////////////////////
TprCustomHSummaryBand = class(TprCustomHBand)
private
  FPrintWithBand : TprCustomHBand;
public
  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  constructor Create(AOwner : TComponent); override;
published
  property PrintWithBand : TprCustomHBand read FPrintWithBand write FPrintWithBand; 
end;

/////////////////////////
//
// TprCustomHPageHeaderBand
//
/////////////////////////
TprCustomHPageHeaderBand = class(TprCustomHBand)
private
  FPrintOnFirstPage : boolean;
public
published
  property PrintOnFirstPage : boolean read FPrintOnFirstPage write FPrintOnFirstPage;
end;

/////////////////////////
//
// TprCustomHPageFooterBand
//
/////////////////////////
TprCustomHPageFooterBand = class(TprCustomHBand)
private
  FPrintOnFirstPage : boolean;
public
  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); override;
published
  property PrintOnFirstPage : boolean read FPrintOnFirstPage write FPrintOnFirstPage;
end;

///////////////////////////////////
//
// TprCustomHDetailBand
//
///////////////////////////////////
TprCustomHDetailBand = class(TprCustomHBand)
private
  FParentDetail : TprCustomHDetailBand;
  FDataSetName  : string;
  FColCount     : integer;
  FColDirection : TprColDirectionType;

  FGroupsNames  : TStrings; //  ,     DetailBand
  FBandsNames   : TStrings;

  procedure SetParentDetail(Value : TprCustomHDetailBand);
protected
  procedure GroupRemoved(Group : TprGroup); override;
  procedure GroupAdded(Group : TprGroup); override;

  procedure ReadGroups(Reader : TReader);
  procedure WriteGroups(Writer : TWriter);
  procedure ReadBands(Reader : TReader);
  procedure WriteBands(Writer : TWriter);

  procedure DefineProperties(Filer : TFiler); override;

  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  DataSet : TprDatasetLink; //    

  Groups  : TprGroups;
  Bands   : TprBands; //  Band        Band

  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); override;

  //   
  function  GenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc; fSimpleCreateCell : boolean) : TprGenCell; override;
  function  GetUseColumns : boolean; override;
  function  GetColCount : integer; override;
  function  GetColDirection : TprColDirectionType; override;
  procedure InitDataSet; override;

  //  
  procedure OnInsertIntoPage(p : TprCustomPage); override;


  procedure AfterReportLoaded; override;


  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;
published
  property DataSetName  : string read FDataSetName write FDataSetName;
  property ColCount     : integer read FColCount write FColCount;
  property ColDirection : TprColDirectionType read FColDirection write FColDirection;
  property ParentDetail : TprCustomHDetailBand read FParentDetail write SetParentDetail;
end;

///////////////////////////
//
// TprCustomHDetailHeaderBand
//
///////////////////////////
TprCustomHDetailHeaderBand = class(TprCustomHBand)
private
  FDetailBand        : TprCustomHDetailBand;
  FColCount          : integer;
  FColDirection      : TprColDirectionType;
  FReprintOnEachPage : boolean;

  FLinkToDetail      : boolean;

  procedure SetDetailBand(Value : TprCustomHDetailBand);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  function  GetUseColumns : boolean; override;
  function  GetColCount : integer; override;
  function  GetColDirection : TprColDirectionType; override;

  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  procedure OnInsertIntoPage(p : TprCustomPage); override;
published
  property DetailBand        : TprCustomHDetailBand read FDetailBand write SetDetailBand;
  property ColCount          : integer read FColCount write FColCount;
  property ColDirection      : TprColDirectionType read FColDirection write FColDirection;
  property ReprintOnEachPage : boolean read FReprintOnEachPage write FReprintOnEachPage;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
end;

////////////////////////////
//
// TprCustomHDetailFooterBand
//
////////////////////////////
TprCustomHDetailFooterBand = class(TprCustomHBand)
private
  FDetailBand   : TprCustomHDetailBand;
  FColCount     : integer;
  FColDirection : TprColDirectionType;

  FLinkToDetail : boolean;
  
  procedure SetDetailBand(Value : TprCustomHDetailBand);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  function  GetUseColumns : boolean; override;
  function  GetColCount : integer; override;
  function  GetColDirection : TprColDirectionType; override;

  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  procedure OnInsertIntoPage(p : TprCustomPage); override;
published
  property DetailBand   : TprCustomHDetailBand read FDetailBand write SetDetailBand;
  property ColCount     : integer read FColCount write FColCount;
  property ColDirection : TprColDirectionType read FColDirection write FColDirection;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
end;

///////////////////////////
//
// TprCustomHGroupHeaderBand
//
///////////////////////////
TprCustomHGroupHeaderBand = class(TprCustomHBand)
private
  FGroup        : TprGroup;
  FColCount     : integer;
  FColDirection : TprColDirectionType;

  FLinkToDetail : boolean;
  FStartNewPage : boolean;

  procedure SetGroup(Value : TprGroup);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  procedure OnInsertIntoPage(p : TprCustomPage); override;

  function  GetUseColumns : boolean; override;
  function  GetColCount : integer; override;
  function  GetColDirection : TprColDirectionType; override;

  function  GetStartNewPage : boolean; override;
  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  function GetDrawDesignCaption : string; override;
published
  property Group        : TprGroup read FGroup write SetGroup;
  property ColCount     : integer read FColCount write FColCount;
  property ColDirection : TprColDirectionType read FColDirection write FColDirection;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
  property StartNewPage : boolean read FStartNewPage write FStartNewPage;
end;

///////////////////////////
//
// TprCustomHGroupFooterBand
//
///////////////////////////
TprCustomHGroupFooterBand = class(TprCustomHBand)
private
  FGroup        : TprGroup;
  FColCount     : integer;
  FColDirection : TprColDirectionType;

  FLinkToDetail : boolean;

  procedure SetGroup(Value : TprGroup);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  procedure OnInsertIntoPage(p : TprCustomPage); override;

  function  GetUseColumns : boolean; override;
  function  GetColCount : integer; override;
  function  GetColDirection : TprColDirectionType; override;

  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  function GetDrawDesignCaption : string; override;
published
  property Group        : TprGroup read FGroup write SetGroup;
  property ColCount     : integer read FColCount write FColCount;
  property ColDirection : TprColDirectionType read FColDirection write FColDirection;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
end;






////////////////////////
//
// TprCustomVBand
//
////////////////////////
TprCustomVBand = class(TprBand)
private
  FUseHorizontalBands : boolean;
protected
  FWidth             : integer;
  FCallerHBand       : TprCustomHBand;
  FCrossTabGenerated : boolean; //      ,
                                //     
public
  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); override;

  //    
  function  PrintOnStartPage : boolean; virtual;

  //   
  function  GenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc; fSimpleCreateCell : boolean) : TprGenCell; override;

  //  
  procedure DesignerResize(oTop,oLeft,oBottom,oRight : integer); override;
  procedure DesignerLink(Linked : pDesignInfo; LinkMode : TprLinkType; var LinkAccepted : boolean); override;

  constructor Create(AOwner : TComponent); override;
published
  property Width : integer read FWidth write FWidth;
  property UseHorizontalBands : boolean read FUseHorizontalBands write FUseHorizontalBands;
end;

///////////////////////////////
//
// TprCustomVTitleBand
//
///////////////////////////////
TprCustomVTitleBand = class(TprCustomVBand)
public
  function PrintOnStartPage : boolean; override;
end;

//////////////////////////////
//
// TprCustomVSummaryBand
//
//////////////////////////////
TprCustomVSummaryBand = class(TprCustomVBand)
private
  FPrintWithBand : TprCustomHBand;
public
  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  constructor Create(AOwner : TComponent); override;
published
  property PrintWithBand : TprCustomHBand read FPrintWithBand write FPrintWithBand; 
end;

/////////////////////////
//
// TprVPageHeaderBand
//
/////////////////////////
TprCustomVPageHeaderBand = class(TprCustomVBand)
private
  FPrintOnFirstPage : boolean;
public
published
  property PrintOnFirstPage : boolean read FPrintOnFirstPage write FPrintOnFirstPage;
end;


/////////////////////////
//
// TprVPageFooterBand
//
/////////////////////////
TprCustomVPageFooterBand = class(TprCustomVBand)
private
  FPrintOnFirstPage : boolean;
public
  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); override;
published
  property PrintOnFirstPage : boolean read FPrintOnFirstPage write FPrintOnFirstPage;
end;

///////////////////////////////////
//
// TprCustomVDetailBand
//
///////////////////////////////////
TprCustomVDetailBand = class(TprCustomVBand)
private
  FParentDetail : TprCustomVDetailBand;
  FDataSetName  : string;

  FGroupsNames  : TStrings;  //  ,     DetailBand
  FBandsNames   : TStrings;

  procedure SetParentDetail(Value : TprCustomVDetailBand);
protected
  procedure GroupRemoved(Group : TprGroup); override;
  procedure GroupAdded(Group : TprGroup); override;

  procedure ReadGroups(Reader : TReader);
  procedure WriteGroups(Writer : TWriter);
  procedure ReadBands(Reader : TReader);
  procedure WriteBands(Writer : TWriter);

  procedure DefineProperties(Filer : TFiler); override;

  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  DataSet : TprDatasetLink; //    

  Groups  : TprGroups;
  Bands   : TprBands; //  Band        Band

  procedure CalcdPageRect(var CurPageRect : TRect; ExData : pointer); override;

  //   
  function  GenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc; fSimpleCreateCell : boolean) : TprGenCell; override;
  procedure InitDataSet; override;

  //  
  procedure OnInsertIntoPage(p : TprCustomPage); override;

  procedure AfterReportLoaded; override;

  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;
published
  property DataSetName  : string read FDataSetName write FDataSetName;
  property ParentDetail : TprCustomVDetailBand read FParentDetail write SetParentDetail;
end;

///////////////////////////
//
// TprCustomVDetailHeaderBand
//
///////////////////////////
TprCustomVDetailHeaderBand = class(TprCustomVBand)
private
  FDetailBand        : TprCustomVDetailBand;
  FReprintOnEachPage : boolean;

  FLinkToDetail      : boolean;

  procedure SetDetailBand(Value : TprCustomVDetailBand);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  procedure OnInsertIntoPage(p : TprCustomPage); override;
published
  property DetailBand        : TprCustomVDetailBand read FDetailBand write SetDetailBand;
  property ReprintOnEachPage : boolean read FReprintOnEachPage write FReprintOnEachPage;

  property LinkToDetail      : boolean read FLinkToDetail write FLinkToDetail;
end;

////////////////////////////
//
// TprCustomVDetailFooterBand
//
////////////////////////////
TprCustomVDetailFooterBand = class(TprCustomVBand)
private
  FDetailBand   : TprCustomVDetailBand;

  FLinkToDetail : boolean;

  procedure SetDetailBand(Value : TprCustomVDetailBand);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  procedure OnInsertIntoPage(p : TprCustomPage); override;
published
  property DetailBand   : TprCustomVDetailBand read FDetailBand write SetDetailBand;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
end;

///////////////////////////
//
// TprCustomVGroupHeaderBand
//
///////////////////////////
TprCustomVGroupHeaderBand = class(TprCustomVBand)
private
  FGroup        : TprGroup;

  FLinkToDetail : boolean;
  FStartNewPage : boolean;

  procedure SetGroup(Value : TprGroup);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  procedure OnInsertIntoPage(p : TprCustomPage); override;

  function  GetStartNewPage : boolean; override;
  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  function GetDrawDesignCaption : string; override;
published
  property Group : TprGroup read FGroup write SetGroup;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
  property StartNewPage : boolean read FStartNewPage write FStartNewPage;
end;

///////////////////////////
//
// TprCustomVGroupFooterBand
//
///////////////////////////
TprCustomVGroupFooterBand = class(TprCustomVBand)
private
  FGroup        : TprGroup;

  FLinkToDetail : boolean;

  procedure SetGroup(Value : TprGroup);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
public
  procedure OnInsertIntoPage(p : TprCustomPage); override;

  function  GetLinkToBand : TprBand; override;
  function  GetLinkType : TprBandLinkType; override;

  function GetDrawDesignCaption : string; override;
published
  property Group : TprGroup read FGroup write SetGroup;

  property LinkToDetail : boolean read FLinkToDetail write FLinkToDetail;
end;





///////////////////////////////
//
// TprBands
//
///////////////////////////////
TprBands = class(TList)
private
  function GetItm(index : integer) : TprBand;
  function GetByBandType(BandType : TprBandType) : TprBand;
  function GetByName(Name : string) : TprBand;
public
  property Items[index : integer] : TprBand read GetItm; default;
  property ByBandType[BandType : TprBandType] : TprBand read GetByBandType;
  property ByName[Name : string] : TprBand read GetByName;

  function IndexByBandType(BandType : TprBandType) : integer;
  function IndexByName(Name : string) : integer;
end;

//////////////////////////
//
// TprGroup
//
//////////////////////////
TprGroup = class(TComponent)
private
  FValid       : string;
  FDetailBand  : TprBand;

  FPrevValue   : Variant;
  FNeedHeaders : boolean;
  FNeedFooters : boolean;
  FLineNo      : integer;

  FReport      : TprCustomReport;

  procedure SetDetailBand(Value : TprBand);
  procedure SetReport(Value : TprCustomReport);
protected
  procedure SetParentComponent(Value : TComponent); override;
  function  GetChildOwner : TComponent; override;
public
  Headers : TprBands;
  Footers : TprBands;

  property  GroupEnded : boolean read FNeedFooters;
  property  LineNo     : integer read FLineNo write FLineNo;
  property  Report     : TprCustomReport read FReport write SetReport;

  function  IndexInReport : integer;

  procedure Reset;
  procedure CalcValue;

  procedure HeadersGenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc);
  procedure FootersGenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc);
  procedure FootersAlwaysGenerateCell(CallerBand : TprBand; CallbackProc : TGenerateCellCallbackProc);

  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;

  function  HasParent : boolean; override;
  function  GetParentComponent : TComponent; override;
published
  property Valid      : string read FValid write FValid;
  property DetailBand : TprBand read FDetailBand write SetDetailBand;
end;

////////////////////////////
//
// TprGroups
//
////////////////////////////
TprGroups = class(TList)
private
  function GetItm(index : integer) : TprGroup;
  function GetByName(Name : string) : TprGroup;
public
  property Items[index : integer] : TprGroup read GetItm; default;
  property ByName[Name : string] : TprGroup read GetByName;

  function IndexByName(Name : string) : integer;
end;

/////////////////////////
//
// TprValueVersion
//
/////////////////////////
TprValueVersion = class(TObject)
public
  V1    : Variant;
  V2    : integer;

  Value : TprValue;  //    
  ID    : integer;   // 
  V     : Variant;   // 

  constructor Create;
end;

/////////////////////////
//
// TprValue
//
/////////////////////////
TprResetValueType = (rvtReport,rvtGroup,rvtPage,rvtDataSetEof);
TprCalcValueType  = (cvtDataSetNext,cvtEventOnReset,cvtCrossTab);
// cvtEventOnReset -    OnReset,  OnCalc
// cvtDataSetNext  -      DataSet   Formula
TprOnValueCalc             = procedure (Value : TprValue) of object;
TprOnGetVersionByVersionID = procedure(ValueVersion : TprValueVersion) of object;

TprValue = class(TCollectionItem)
private
  FName                    : string;
  FResetOn                 : TprResetValueType;// 
  FCalcOn                  : TprCalcValueType; //  
  FDataSetName             : string;           //  DataSet,   
  FResetDataSetName        : string;           //  DataSet,    Reset
  FCrossTabHorzDataSetName : string;           //
  FGroup                   : TprGroup;


  FVersions                : TList;
  FCurrentValueExists      : boolean;

  function  GetVersion(index : integer) : TprValueVersion;
  procedure SetCurrentValue(Value : Variant);
  procedure SetName(Value : string);
public
  SavedIndex              : integer;        //   CrossTab
  OnCalc                  : TprOnValueCalc; // cvtEventOnReset
  OnGetVersionByVersionID : TprOnGetVersionByVersionID;

  DataSet                 : TprDatasetLink;
  ResetDataSet            : TprDatasetLink;
  CrossTabHorzDataSet     : TprDatasetLink;

  FFormula                : string;           //     

  function Report : TprCustomReport;

  property CurrentValue : Variant write SetCurrentValue;
  property Versions[index : integer] : TprValueVersion read GetVersion;

  function  VersionsCount : integer;
  function  GetCurrentVersion : TprValueVersion; overload;
  function  GetCurrentVersion(var fCreated : boolean) : TprValueVersion; overload;
  function  GetCurrentVersionID : string;
  procedure Init;
  procedure Reset;
  procedure Clear;

  constructor Create(Collection : TCollection); override;
  destructor Destroy; override;
published
  property Name                    : string read FName write SetName;
  property Group                   : TprGroup read FGroup write FGroup;
  property Formula                 : string read FFormula write FFormula;
  property ResetOn                 : TprResetValueType read FResetOn write FResetOn;
  property CalcOn                  : TprCalcValueType read FCalcOn write FCalcOn;
  property DataSetName             : string read FDataSetName write FDataSetName;
  property ResetDataSetName        : string read FResetDataSetName write FResetDataSetName;
  property CrossTabHorzDataSetName : string read FCrossTabHorzDataSetName write FCrossTabHorzDataSetName;
end;

//////////////////////////
//
// TprValues
//
//////////////////////////
TprValues = class(TCollection)
private
  function GetItm(index : integer) : TprValue;
  function GetByName(name : string) : TprValue;
public
  Report : TprCustomReport;

  property Items[index : integer] : TprValue read GetItm; default;
  property ByName[Name : string] : TprValue read GetByName;

  function IndexByName(name : string) : integer;
  function VersionByVersionID(id : string) : TprValueVersion;
end;

//////////////////////////
//
// TprCustomEndPage
//    ,   
//    
//
//////////////////////////
TprCustomEndPage = class(TObject)
private
  FoRecs : TList;

  function GetoRec(index : integer) : TprObjRec;
public
  Report        : TprCustomReport;
  PageX         : integer; //     -   X
  PageY         : integer; //                                   -   Y

  Width         : integer;
  Height        : integer;

  Rect          : TRect;   // ,     ,
                           //   Band-   , 
                           //  TprReportTitleBand   
  AvailableRect : TRect;   //  ,
                           //   :
                           // vTitle,hTitle,vPageHeader,hPageHeader,vPageFooter,hPageFooter

  function CurrentWidth : integer;  // Rect.Right-Rect.Left
  function CurrentHeight : integer; // Rect.Bottom-Rect.Top

  property oRec[index : integer] : TprObjRec read GetoRec;
  function oRecsCount : integer;

  procedure ThirdPass; virtual; abstract;
  function  GetPageRect : TRect; virtual; abstract;

  procedure FreeGeneratedObjects;

  constructor Create(_Page : TprCustomPage); virtual;
  constructor CreateEmpty(_Report : TprCustomReport); virtual;
  destructor Destroy; override;
end;

//////////////////////////
//
// TprGenCell
//
//////////////////////////
TprGenCell = class(TObject)
private
  FObjRecs : TList;

  function GetObjRec(i : integer) : TprObjRec;
public
  Page    : TprCustomEndPage; //  ,    
  XOffs   : integer;          //      
  YOffs   : integer;          //

  Width   : integer;
  Height  : integer;

  HVector : TprGenVector;
  VVector : TprGenVector;

  property ObjRecs[i : integer] : TprObjRec read GetObjRec;
  function ObjRecsCount : integer;

  procedure Add(ObjRec : TprObjRec);

  constructor Create;
  destructor Destroy; override;
end;

////////////////////////////
//
// TprGenVector
//
////////////////////////////
TprGenVector = class(TList)
private
  FLinkToVector : TprGenVector;

  function GetItm(i : integer) : TprGenCell;
public
  Page   : integer; //   - PageX  EndPage,   PageY  EndPage
  XOffs  : integer;
  YOffs  : integer;
  Height : integer;
  Width  : integer;
  Band   : TprBand;

  property Itms[i : integer] : TprGenCell read GetItm; default;

  procedure AddCell(Cell : TprGenCell);

  constructor Create(_Band : TprBand);
end;

///////////////////////////
//
// TprGenGrid
//
///////////////////////////
TprGenGrid = class(TObject)
private
  FCurHVector      : TprGenVector;
  FCurVVectorIndex : integer;
  FVVectors        : TList;
  FHVectors        : TList;
  FInsertedVectors : TList;
  FCells           : TList;

  function GetVVector(i : integer) : TprGenVector;
  function GetHVector(i : integer) : TprGenVector;
  function GetCell(i : integer) : TprGenCell;
public
  Page         : TprCustomPage;
  vTitleCell   : TprGenCell;
  hTitleCell   : TprGenCell;
  vSummaryCell : TprGenCell;
  hSummaryCell : TprGenCell;

  property VVectors[i : integer] : TprGenVector read GetVVector;
  property HVectors[i : integer] : TprGenVector read GetHVector;
  property Cells[i : integer] : TprGenCell read GetCell;

  function VVectorsCount : integer;
  function HVectorsCount : integer;
  function CellsCount : integer;

  function IndexOfVVector(Vector : TprGenVector) : integer;
  function FindPriorVVectorOnPage(FromIndex : integer; Band : TprBand; EndPagePageX : integer) : TprGenVector;
  function FindPriorVVector(FromIndex : integer; Band : TprBand; MinEndPageIndex : integer) : TprGenVector;
  function CopyVVector(SourceVector : TprGenVector; Index : integer) : TprGenVector;

  function IndexOfHVector(Vector : TprGenVector) : integer;
  function FindPriorHVectorOnPage(FromIndex : integer; Band : TprBand; EndPagePageY : integer) : TprGenVector;
  function FindPriorHVector(FromIndex : integer; Band : TprBand; MinEndPageIndex : integer) : TprGenVector;
  function CopyHVector(SourceVector : TprGenVector; Index : integer) : TprGenVector;

  function  AddCell(_HBand : TprCustomHBand; _VBand : TprCustomVBand) : TprGenCell;
  function  AddCellSimple : TprGenCell;
  procedure EndOfLine;

  procedure Clear;

  constructor Create(_Page : TprCustomPage);
  destructor Destroy; override;
end;

//////////////////////////
//
// TprCustomPage
//
//////////////////////////
TprCustomPage = class(TComponent)
private
  FWidth       : integer;
  FHeight      : integer;

  FReport      : TprCustomReport;

  FOldEndPagesRects : array of TRect;

  procedure SetReport(Value : TprCustomReport);
protected
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
  procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
  procedure SetParentComponent(Value : TComponent); override;
  function  GetChildOwner : TComponent; override;

  procedure ReportSetted; virtual;
public
  Grid       : TprGenGrid;
  Bands      : TprBands; //   Band
  dPageRect  : PRect;    //   ,    
                         //       
                         // 

  property Report : TprCustomReport read FReport write SetReport;

  function IndexInReport : integer;

  //  
  function  GetPageRect : TRect; virtual; abstract;
  function  HasVerticalBands : boolean;

  //
  procedure FirstPassGenerateGrid;
  procedure SecondPassPlaceGridOnEndPage;

  // 
  procedure UpdateBandsPageRect;
  procedure DrawDesign(Canvas : TCanvas; ExData : pointer); virtual;

  // 
  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;

  function  HasParent : boolean; override;
  function  GetParentComponent : TComponent; override;
published
  property Width  : integer read FWidth write FWidth;
  property Height : integer read FHeight write FHeight;
end;

///////////////////////////
//
// TprCustomReport
//
///////////////////////////
rDataSetRecNo = record
  DataSet : TprDatasetLink;
  RecNo   : integer;
end;
TDataSetRecNoArray = array of rDataSetRecNo;

TOnBandGenerateCell       = procedure (Sender : TObject; Band : TprBand) of object;
TOnPageStart              = procedure (Sender : TObject; EndPage : TprCustomEndPage) of object;
TOnPageEnd                = procedure (Sender : TObject; EndPage : TprCustomEndPage) of object;
TOnFirstPassObject        = procedure (Sender : TObject; Obj : TprObj; var ManuallyProcessed : boolean) of object;
TOnUnknownVariable        = procedure (Sender : TObject; const VarName : string;
                                                         var Value : TprVarValue;
                                                         var IsProcessed : boolean) of object;
TOnUnknownFunction        = procedure (Sender : TObject; const FuncName   : string;
                                                         const Parameters : TprVarsArray;
                                                         ParametersCount  : integer;
                                                         var Value        : TprVarValue;
                                                         var IsProcessed  : boolean) of object;
TOnUnknownObjFunction     = procedure (Sender : TObject; Component        : TComponent;
                                                         const FuncName   : string;
                                                         const Parameters : TprVarsArray;
                                                         ParametersCount  : integer;
                                                         var Value        : TprVarValue;
                                                         var IsProcessed  : boolean) of object;

TOnDesignerSave           = procedure (Sender : TObject; var FileName : string; SaveAs : Boolean) of object;
TOnGetAvailableComponents = procedure (Sender : TObject; L : TStrings) of object;
TOnGetAvailableDatasets   = procedure (Sender : TObject; L : TStrings) of object;
TOnCloseSetupDialogEvent  = procedure (Sender : TObject; fOK : boolean) of object; 

TprReportMode     = (rmDesign, rmPrinting, rmNone);
TprPrintPagesMode = (ppmAll,ppmPagesRange,ppmSelection,ppmPagesList);
TprFormMode       = (fmNormal,fmMDIChild);

TprCustomReport = class(TComponent)
private
  FTitle                : string;

  FCollate              : boolean;

  FShowProgress         : boolean;

  FDesignerFormMode     : TprFormMode;
  FPreviewFormMode      : TprFormMode;

  FValues               : TprValues;

  FCustomActionInPreviewCaption : string;

  FOnBandGenerateCell           : TOnBandGenerateCell;
  FOnPageStart                  : TOnPageStart;
  FOnPageEnd                    : TOnPageEnd;
  FOnFirstPassObject            : TOnFirstPassObject;
  FOnUnknownVariable            : TOnUnknownVariable;
  FOnUnknownFunction            : TOnUnknownFunction;
  FOnUnknownObjFunction         : TOnUnknownObjFunction;
  FOnGetAvailableComponents     : TOnGetAvailableComponents;
  FOnGetAvailableDatasets       : TOnGetAvailableDatasets;
  FOnCustomActionInPreview      : TNotifyEvent;

  FOnDesignerSave               : TOnDesignerSave;
  FOnCreateDesigner             : TNotifyEvent;
  FOnDestroyDesigner            : TNotifyEvent;
  FOnCreatePreview              : TNotifyEvent;
  FOnDestroyPreview             : TNotifyEvent;
  FOnCloseSetupDialog           : TOnCloseSetupDialogEvent;

  FEndPages             : TList;
  FPages                : TList;
  FParser               : TObject;

  FDesignerForm         : TprDesigner;
  FPreviewForm          : TprPreview;

  ADSRN                 : TDataSetRecNoArray;

  FReportPrepared       : boolean;
  FTemplateChanged      : boolean;
  FGACList              : TStringList;

  function  GetPage(index : integer) : TprCustomPage;
  function  GetEndPage(index : integer) : TprCustomEndPage;

  function  GetAllValuesCount : integer;
  function  GetAllValue(i : integer) : TprValue;

  procedure OnCalcPagesCount(Value : TprValue);
  procedure OnCalcPageNo(ValueVersion : TprValueVersion);
  procedure OnCalcPageNo2(Value : TprValue);

  function  GetprOwner : TComponent;
protected
  FCopies         : integer;
  FFromPage       : integer;
  FToPage         : integer;
  FPrintPages     : string;
  FPrintPagesMode : TprPrintPagesMode;

  FActionCanceled : boolean;
  FProgressForm   : TprProgressForm;

  function  GetPrinterName : string; virtual; abstract;
  procedure SetPrinterName(Value : string); virtual; abstract;

  procedure Loaded; override;
  procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
  procedure Notification(AComponent : TComponent; Operation : TOperation); override;
  function  GetChildOwner : TComponent; override;

  function GetDesignerFormClass : string; virtual; abstract;
  function GetPreviewFormClass : string; virtual; abstract;

  procedure SetTemplateChanged(Value : boolean); virtual;
public
  Designer               : IFormDesigner; //    Delphi IDE

  StartDateTime          : TDateTime;

  Groups                 : TprGroups;
  IndexCurEndPage        : integer; //    

  SystemValues           : TprValues;

  FPrintPagesList        : TList;
  ObjectCalcSizesDevice  : TObject;

  property DesignerForm    : TprDesigner read FDesignerForm write FDesignerForm;
  property PreviewForm     : TprPreview read FPreviewForm write FPreviewForm;
  property ProgressForm    : TprProgressForm read FProgressForm write FProgressForm;

  property ActionCanceled  : boolean read fActionCanceled;
  property TemplateChanged : boolean read FTemplateChanged write SetTemplateChanged;

  property Pages[index : integer]    : TprCustomPage read GetPage;
  property EndPages[index : integer] : TprCustomEndPage read GetEndPage;

  property AllValues[i : integer]    : TprValue read GetAllValue;
  property AllValuesCount            : integer read GetAllValuesCount;

  property prOwner                   : TComponent read GetprOwner;

  property ReportPrepared            : boolean read FReportPrepared;

  function PagesCount : integer;
  function EndPagesCount : integer;

  function  CreateEndPage(Page : TprCustomPage) : TprCustomEndPage; virtual; abstract;
  function  CreateEmptyEndPage : TprCustomEndPage; virtual; abstract;
  function  CurEndPage : TprCustomEndPage;
  procedure AddEndPage(Page : TprCustomPage);
  procedure AddEndPageToList(EndPage : TprCustomEndPage);
  procedure DeleteEndPage(index : integer);
  procedure ClearEndPages;

  procedure ConvertToDesignerCoords(const rSource : TRect; var rDest : TRect);

  function  GetBandClass(BandType : TprBandType) : TprBandClass; virtual;

  procedure ChangePrinter(OldIndex,NewIndex : integer); virtual; abstract;

  function  PrepareReport : boolean; virtual;
  procedure DesignReport(Modal : boolean);
  procedure PreviewPreparedReport(Modal : boolean);
  function  SetupPrintParams : boolean; virtual; abstract;
  function  PrintPreparedReport : boolean; virtual; abstract;

  function  Calc(var Expression : string; Value : TprValue) : Variant;
  function  FormatTemplate(Template : string; var Res : string) : boolean;
  function  FormatStrings(lSource,lDest : TStrings; DeleteEmptyLines,DeleteEmptyLinesAtEnd : boolean) : boolean;

  procedure DoOnBandGenerateCell(Band : TprBand);
  procedure DoOnDataSetNext(DataSet : TprDatasetLink; HorizontalBand : TprCustomHBand; VerticalBand : TprCustomVBand);
  procedure DoOnDataSetEof(DataSet : TprDatasetLink);
  procedure DoOnGroupEnd(Group : TprGroup);

  procedure ClearTemplate;
  procedure SaveTemplate(Stream : TStream; InBinaryFormat : boolean);
  procedure SaveTemplateToFile(FileName : string; InBinaryFormat : boolean);
  procedure SaveTemplateToStrings(Dest : TStrings);
  procedure LoadTemplate(Stream : TStream; InBinaryFormat : boolean);
  procedure LoadTemplateFromFile(FileName : string; InBinaryFormat : boolean);
  procedure LoadTemplateFromStrings(Source : TStrings);

  procedure LoadPreparedReport(Stream : TStream); virtual;
  procedure SavePreparedReport(Stream : TStream); virtual;
  procedure LoadPreparedReportFromFile(FileName : string); virtual;
  procedure SavePreparedReportToFile(FileName : string); virtual;

  procedure TranslateObjectName(Ident : string; var Component : TComponent; var LastName : string);

  procedure GetAvailableDataSets(L : TStrings);
  procedure GetAvailableComponents(L : TStrings);
  function  GetDataSetByName(DataSetName : string) : TObject;

  function  GetDataSetRecNo(Dataset : TprDatasetLink) : integer; overload;
  function  GetDataSetRecNo(Dataset : TObject) : integer; overload;

  procedure DoOnPageStart(EndPage : TprCustomEndPage);
  procedure DoOnPageEnd(EndPage : TprCustomEndPage);
  procedure DoOnFirstPassObject(Sender : TprObj; var ManuallyProcessed : boolean);

  function  ADSRN_GetIndex(DataSet : TprDatasetLink) : integer; overload;
  function  ADSRN_GetIndex(DataSet : TObject) : integer; overload;
  procedure ADSRN_First(DataSet : TprDatasetLink);
  procedure ADSRN_Next(DataSet : TprDatasetLink);

  procedure UpdateProgressForm(Text : string);

  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;
published
  property ShowProgress                 : boolean read FShowProgress write FShowProgress;
                                        
  property DesignerFormMode             : TprFormMode read FDesignerFormMode write FDesignerFormMode;
  property PreviewFormMode              : TprFormMode read FPreviewFormMode write FPreviewFormMode;
                                        
  property Collate                      : boolean read FCollate write FCollate;
  property Copies                       : integer read FCopies write FCopies;
  property FromPage                     : integer read FFromPage write FFromPage;
  property ToPage                       : integer read FToPage write FToPage;
  property PrintPages                   : string read FPrintPages write FPrintPages;
  property PrintPagesMode               : TprPrintPagesMode read FPrintPagesMode write FPrintPagesMode;
  property Title                        : string read FTitle write FTitle;
                                        
  property Values                       : TprValues read FValues write FValues;
                                        
  property PrinterName                  : string read GetPrinterName write SetPrinterName;
  property CustomActionInPreviewCaption : string read FCustomActionInPreviewCaption write FCustomActionInPreviewCaption;

  property OnPageStart              : TOnPageStart read FOnPageStart write FOnPageStart;
  property OnPageEnd                : TOnPageEnd read FOnPageEnd write FOnPageEnd;
  property OnFirstPassObject        : TOnFirstPassObject read FOnFirstPassObject write FOnFirstPassObject;
  property OnUnknownVariable        : TOnUnknownVariable read FOnUnknownVariable write FOnUnknownVariable;
  property OnUnknownFunction        : TOnUnknownFunction read FOnUnknownFunction write FOnUnknownFunction;
  property OnUnknownObjFunction     : TOnUnknownObjFunction read FOnUnknownObjFunction write FOnUnknownObjFunction;
  property OnGetAvailableComponents : TOnGetAvailableComponents read FOnGetAvailableComponents write FOnGetAvailableComponents;
  property OnGetAvailableDatasets   : TOnGetAvailableDatasets read FOnGetAvailableDatasets write FOnGetAvailableDatasets;
  property OnDesignerSave           : TOnDesignerSave read FOnDesignerSave write FOnDesignerSave;
  property OnCreateDesigner         : TNotifyEvent read FOnCreateDesigner write FOnCreateDesigner;
  property OnDestroyDesigner        : TNotifyEvent read FOnDestroyDesigner write FOnDestroyDesigner;
  property OnCreatePreview          : TNotifyEvent read FOnCreatePreview write FOnCreatePreview;
  property OnDestroyPreview         : TNotifyEvent read FOnDestroyPreview write FOnDestroyPreview;
  property OnCustomActionInPreview  : TNotifyEvent read FOnCustomActionInPreview write FOnCustomActionInPreview;
  property OnCloseSetupDialog       : TOnCloseSetupDialogEvent read FOnCloseSetupDialog write FOnCloseSetupDialog;
  property OnBandGenerateCell       : TOnBandGenerateCell read FOnBandGenerateCell write FOnBandGenerateCell;
end;

function  prGetDefaultPrinterName : string;

function  ReadString(Stream : TStream) : string;
procedure ReadRect(Stream : TStream; var r : TRect);
procedure WriteString(Stream : TStream; s : string);
procedure WriteRect(Stream : TStream; r : TRect);

procedure prLoadResImages(Form : TForm; IL : TImageList);
procedure LoadResImage(b : TBitmap; ResID : string);
function  GetValidComponentName(Component : TComponent) : string;
procedure prWriteCompListNames(Writer : TWriter; L : TList);
procedure prReadStringList(Reader : TReader; L : TStrings);
procedure GetEnumNamesToStrings(ti : PTypeInfo; Max : integer; L : TStrings);

procedure DrawRect(DC : HDC; const R : TRect);
function  Create90Font(Font : TFont) : HFont;
function  CreateAPIFont(Font : TFont) : HFont;
function  CreateDefFont(DC : HDC; Size : integer; Color : TColor) : HFont;
function  Create90DefFont(DC : HDC; Size : integer; Color : TColor) : HFont;

function  PointInRect(X,Y : integer; const R : TRect) : boolean;
function  RectInRect(const r1 : TRect; const r2 : TRect) : boolean;
function  RectOverRect(const r1 : TRect; const r2 : TRect) : boolean;
function  MulRect(const R : TRect; cx,cy : integer) : TRect;
function  DivRect(const R : TRect; cx,cy : integer) : TRect;

procedure DrawAngleRect(DC : HDC; r : TRect);

procedure prRegisterObj(_ClassRef           : TprObjClass;
                        _RecClassRef        : TprObjRecClass;
                        _ReportRef          : TprCustomReportClass;
                        _CaptionResID       : integer;
                        _HintResID          : integer;
                        _PropsFormClassName : string);
function  CreatePrObj(ClassRef  : TprObjClass;
                      Band      : TprBand) : TprObj;
function  CreatePrObjRec(ClassRef  : TprObjRecClass;
                         EndPage   : TprCustomEndPage) : TprObjRec;

type

TprPaperInfo = record
  Typ  : Integer;
  Name : String;
  X,Y  : Integer;
end;

rBandInfo = record
  Title    : string;
  ClassRef : TprBandClass;
end;

TprObjRegInfo = record
  ClassRef           : TprObjClass;
  RecClassRef        : TprObjRecClass;
  ReportRef          : TprCustomReportClass;
  CaptionResID       : integer;
  HintResID          : integer;
  PropsFormClassName : string;
end;

var
  prObjRegInfos             : array of TprObjRegInfo;
  prIniFileName             : string;
  prGetComponentsReportDesc : string;

const
  PageBandsOrder          : array [0..MAX_PAGEBANDSORDER] of TprBandType = (bthTitle,bthPageHeader,bthPageFooter,bthDetail,bthSummary,bthDetailHeader,bthDetailFooter,bthGroupHeader,bthGroupFooter);
  PageBandsOrderVert      : array [0..MAX_PAGEBANDSORDERVERT] of TprBandType = (btvTitle,btvPageHeader,btvPageFooter,btvDetail,btvSummary,btvDetailHeader,btvDetailFooter,btvGroupHeader,btvGroupFooter);
  PageHeadersFootersBands : array [0..MAX_PAGEHEADERSFOOTERS] of TprBandType = (btvPageHeader,(btvPageFooter),bthPageHeader,bthPageFooter);
  
  VerticalBands : set of TprBandType = [btvTitle,
                                        btvSummary,
                                        btvPageHeader,
                                        btvPageFooter,
                                        btvDetail,
                                        btvDetailHeader,
                                        btvDetailFooter,
                                        btvGroupHeader,
                                        btvGroupFooter];
  HorizontalBands : set of TprBandType = [bthTitle,
                                          bthSummary,
                                          bthPageHeader,
                                          bthPageFooter,
                                          bthDetail,
                                          bthDetailHeader,
                                          bthDetailFooter,
                                          bthGroupHeader,
                                          bthGroupFooter];

  BandTitles : array [TprBandType] of string = (
    (' '),
    ('  '),
    (' '),
    (' '),
    (''),
    (' '),
    (' '),
    (' '),
    (' '),
    ('  '),
    ('   '),
    ('  '),
    ('  '),
    (' '),
    ('  '),
    ('  '),
    ('  '),
    ('  '));

implementation

uses
  pr_Strings, pr_Parser;

///////////////////////////
//
// 
//
///////////////////////////
function prGetDefaultPrinterName;
var
  n : integer;
  Buffer : pointer;
  BytesNeeded,NumInfo : cardinal;
begin
Result     :='';
Buffer     :=nil;
BytesNeeded:=0;

case Win32Platform of
  VER_PLATFORM_WIN32_WINDOWS:
    begin
      EnumPrinters(PRINTER_ENUM_DEFAULT,nil,5,Buffer,0,BytesNeeded,NumInfo);
      if BytesNeeded<>0 then
        begin
          GetMem(Buffer,BytesNeeded);
          try
            if EnumPrinters(PRINTER_ENUM_DEFAULT,nil,5,Buffer,BytesNeeded,BytesNeeded,NumInfo) then
              if NumInfo>0 then
                Result:=StrPas(PPrinterInfo5(Buffer).pPrinterName);
          finally
            FreeMem(Buffer);
          end
        end;
    end;
  VER_PLATFORM_WIN32_NT:
    begin
    {
      EnumPrinters(PRINTER_ENUM_DEFAULT,nil,4,Buffer,0,BytesNeeded,NumInfo);
      if BytesNeeded<>0 then
        begin
          GetMem(Buffer,BytesNeeded);
          try
            if EnumPrinters(PRINTER_ENUM_DEFAULT,nil,4,Buffer,BytesNeeded,BytesNeeded,NumInfo) then
              if NumInfo>0 then
                Result:=StrPas(PPrinterInfo4(Buffer).pPrinterName);
          finally
            FreeMem(Buffer);
          end
        end
      else
        outputdebugstring(pchar(syserrormessage(getlasterror)));
    }
    end;
end;

if Result='' then
  begin
    SetLength(Result,80);
    n:=GetProfileString('windows',
                        'device',
                        '',
                        @(Result[1]),
                        79);
    SetLength(Result,n);
    n:=pos(',',Result);
    if n<>0 then
      Result:=Copy(Result,1,n-1);
  end;
end;

procedure prLoadResImages;
var
  i : integer;
  c : TComponent;
  l : TStringList;
  b : TBitmap;

  function LoadImage(ResName : string) : integer;
  begin
  ResName:=AnsiUpperCase(Copy(ResName,2,Length(ResName)));
  Result :=l.IndexOf(ResName);
  if Result=-1 then
    begin
      if FindResource(hInstance,PChar(ResName),RT_BITMAP)<>0 then
        try
          b.LoadFromResourceName(hInstance,ResName);
          l.AddObject(ResName,TObject(IL.AddMasked(b,b.TransparentColor)));
        except
          l.AddObject(ResName,TObject(-1));
        end
      else
        l.AddObject(ResName,TObject(-1));
      Result:=integer(l.Objects[l.Count-1])
    end
  else
    begin
      Result:=integer(l.Objects[Result]);
    end;
  end;

begin
l:=TStringList.Create;
b:=TBitmap.Create;
try
  for i:=0 to Form.ComponentCount-1 do
    begin
      c:=Form.Components[i];
      if (c is TToolButton) and (TToolButton(c).Action=nil) and (TToolButton(c).Style<>tbsSeparator) then
        begin
          TToolButton(c).ImageIndex:=LoadImage(TToolButton(c).Name);
        end;
  
      if Form.Components[i] is TAction then
        begin
          TAction(c).ImageIndex:=LoadImage(TAction(c).Name);
        end
    end;
finally
  l.Free;
  b.Free;
end;
end;

procedure GetEnumNamesToStrings;
var
  i : integer;
begin
for i:=0 to Max do
  L.Add(GetEnumName(ti,i));
end;

function MulRect;
begin
Result.Left  :=r.Left*cx;
Result.Right :=r.Right*cx;
Result.Top   :=r.Top*cy;
Result.Bottom:=r.Bottom*cy;
end;

function DivRect;
begin
Result.Left  :=r.Left div cx;
Result.Right :=r.Right div cx;
Result.Top   :=r.Top div cy;
Result.Bottom:=r.Bottom div cy;
end;

procedure prRegisterObj;
begin
SetLength(prObjRegInfos,Length(prObjRegInfos)+1);
with prObjRegInfos[High(prObjRegInfos)] do
  begin
    ClassRef           :=_ClassRef;
    RecClassRef        :=_RecClassRef;
    ReportRef          :=_ReportRef;
    CaptionResID       :=_CaptionResID;
    HintResID          :=_HintResID;
    PropsFormClassName :=_PropsFormClassName;
  end;
end;

procedure DrawAngleRect(DC : HDC; r : TRect);
var
  npn,opn : HPen;
begin
npn:=CreatePen(PS_SOLID,1,clBlack);
opn:=SelectObject(DC,npn);

MoveToEx(DC,r.Left+AngleSize,r.Top,nil);
LineTo(DC,r.Left,r.Top);
LineTo(DC,r.Left,r.Top+AngleSize);

MoveToEx(DC,r.Left+AngleSize,r.Bottom-1,nil);
LineTo(DC,r.Left,r.Bottom-1);
LineTo(DC,r.Left,r.Bottom-AngleSize-1);

MoveToEx(DC,r.Right-AngleSize-1,r.Top,nil);
LineTo(DC,r.Right-1,r.Top);
LineTo(DC,r.Right-1,r.Top+AngleSize);

MoveToEx(DC,r.Right-AngleSize-1,r.Bottom-1,nil);
LineTo(DC,r.Right-1,r.Bottom-1);
LineTo(DC,r.Right-1,r.Bottom-AngleSize-1);

SelectObject(DC,opn);
DeleteObject(npn);
end;

function PointInRect;
begin
Result:=((X>=R.Left) and (X<=R.Right)) and ((Y>=R.Top) and (Y<=R.Bottom))
end;

function RectInRect;
begin
Result:=PointInRect(r1.Left,r1.Top,r2) and
        PointInRect(r1.Right,r1.Top,r2) and
        PointInRect(r1.Right,r1.Bottom,r2) and
        PointInRect(r1.Left,r1.Bottom,r2);
end;

function RectOverRect;
begin
Result:=PointInRect(r1.Left,r1.Top,r2) or
        PointInRect(r1.Right,r1.Top,r2) or
        PointInRect(r1.Right,r1.Bottom,r2) or
        PointInRect(r1.Left,r1.Bottom,r2) or

        PointInRect(r2.Left,r2.Top,r1) or
        PointInRect(r2.Right,r2.Top,r1) or
        PointInRect(r2.Right,r2.Bottom,r1) or
        PointInRect(r2.Left,r2.Bottom,r1) or

        ((r1.Left>r2.Left) and (r1.Right<r2.Right) and (r1.Top<r2.Top) and (r1.Bottom>r2.Bottom)) or
        ((r1.Left<r2.Left) and (r1.Right>r2.Right) and (r1.Top>r2.Top) and (r1.Bottom<r2.Bottom)) or

        ((r2.Left>r1.Left) and (r2.Right<r1.Right) and (r2.Top<r1.Top) and (r2.Bottom>r1.Bottom)) or
        ((r2.Left<r1.Left) and (r2.Right>r1.Right) and (r2.Top>r1.Top) and (r2.Bottom<r1.Bottom));
end;

function CreatePrObj;
begin
Result     :=ClassRef.Create(Band.Report.prOwner);
Result.Band:=Band;
end;

function CreatePrObjRec;
begin
Result:=ClassRef.Create(EndPage,nil);
end;

procedure DrawRect;
begin
MoveToEx(DC,R.Left,R.Top,nil);
LineTo(DC,R.Right-1,R.Top);
MoveToEx(DC,R.Left,R.Bottom-1,nil);
LineTo(DC,R.Right-1,R.Bottom-1);
MoveToEx(DC,R.Left,R.Top+1,nil);
LineTo(DC,R.Left,R.Bottom-1);
MoveToEx(DC,R.Right-1,R.Top,nil);
LineTo(DC,R.Right-1,R.Bottom);
end;

procedure WriteString(Stream : TStream; s : string);
var
  b : integer;
begin
b:=Length(s);
Stream.Write(b,4);
Stream.Write(s[1],b);
end;

procedure WriteRect(Stream : TStream; r : TRect);
begin
Stream.Write(r.Left,4);
Stream.Write(r.Top,4);
Stream.Write(r.Right,4);
Stream.Write(r.Bottom,4);
end;

function ReadString(Stream : TStream) : string;
var
  b : integer;
begin
Stream.Read(b,4);
SetLength(Result,b);
Stream.Read(Result[1],b);
end;

procedure ReadRect(Stream : TStream; var r : TRect);
begin
Stream.Read(r.Left,4);
Stream.Read(r.Top,4);
Stream.Read(r.Right,4);
Stream.Read(r.Bottom,4);
end;

procedure LoadResImage;
begin
try
  if FindResource(hInstance,PChar(ResID),RT_BITMAP)<>0 then
    b.LoadFromResourceName(hInstance,ResID);
except
end;
end;

function GetValidComponentName;
var
  i : integer;
  s : string;
begin
s:=Component.ClassName;
i:=1;
while Component.Owner.FindComponent(Copy(s,2,Length(s))+IntToStr(i))<>nil do Inc(i);
Result:=Copy(s,2,Length(s))+IntToStr(i);
end;

procedure prWriteCompListNames(Writer : TWriter; L : TList);
var
  i : integer;
begin
Writer.WriteListBegin;
for i:=0 to L.Count-1 do
  Writer.WriteString(TComponent(L[i]).Name);
Writer.WriteListEnd;
end;

procedure prReadStringList(Reader : TReader; L : TStrings);
begin
Reader.ReadListBegin;
while not Reader.EndOfList do
  L.Add(Reader.ReadString);
Reader.ReadListEnd;
end;

function Create90Font;
var
  F : TLogFont;
begin
  GetObject(Font.Handle, SizeOf(TLogFont), @F);
  F.lfEscapement  := 900;
  F.lfOrientation := 900;
  Result          := CreateFontIndirect(F);
end;

function CreateAPIFont;
var
  F : TLogFont;
begin
  GetObject(Font.Handle, SizeOf(TLogFont), @F);
  Result:= CreateFontIndirect(F);
end;

function  CreateDefFont;
var
  F : TLogFont;
begin
  SetTextColor(DC,Color);
  ZeroMemory(@F,sizeof(F));
  F.lfHeight  :=-MulDiv(Size, GetDeviceCaps(DC, LOGPIXELSY), 72);
  F.lfFaceName:='Arial';
  F.lfCharSet :=DEFAULT_CHARSET;
  Result      :=CreateFontIndirect(F);
end;

function  Create90DefFont;
var
  F : TLogFont;
begin
  SetTextColor(DC,Color);
  ZeroMemory(@F,sizeof(F));
  F.lfHeight      :=-MulDiv(Size, GetDeviceCaps(DC, LOGPIXELSY), 72);
  F.lfFaceName    :='Arial';
  F.lfCharSet     :=DEFAULT_CHARSET;
  F.lfEscapement  := 900;
  F.lfOrientation := 900;
  Result          :=CreateFontIndirect(F);
end;

///////////////////////////
//
// TprScrollBox
//
///////////////////////////
procedure TprScrollBox.mWM_VSCROLL;
begin
inherited;
if Assigned(OnVScroll) then
  OnVScroll(Self,Msg);
end;

procedure TprScrollBox.mWM_HSCROLL;
begin
inherited;
if Assigned(OnHScroll) then
  OnHScroll(Self,Msg);
end;

/////////////////////////////
//
// TprForm
//
/////////////////////////////
procedure TprForm.AfterConstruction;
var
  Ini : TIniFile;
  buf : array [0..254] of char;
begin
inherited;

GetWindowsDirectory(buf,254);
Ini:=TIniFile.Create(prIniFileName);
try
  prRestoreProperties(Ini,ClassName);
finally
  Ini.Free;
end;
end;

procedure TprForm.BeforeDestruction;
var
  Ini : TIniFile;
  buf : array [0..254] of char;
begin
inherited;

GetWindowsDirectory(buf,254);
Ini:=TIniFile.Create(prIniFileName);
try
  prSaveProperties(Ini,ClassName);
finally
  Ini.Free;
end;
end;

procedure TprForm.prRestoreProperties;
var
  ws : TWindowState;
begin
//   
ws:=TWindowState(Ini.ReadInteger(sn,'WindowState',integer(wsNormal)));
if ws<>wsMinimized then
  begin
    if ws=wsNormal then
      begin
        Left  :=Ini.ReadInteger(sn,'Left',Left);
        Top   :=Ini.ReadInteger(sn,'Top',Top);
        Width :=Ini.ReadInteger(sn,'Width',Width);
        Height:=Ini.ReadInteger(sn,'Height',Height);
      end
    else
      WindowState:=ws;
  end;
end;

procedure TprForm.prSaveProperties;
begin
Ini.WriteInteger(sn,'WindowState',integer(WindowState));
Ini.WriteInteger(sn,'Left',Left);
Ini.WriteInteger(sn,'Top',Top);
Ini.WriteInteger(sn,'Width',Width);
Ini.WriteInteger(sn,'Height',Height);
end;

/////////////////////////////
//
// TprDesigner
//
/////////////////////////////
constructor TprDesigner.CreateDesigner;
begin
Report              :=_Report;
Report.FDesignerForm:=Self;
inherited Create(AOwner);
end;

procedure TprDesigner.BeforeDestruction;
begin
inherited;
Report.FDesignerForm:=nil;
if Assigned(Report.OnDestroyDesigner) then
  Report.OnDestroyDesigner(Report);
end;

procedure TprDesigner.Loaded;
begin
inherited;
if Report.DesignerFormMode=fmNormal then
  begin
    FormStyle:=fsNormal;
    Visible  :=false;
  end;
end;

procedure TprDesigner.CreateWnd;
begin
if Report.DesignerFormMode=fmNormal then
  begin
    FormStyle:=fsNormal;
    Visible  :=false;
  end;
inherited;
end;

procedure TprDesigner.ConvertToDesignerCoords;
begin
rDest:=rSource;
end;

function TprDesigner.ConvertXToDesignerCoords;
begin
Result:=X;
end;

function TprDesigner.ConvertYToDesignerCoords;
begin
Result:=Y;
end;

procedure TprDesigner.ConvertFromDesignerCoords;
begin
rDest:=rSource;
end;

function TprDesigner.ConvertXFromDesignerCoords;
begin
Result:=X;
end;

function TprDesigner.ConvertYFromDesignerCoords;
begin
Result:=Y;
end;

/////////////////////////////
//
// TprPreview
//
/////////////////////////////
constructor TprPreview.CreatePreview;
begin
Report             :=_Report;
Report.FPreviewForm:=Self;
inherited Create(AOwner);
end;

procedure TprPreview.BeforeDestruction;
begin
inherited;
Report.FPreviewForm:=nil;
if Assigned(Report.OnDestroyPreview) then
  Report.OnDestroyPreview(Report);
end;

procedure TprPreview.Loaded;
begin
inherited;
if Report.PreviewFormMode=fmNormal then
  begin
    FormStyle:=fsNormal;
    Visible  :=false;
  end;
end;

procedure TprPreview.CreateWnd;
begin
if Report.PreviewFormMode=fmNormal then
  begin
    FormStyle:=fsNormal;
    Visible  :=false;
  end;
inherited;
end;

procedure TprPreview.DoOnCustomAction;
begin
if Assigned(Report.OnCustomActionInPreview) then
  Report.OnCustomActionInPreview(Self);
end;

/////////////////////
//
// TprObjRecVersion
//
/////////////////////
constructor TprObjRecVersion.Create;
begin
inherited;
FVisible:=true;
end;

procedure TprObjRecVersion.Assign;
begin
with Source as TprObjRecVersion do
  begin
    Self.Formula                        :=Formula;
    Self.Visible                        :=Visible;
    Self.FCompiledFormula               :=FCompiledFormula;
    Self.FMayBeUse                      :=FMayBeUse;
    Self.FSecondPassCalcCurVersionNeeded:=FSecondPassCalcCurVersionNeeded;
  end;
end;

///////////////////////
//
// TprObjRecVersions
//
///////////////////////
function TprObjRecVersions.GetItm;
begin
Result:=TprObjRecVersion(inherited Items[i]);
end;

/////////////////////
//
// TprObjRec
//
/////////////////////
constructor TprObjRec.Create;
begin
inherited Create;
Page      :=_Page;
Obj       :=_Obj;
DefVersion:=0;
CurVersion:=-1;

FVersions:=CreateVersions;
end;

destructor TprObjRec.Destroy;
begin
FVersions.Free;
inherited;
end;

procedure TprObjRec.Save;
var
  Writer : TWriter;
begin
Stream.Write(FDefVersion,4);
WriteRect(Stream,pRect);

Writer:=TWriter.Create(Stream,1024);
try
  Writer.WriteCollection(FVersions);
finally
  Writer.Free;
end;
end;

procedure TprObjRec.Load;
var
  Reader : TReader;
begin
Stream.Read(FDefVersion,4);
ReadRect(Stream,pRect);

Reader:=TReader.Create(Stream,1024);
try
  Reader.ReadValue;
  Reader.ReadCollection(FVersions);
finally
  Reader.Free;
end;
end;

function TprObjRec.GetLeft;
begin
Result:=pRect.Left;
end;

function TprObjRec.GetTop;
begin
Result:=pRect.Top;
end;

function TprObjRec.GetRight;
begin
Result:=pRect.Right;
end;

function TprObjRec.GetBottom;
begin
Result:=pRect.Bottom;
end;

procedure TprObjRec.SetLeft;
begin
pRect.Left:=value;
end;

procedure TprObjRec.SetTop;
begin
pRect.Top:=value;
end;

procedure TprObjRec.SetRight;
begin
pRect.Right:=value;
end;

procedure TprObjRec.SetBottom;
begin
pRect.Bottom:=value;
end;

procedure TprObjRec.Assign;
begin
with Source as TprObjRec do
  begin
    Self.pRect                          :=pRect;
    Self.X                              :=X;
    Self.Y                              :=Y;
    Self.DX                             :=DX;
    Self.DY                             :=DY;
    Self.DefVersion                     :=DefVersion;
    Self.CurVersion                     :=CurVersion;
    Self.FSecondPassCalcCurVersionNeeded:=FSecondPassCalcCurVersionNeeded;

    Self.FVersions.Assign(Versions);
  end;
end;

procedure TprObjRec.FirstPassCalcCurVersion;
var
  i : integer;
  p : TprParser;
  Res : TprVarValue;
begin
CurVersion:=DefVersion;

for i:=0 to Versions.Count-1 do
  if i<>DefVersion then
    begin
      //    ,     
      // ,     FCompiledFormula
      p:=TprParser(Obj.Band.Report.FParser);
      Versions[i].FCompiledFormula:=Versions[i].Formula;

      Versions[i].FSecondPassCalcCurVersionNeeded:=not p.Calc(Versions[i].FCompiledFormula,Res);
      if not Versions[i].FSecondPassCalcCurVersionNeeded then
        begin
          Versions[i].FMayBeUse:=_vAsBoolean(Res);
          if CurVersion=DefVersion then
            CurVersion:=i;
        end;
    end;
end;

procedure TprObjRec.SecondPass;
var
  i : integer;
begin
//  
CurVersion:=DefVersion;

for i:=0 to Versions.Count-1 do
  if i<>DefVersion then
    begin
      if Versions[i].FSecondPassCalcCurVersionNeeded then
        begin
          if Obj.Band.Report.Calc(Versions[i].FCompiledFormula,nil) then
            begin
              CurVersion:=i;
              break;
            end;
        end
      else
        if Versions[i].FMayBeUse then
          begin
            CurVersion:=i;
            break;
          end;
    end;
end;

//////////////////
//
// TprObjs
//
//////////////////
function TprObjs.GetItm;
begin
Result:=TprObj(inherited Items[index]);
end;

/////////////////////
//
// TprObj
//
/////////////////////
constructor TprObj.Create;
begin
inherited Create(AOwner);

TopObjs   :=TprObjs.Create;
LeftObjs  :=TprObjs.Create;
WidthObjs :=TprObjs.Create;
HeightObjs:=TprObjs.Create;

TopMode   :=prlmMaxBottom;
LeftMode  :=prlmMaxRight;
WidthMode :=prrmMaxRight;
HeightMode:=prrmMaxBottom;

Name      :=GetValidComponentName(Self);

PDI       :=nil;

InitdRec; //  
end;

destructor TprObj.Destroy;
begin
{$IFDEF DEBUG}
WriteToLog('TprObj.Destroy : '+Name);
{$ENDIF}
TopObjs.Free;
LeftObjs.Free;
WidthObjs.Free;
HeightObjs.Free;

FdRec.Free;
aRec.Free;

if PDI<>nil then
  FreeMem(PDI);

{$IFDEF DEBUG}
WriteToLog('TprObj.Destroy : '+Name+' destroyed');
{$ENDIF}
inherited;
end;

procedure TprObj.SetParentComponent;
begin
Band:=Value as TprBand;
end;

function TprObj.GetChildOwner;
begin
Result:=nil;
if Band<>nil then
  Result:=Band.Report;
end;

function TprObj.HasParent;
begin
Result:=true;
end;

function TprObj.GetParentComponent;
begin
Result:=Band;
end;

procedure TprObj.Notification;
begin
{$IFDEF DEBUG}
WriteToLog('TprObj.Notification : '+Name+' '+GetEnumName(TypeInfo(TOperation),integer(Operation))+' AComponent : '+AComponent.Name);
{$ENDIF}
if (Operation=opRemove) and (AComponent<>Self) and (AComponent is TprObj) then
  begin
    TopObjs.Remove(AComponent);
    LeftObjs.Remove(AComponent);
    WidthObjs.Remove(AComponent);
    HeightObjs.Remove(AComponent);
  end;
end;

procedure TprObj.DefineProperties;
begin
inherited;
Filer.DefineProperty('LeftObjs',ReadLeft,WriteLeft,LeftObjs.Count>0);
Filer.DefineProperty('TopObjs',ReadTop,WriteTop,TopObjs.Count>0);
Filer.DefineProperty('WidthObjs',ReadWidth,WriteWidth,WidthObjs.Count>0);
Filer.DefineProperty('HeightObjs',ReadHeight,WriteHeight,HeightObjs.Count>0);
end;

procedure TprObj.ReadLeft;
begin
FLeftObjsNames:=TStringList.Create;
prReadStringList(Reader,FLeftObjsNames);
end;

procedure TprObj.WriteLeft;
begin
prWriteCompListNames(Writer,LeftObjs);
end;

procedure TprObj.ReadTop;
begin
FTopObjsNames:=TStringList.Create;
prReadStringList(Reader,FTopObjsNames);
end;

procedure TprObj.WriteTop;
begin
prWriteCompListNames(Writer,TopObjs);
end;

procedure TprObj.ReadWidth;
begin
FWidthObjsNames:=TStringList.Create;
prReadStringList(Reader,FWidthObjsNames);
end;

procedure TprObj.WriteWidth;
begin
prWriteCompListNames(Writer,WidthObjs);
end;

procedure TprObj.ReadHeight;
begin
FHeightObjsNames:=TStringList.Create;
prReadStringList(Reader,FHeightObjsNames);
end;

procedure TprObj.WriteHeight;
begin
prWriteCompListNames(Writer,HeightObjs);
end;

procedure TprObj.AfterReportLoaded;

  procedure Update(LO : TprObjs; LS : TStrings);
  var
    i : integer;
    c : TComponent;
  begin
  if LS<>nil then
    for i:=0 to LS.Count-1 do
      begin
        c:=Band.Report.FindComponent(LS[i]);
        if c<>nil then
          LO.Add(c);
      end;
  end;

begin
try
  Update(LeftObjs,FLeftObjsNames);
  Update(TopObjs,FTopObjsNames);
  Update(WidthObjs,FWidthObjsNames);
  Update(HeightObjs,FHeightObjsNames);
finally
  FLeftObjsNames.Free;
  FTopObjsNames.Free;
  FWidthObjsNames.Free;
  FHeightObjsNames.Free;

  FLeftObjsNames  :=nil;
  FTopObjsNames   :=nil;
  FWidthObjsNames :=nil;
  FHeightObjsNames:=nil;
end;
end;

procedure TprObj.SetBand;
begin
if FBand=Value then exit;
if FBand<>nil then
  FBand.Objects.Remove(Self);
FBand:=Value;
if (FBand<>nil) and (FBand.Objects.IndexOf(Self)=-1) then
  FBand.Objects.Add(Self);
end;

function TprObj.GetReport;
begin
Result:=nil;
if Band<>nil then
  Result:=Band.Report;
end;

function TprObj.AllowInplaceEdit;
begin
Result:=false;
end;

procedure TprObj.InplaceEdit;
begin
end;

procedure TprObj.SaveInplaceEdit;
begin
end;

function TprObj.GetDesc;
begin
Result:=Name;
end;

procedure TprObj.UpdatePDIdPageRect;
begin
Band.Report.ConvertToDesignerCoords(dRec.pRect,pdi.dPageRect);
pdi.dPageRect.Left  :=pdi.dPageRect.Left  +Band.pdi.dPageRect.Left;
pdi.dPageRect.Top   :=pdi.dPageRect.Top   +Band.pdi.dPageRect.Top;
pdi.dPageRect.Right :=pdi.dPageRect.Right +Band.pdi.dPageRect.Left;
pdi.dPageRect.Bottom:=pdi.dPageRect.Bottom+Band.pdi.dPageRect.Top;
end;

procedure TprObj.UpdatePDI;
begin
if PDI=nil then
  begin
    GetMem(PDI,sizeof(rDesignInfo));
    with PDI^ do
      begin
        AllowResizeTypes :=[ppLeftTop,ppTop,ppRightTop,ppRight,ppRightBottom,ppBottom,ppLeftBottom,ppLeft];
        AllowDrag        :=true;
        AllowLinkTypes   :=[ltLeft,ltTop,ltRight,ltBottom];
        Parent           :=Self;
      end;
  end;
UpdatePDIdPageRect;
end;

procedure TprObj.DesignerResize;
begin
if not RectInRect(Rect(dRec.pRect.Left+oLeft+Band.dPageRect.Left,
                       dRec.pRect.Top+oTop+Band.dPageRect.Top,
                       dRec.pRect.Right+oRight+Band.dPageRect.Left,
                       dRec.pRect.Bottom+oBottom+Band.dPageRect.Top),
                  Band.dPageRect) then exit; //        

dRec.pRect.Top   :=dRec.pRect.Top+oTop;
dRec.pRect.Left  :=dRec.pRect.Left+oLeft;
dRec.pRect.Bottom:=dRec.pRect.Bottom+oBottom;
dRec.pRect.Right :=dRec.pRect.Right+oRight;
UpdatePDIdPageRect;
end;

procedure TprObj.DesignerDrag;
begin
if not RectInRect(Rect(dRec.pRect.Left+dx+Band.dPageRect.Left,
                       dRec.pRect.Top+dy+Band.dPageRect.Top,
                       dRec.pRect.Right+dx+Band.dPageRect.Left,
                       dRec.pRect.Bottom+dy+Band.dPageRect.Top),
                  Band.dPageRect) then exit; //        

dRec.pRect.Top   :=dRec.pRect.Top+dy;
dRec.pRect.Left  :=dRec.pRect.Left+dx;
dRec.pRect.Bottom:=dRec.pRect.Bottom+dy;
dRec.pRect.Right :=dRec.pRect.Right+dx;
UpdatePDIdPageRect;
end;

procedure TprObj.DesignerLink;
begin
if (Linked.Parent is TprObj) and
   (Band=TprObj(Linked.Parent).Band) then
  begin
    case LinkMode of
      ltLeft  : if LeftObjs.IndexOf(Linked.Parent)=-1 then
                  LeftObjs.Add(Linked.Parent);
      ltTop   : if TopObjs.IndexOf(Linked.Parent)=-1 then
                  TopObjs.Add(Linked.Parent);
      ltRight : if WidthObjs.IndexOf(Linked.Parent)=-1 then
                  WidthObjs.Add(Linked.Parent);
      ltBottom: if HeightObjs.IndexOf(Linked.Parent)=-1 then
                  HeightObjs.Add(Linked.Parent);
    end;
    LinkAccepted:=true;
  end;
end;

procedure TprObj.DesignerDelete;
begin
Free;
end;

function TprObj.IsLinkedTo;
begin
Result:=false;
case LinkMode of
  prlLeft  : Result:=LeftObjs.IndexOf(Obj)<>-1;
  prlTop   : Result:=TopObjs.IndexOf(Obj)<>-1;
  prlWidth : Result:=WidthObjs.IndexOf(Obj)<>-1;
  prlHeight: Result:=HeightObjs.IndexOf(Obj)<>-1;
end;
end;

//
//   aRec     
//   DX  DY (       )
//
procedure TprObj.FirstPass;
var
  MinBottomObj,MaxBottomObj,MinRightObj,MaxRightObj : TprObj;

  procedure CalcObjs(o : TprObjs; var MinY,MaxY,MinX,MaxX : TprObj; LinkMode : TprLinkMode);
  var
    i : integer;
    fCrlLink : boolean;
  begin
  if not o[0].FirstPassProcessed then
    o[0].FirstPass;
    
  MinY    :=o[0];
  MaxY    :=o[0];
  MinX    :=o[0];
  MaxX    :=o[0];
  fCrlLink:=o[0].IsLinkedTo(LinkMode,Self);
  for i:=1 to o.Count-1 do
    begin
      if not o[i].FirstPassProcessed then
        o[i].FirstPass;

      fCrlLink:=fCrlLink or o[i].IsLinkedTo(LinkMode,Self);

      if o[i].aRec.Y+o[i].aRec.DY>MaxY.aRec.Y+MaxY.aRec.DY then
        MaxY:=o[i];
      if o[i].aRec.Y+o[i].aRec.DY<MinY.aRec.Y+MinY.aRec.DY then
        MinY:=o[i];
      if o[i].aRec.X+o[i].aRec.DX>MaxX.aRec.X+MaxX.aRec.DX then
        MaxX:=o[i];
      if o[i].aRec.X+o[i].aRec.DX<MinX.aRec.X+MinX.aRec.DX then
        MinX:=o[i];
    end;
    
  if fCrlLink then
    begin
      if Self.aRec.Y+Self.aRec.DY>MaxY.aRec.Y+MaxY.aRec.DY then
        MaxY:=Self;
      if Self.aRec.Y+Self.aRec.DY<MinY.aRec.Y+MinY.aRec.DY then
        MinY:=Self;
      if Self.aRec.X+Self.aRec.DX>MaxX.aRec.X+MaxX.aRec.DX then
        MaxX:=Self;
      if Self.aRec.X+Self.aRec.DX<MinX.aRec.X+MinX.aRec.DX then
        MinX:=Self;
    end;
  end;

begin
if aRec=nil then
  raise Exception.Create(prLoadStr(saRecNotInitializated));
FirstPassProcessed:=true;

aRec.X :=dRec.pRect.Left;
aRec.Y :=dRec.pRect.Top;

//  TopObjs
if TopObjs.Count>0 then
  begin
    CalcObjs(TopObjs,MinBottomObj,MaxBottomObj,MinRightObj,MaxRightObj,prlTop);

    case TopMode of
      prlmMaxBottom:  //    ,     Bottom
        begin
          if MaxBottomObj<>Self then
            aRec.Y:=aRec.Y+(MaxBottomObj.aRec.DY-(MaxBottomObj.dRec.pRect.Bottom-MaxBottomObj.dRec.pRect.Top));
        end;
      prlmMinBottom:  //    
        begin
          if MinBottomObj<>Self then
            aRec.Y:=aRec.Y+(MinBottomObj.aRec.DY-(MinBottomObj.dRec.pRect.Bottom-MinBottomObj.dRec.pRect.Top));
        end;
    end;
  end;

//  LeftObjs
if LeftObjs.Count>0 then
  begin
    CalcObjs(LeftObjs,MinBottomObj,MaxBottomObj,MinRightObj,MaxRightObj,prlLeft);

    case LeftMode of
      prlmMaxRight:  //    ,     Right
        begin
          if MaxRightObj<>Self then
            aRec.X:=aRec.X+(MaxRightObj.aRec.DX-(MaxRightObj.dRec.pRect.Right-MaxRightObj.dRec.pRect.Left));
        end;
      prlmMinRight:  //    
        begin
          if MinRightObj<>Self then
            aRec.X:=aRec.X+(MinRightObj.aRec.DX-(MinRightObj.dRec.pRect.Right-MinRightObj.dRec.pRect.Left));
        end;
    end;
  end;

//  WidthObjs
if WidthObjs.Count>0 then
  begin
    CalcObjs(WidthObjs,MinBottomObj,MaxBottomObj,MinRightObj,MaxRightObj,prlWidth);

    case WidthMode of
      prrmMaxRight:  //    ,     Right
        begin
          if MaxRightObj<>Self then
            aRec.DX:=dRec.pRect.Right-dRec.pRect.Left+MaxRightObj.aRec.DX-(MaxRightObj.dRec.pRect.Right-MaxRightObj.dRec.pRect.Left);
        end;
      prrmMinRight:  //    
        begin
          if MinRightObj<>Self then
            aRec.DX:=dRec.pRect.Right-dRec.pRect.Left+MinRightObj.aRec.DX-(MinRightObj.dRec.pRect.Right-MinRightObj.dRec.pRect.Left);
        end;
    end;
  end;


//  HeightObjs
if HeightObjs.Count>0 then
  begin
    CalcObjs(HeightObjs,MinBottomObj,MaxBottomObj,MinRightObj,MaxRightObj,prlHeight);

    case HeightMode of
      prrmMaxBottom:  //    ,     Right
        begin
          if MaxBottomObj<>Self then
            aRec.DY:=dRec.pRect.Bottom-dRec.pRect.Top+MaxBottomObj.aRec.DY-(MaxBottomObj.dRec.pRect.Bottom-MaxBottomObj.dRec.pRect.Top);
        end;
      prrmMinBottom:  //    
        begin
          if MinBottomObj<>Self then
            aRec.DY:=dRec.pRect.Bottom-dRec.pRect.Top+MinBottomObj.aRec.DY-(MinBottomObj.dRec.pRect.Bottom-MinBottomObj.dRec.pRect.Top);
        end;
    end;
  end;
end;

procedure TprObj.DrawDesign;
begin
end;

////////////////////////
//
// TprValueVersion
//
////////////////////////
constructor TprValueVersion.Create;
begin
inherited;
end;

////////////////////////
//
// TprValue
//
////////////////////////
constructor TprValue.Create;
var
  i : integer;
  cn : string;
begin
inherited;
DataSet            :=TprDatasetLink.Create;
ResetDataSet       :=TprDatasetLink.Create;
CrossTabHorzDataSet:=TprDatasetLink.Create;

SavedIndex :=0;
DataSetName:='';
Group      :=nil;

FVersions  :=TList.Create;

i:=1;
cn:=ClassName;
while Report.Values.IndexByName(Copy(cn,2,Length(cn))+IntToStr(i))<>-1 do Inc(i);
Name:=Copy(cn,2,Length(cn))+IntToStr(i);
end;

destructor TprValue.Destroy;
begin
Clear;
Dataset.Free;
ResetDataset.Free;
CrossTabHorzDataset.Free;
inherited;
end;

function TprValue.Report;
begin
Result:=TprValues(Collection).Report;
end;

procedure TprValue.SetName;
begin
if CompText(Value,Name)=0 then
  exit;
if Report.Values.IndexByName(Value)<>-1 then
  raise Exception.CreateFmt(prLoadStr(sVarAlreadyExists),[Value,Report.Name]);
FName:=Value;
end;

function TprValue.GetVersion;
begin
Result:=TprValueVersion(FVersions[index]);
end;

function TprValue.VersionsCount;
begin
Result:=FVersions.Count;
end;

function TprValue.GetCurrentVersionID;
begin
if CalcOn=cvtCrossTab then
  begin
    Result:=Format('%s_%d',[Name,Report.GetDataSetRecNo(CrossTabHorzDataSet.Dataset)+SavedIndex-1])
  end
else
  begin
    if FCurrentValueExists then
      Result:=Format('%s_%d',[Name,FVersions.Count-1])
    else
      Result:=Format('%s_%d',[Name,FVersions.Count])
  end;
end;

procedure TprValue.SetCurrentValue;
var
  ver : TprValueVersion;
begin
if CalcOn=cvtCrossTab then
  begin
    if Report.GetDataSetRecNo(CrossTabHorzDataSet)+SavedIndex-1<VersionsCount then
      Versions[Report.GetDataSetRecNo(CrossTabHorzDataSet)+SavedIndex-1].V:=Value
    else
      begin
        ver      :=TprValueVersion.Create;
        ver.Value:=Self;
        ver.ID   :=FVersions.Count;
        ver.V    :=Value;
        FVersions.Add(ver);
      end;
  end
else
  begin
    if FCurrentValueExists then
      Versions[FVersions.Count-1].V:=Value
    else
      begin
        ver      :=TprValueVersion.Create;
        ver.Value:=Self;
        ver.ID   :=FVersions.Count;
        ver.V    :=Value;
        FVersions.Add(ver);
        FCurrentValueExists:=true;
      end;
  end;
end;

function TprValue.GetCurrentVersion : TprValueVersion;
var
  fCreated : boolean;
begin
Result:=GetCurrentVersion(fCreated);
end;

function TprValue.GetCurrentVersion(var fCreated : boolean) : TprValueVersion;
begin
fCreated:=false;
if CalcOn=cvtCrossTab then
  begin
    if Report.GetDataSetRecNo(CrossTabHorzDataSet.Dataset)+SavedIndex-1<VersionsCount then
      Result:=Versions[Report.GetDataSetRecNo(CrossTabHorzDataSet.Dataset)+SavedIndex-1]
    else
      begin
        Result  :=TprValueVersion.Create;
        fCreated:=true;
        FVersions.Add(Result);
      end;
  end
else
  begin
    if FCurrentValueExists then
      Result:=Versions[FVersions.Count-1]
    else
      begin
        Result  :=TprValueVersion.Create;
        fCreated:=true;
        FVersions.Add(Result);
        FCurrentValueExists:=true;
      end;
  end;
end;

procedure TprValue.Reset;
var
  v : TprValueVersion;
  fCreated : boolean;
begin
v:=GetCurrentVersion(fCreated);
if fCreated then
  begin
    v.V :=NULL;
    v.V1:=NULL;
  end;

case CalcOn of
  cvtEventOnReset:
    begin
      if Assigned(OnCalc) then
        OnCalc(Self)
      else
        raise Exception.CreateFmt(prLoadStr(sReportVarNotCalced),[Name]);
    end;
  cvtCrossTab:
    begin
      SavedIndex:=VersionsCount;
    end
  else
    begin
      FCurrentValueExists:=false;
    end
end;
end;

procedure TprValue.Init;

  procedure GetDataSet(DataSet : TprDatasetLink; DataSetName : string);
  begin
  Dataset.Dataset:=Report.GetDataSetByName(DataSetName);
  if DataSet.Dataset=nil then
    raise Exception.CreateFmt(prLoadStr(sDataSetNotFound),[DataSetName]);
  end;

begin
if CalcOn in [cvtDataSetNext,cvtCrossTab] then
  GetDataSet(DataSet,DataSetName);
if ResetOn in [rvtDataSetEof] then
  GetDataSet(ResetDataSet,ResetDataSetName);
if CalcOn  in [cvtCrossTab] then
  GetDataSet(CrossTabHorzDataSet,CrossTabHorzDataSetName);
end;

procedure TprValue.Clear;
begin
FCurrentValueExists:=false;
SavedIndex:=0;
while FVersions.Count>0 do
  begin
    Versions[0].Free;
    FVersions.Delete(0);
  end;
end;

/////////////////////////
//
// TprValues
//
/////////////////////////
function TprValues.GetItm;
begin
Result:=TprValue(inherited Items[index]);
end;

function TprValues.GetByName;
var
  i : integer;
begin
i:=IndexByName(Name);
if i=-1 then
  raise Exception.CreateFmt(prLoadStr(sVarNotExists),[Name])
else
  Result:=Items[i];
end;

function TprValues.IndexByName;
begin
Result:=0;
while (Result<Count) and
      (CompText(Items[Result].Name,Name)<>0) do Inc(Result);
if Result>=Count then
  Result:=-1;
end;

function TprValues.VersionByVersionId;
var
  name : string;
  p,ind : integer;
begin
Result:=nil;
p     :=PosLastChar(id,'_');
if p<>0 then
  begin
    ind :=StrToIntDef(Copy(id,p+1,length(id)),-1);
    if Ind<>-1 then
      begin
        name:=Copy(id,1,p-1);
        p   :=IndexByName(name);
        if p<>-1 then
          begin
            if ind<Items[p].VersionsCount then
              begin
                Result:=Items[p].Versions[ind];
                if Assigned(Items[p].OnGetVersionByVersionID) then
                  Items[p].OnGetVersionByVersionID(Result);
              end
            else
              raise Exception.CreateFmt(prLoadStr(sNoVarVersion),[name,ind]);
          end;
      end;
  end;
end;

/////////////////////////
//
// TprGroup
//
/////////////////////////
constructor TprGroup.Create;
begin
inherited;

Headers   :=TprBands.Create;
Footers   :=TprBands.Create;

FPrevValue:=Unassigned;

Name      :=GetValidComponentName(Self);
end;

destructor TprGroup.Destroy;
begin
Headers.Free;
Footers.Free;
Report:=nil;

inherited;
end;

function TprGroup.GetChildOwner;
begin
Result:=Report;
end;

procedure TprGroup.SetParentComponent;
begin
Report:=Value as TprCustomReport;
end;

function TprGroup.HasParent;
begin
Result:=true;
end;

function TprGroup.GetParentComponent;
begin
Result:=Report;
end;

procedure TprGroup.SetReport;
begin
if FReport=Value then exit;
if FReport<>nil then
  FReport.Groups.Remove(Self);
FReport:=Value;
if (FReport<>nil) and (FReport.Groups.IndexOf(Self)=-1) then
  FReport.Groups.Add(Self);
end;

procedure TprGroup.SetDetailBand;
begin
if Value=FDetailBand then exit;
if FDetailBand<>nil then
  FDetailBand.GroupRemoved(Self);
FDetailBand:=Value;
if FDetailBand<>nil then
  FDetailBand.GroupAdded(Self);
end;

function TprGroup.IndexInReport;
begin
Result:=Report.Groups.IndexOf(Self);
end;

procedure TprGroup.Reset;
begin
FLineNo   :=0;
FPrevValue:=Unassigned;
end;

procedure TprGroup.CalcValue;
var
  CurValue : Variant;
begin
FLineNo     :=FLineNo+1;     
CurValue    :=Report.Calc(FValid,nil);
FNeedHeaders:=VarIsEmpty(FPrevValue) or (FPrevValue<>CurValue);
FNeedFooters:=not VarIsEmpty(FPrevValue) and (FPrevValue<>CurValue);
if FNeedFooters then
  FLineNo:=1;
FPrevValue  :=CurValue;
end;

procedure TprGroup.HeadersGenerateCell;
var
  i : integer;
begin
if FNeedHeaders then
  //   
  for i:=0 to Headers.Count-1 do
    Headers[i].GenerateCell(CallerBand,CallbackProc,false);
end;

procedure TprGroup.FootersGenerateCell;
var
  i : integer;
begin
if FNeedFooters then
  begin
    //   
    for i:=0 to Footers.Count-1 do
      Footers[i].GenerateCell(CallerBand,CallbackProc,false);
    Report.DoOnGroupEnd(Self);
  end;
end;

procedure TprGroup.FootersAlwaysGenerateCell;
var
  i : integer;
begin
//   
for i:=0 to Footers.Count-1 do
  Footers[i].GenerateCell(CallerBand,CallbackProc,false);
Report.DoOnGroupEnd(Self);
end;

///////////////////////////////
//
// TprGenCell
//
///////////////////////////////
constructor TprGenCell.Create;
begin
inherited;
FObjRecs:=TList.Create;
end;

destructor TprGenCell.Destroy;
begin
FObjRecs.Free;
inherited;
end;

procedure TprGenCell.Add;
begin
FObjRecs.Add(ObjRec);
end;

function TprGenCell.GetObjRec;
begin
Result:=TprObjRec(FObjRecs[i]);
end;

function TprGenCell.ObjRecsCount;
begin
Result:=FObjRecs.Count;
end;

///////////////////////////////
//
// TprGenVector
//
///////////////////////////////
constructor TprGenVector.Create;
begin
inherited Create;
Band:=_Band;
end;

procedure TprGenVector.AddCell;
begin
inherited Add(Cell);
end;

function TprGenVector.GetItm;
begin
Result:=TprGenCell(inherited Items[i]);
end;

///////////////////////////////
//
// TprGenGrid
//
///////////////////////////////
constructor TprGenGrid.Create;
begin
inherited Create;
Page            :=_Page;
FCells          :=TList.Create;
FVVectors       :=TList.Create;
FHVectors       :=TList.Create;
FInsertedVectors:=TList.Create;
end;

destructor TprGenGrid.Destroy;
begin
Clear;
FCells.Free;
FVVectors.Free;
FHVectors.Free;
FInsertedVectors.Free;
inherited Destroy;
end;

procedure TprGenGrid.Clear;

  procedure FreeVectorsList(L : TList);
  begin
  while L.Count>0 do
    begin
      TprGenVector(L[0]).Free;
      L.Delete(0);
    end;
  end;

begin
FCurHVector     :=nil;
FCurVVectorIndex:=0;

FreeVectorsList(FVVectors);
FreeVectorsList(FHVectors);
FreeVectorsList(FInsertedVectors);
while FCells.Count>0 do
  begin
    Cells[0].Free;
    FCells.Delete(0);
  end;
end;

function TprGenGrid.AddCellSimple;
begin
Result:=TprGenCell.Create;
FCells.Add(Result);
end;

function TprGenGrid.AddCell;
begin
Result:=TprGenCell.Create;
if _HBand<>nil then
  begin
    if FCurHVector=nil then
      begin
        FCurHVector:=TprGenVector.Create(_HBand);
        FHVectors.Add(FCurHVector);
      end;
  end;

if _VBand<>nil then
  begin
    Inc(FCurVVectorIndex);
    if FCurVVectorIndex>=VVectorsCount then
      FVVectors.Add(TprGenVector.Create(_VBand));
  end;

FCells.Add(Result);
if (_HBand<>nil) and (FCurHVector<>nil) then
  begin
    Result.HVector:=FCurHVector;
    FCurHVector.AddCell(Result);
  end;
if (FCurVVectorIndex>=0) and (FCurVVectorIndex<VVectorsCount) then
  begin
    Result.VVector:=VVectors[FCurVVectorIndex];
    Result.VVector.AddCell(Result);
  end;

Page.Report.UpdateProgressForm(Format(prLoadStr(sFirstPassCaption),[HVectorsCount,VVectorsCount,CellsCount]));
end;

procedure TprGenGrid.EndOfLine;
begin
FCurHVector     :=nil;
FCurVVectorIndex:=-1;
end;

function TprGenGrid.VVectorsCount;
begin
Result:=FVVectors.Count;
end;

function TprGenGrid.HVectorsCount;
begin
Result:=FHVectors.Count;
end;

function TprGenGrid.CellsCount;
begin
Result:=FCells.Count;
end;

function TprGenGrid.GetVVector;
begin
Result:=TprGenVector(FVVectors[i]);
end;

function TprGenGrid.GetHVector;
begin
Result:=TprGenVector(FHVectors[i]);
end;

function TprGenGrid.GetCell;
begin
Result:=TprGenCell(FCells[i]);
end;

function TprGenGrid.FindPriorVVector;
var
  i : integer;
begin
i:=FromIndex-1;
while (i>=MinEndPageIndex) and (VVectors[i].Band<>Band) do Dec(i);
if i<MinEndPageIndex then
  Result:=nil
else
  Result:=VVectors[i];
end;

function TprGenGrid.FindPriorVVectorOnPage;
var
  i : integer;
begin
i:=FromIndex-1;
while (i>=0) and (VVectors[i].Page>=EndPagePageX) and (VVectors[i].Band<>Band) do Dec(i);
if (i>=0) and (VVectors[i].Page=EndPagePageX) and (VVectors[i].Band=Band) then
  Result:=VVectors[i]
else
  Result:=nil;
end;

//   SourceVector    
//   Index
function TprGenGrid.CopyVVector;
var
  i,j : integer;
  Cell : TprGenCell;
begin
Result       :=TprGenVector.Create(SourceVector.Band);
Result.Width :=SourceVector.Width;
Result.Height:=SourceVector.Height;
for i:=0 to SourceVector.Count-1 do
  begin
    Cell        :=AddCellSimple;
    Cell.Height :=SourceVector[i].Height;
    Cell.Width  :=SourceVector[i].Width;
    Cell.VVector:=Result;
    Cell.HVector:=SourceVector[i].HVector;

    for j:=0 to SourceVector[i].ObjRecsCount-1 do
      Cell.Add(SourceVector[i].ObjRecs[j].CreateCopy);

    Result.AddCell(Cell);
  end;
FInsertedVectors.Add(Result);
end;

function TprGenGrid.IndexOfVVector;
begin
Result:=FVVectors.IndexOf(Vector);
end;


function TprGenGrid.FindPriorHVector;
var
  i : integer;
begin
i:=FromIndex-1;
while (i>=MinEndPageIndex) and (HVectors[i].Band<>Band) do Dec(i);
if i<MinEndPageIndex then
  Result:=nil
else
  Result:=HVectors[i];
end;

function TprGenGrid.FindPriorHVectorOnPage;
var
  i : integer;
begin
i:=FromIndex-1;
while (i>=0) and (HVectors[i].Page>=EndPagePageY) and (HVectors[i].Band<>Band) do Dec(i);
if (i>=0) and (HVectors[i].Page=EndPagePageY) and (HVectors[i].Band=Band) then
  Result:=HVectors[i]
else
  Result:=nil;
end;

//   SourceVector    
//   Index
function TprGenGrid.CopyHVector;
var
  i,j : integer;
  Cell : TprGenCell;
begin
Result       :=TprGenVector.Create(SourceVector.Band);
Result.Width :=SourceVector.Width;
Result.Height:=SourceVector.Height;
for i:=0 to SourceVector.Count-1 do
  begin
    Cell        :=AddCellSimple;
    Cell.Height :=SourceVector[i].Height;
    Cell.Width  :=SourceVector[i].Width;
    Cell.HVector:=Result;
    Cell.VVector:=SourceVector[i].VVector;

    for j:=0 to SourceVector[i].ObjRecsCount-1 do
      Cell.Add(SourceVector[i].ObjRecs[j].CreateCopy);

    Result.AddCell(Cell);
  end;
FInsertedVectors.Add(Result);
end;

function TprGenGrid.IndexOfHVector;
begin
Result:=FHVectors.IndexOf(Vector);
end;


/////////////////////////
//
// TprGroups
//
/////////////////////////
function TprGroups.GetItm;
begin
Result:=TprGroup(inherited Items[index]);
end;

function TprGroups.GetByName;
var
  i : integer;
begin
i:=IndexByName(Name);
if i=-1 then
  raise Exception.CreateFmt(prLoadStr(sGroupNotExists),[name])
else
  result:=Items[i];
end;

function TprGroups.IndexByName;
begin
Result:=0;
while (Result<Count) and (CompText(Items[Result].Name,Name)<>0) do Inc(Result);
if Result>=Count then
  Result:=-1;
end;

///////////////////////
//
// TprBand
//
///////////////////////
constructor TprBand.Create;
var
  i : integer;
  ClassRef: TClass;
begin
{$IFDEF DEBUG}
if AOwner=nil then
  WriteToLog('TprBand.Create : AOwner : nil')
else
  WriteToLog('TprBand.Create : AOwner : '+AOwner.Name);
{$ENDIF}

inherited Create(AOwner);

FVisible  :=true;
PDI       :=nil;
ResizeMode:=prbrmNone;
ResizeObjs:=TprObjs.Create;
Objects   :=TprObjs.Create;


ClassRef := ClassType;
while ClassRef<>nil do
  begin
    i:=GetEnumValue(TypeInfo(TprBandType),'bt'+Copy(ClassRef.ClassName,10,Length(ClassRef.ClassName)-13));
    if i<>-1 then
      begin
        BandType:=TprBandType(i);
        break;
      end;
    ClassRef:=ClassRef.ClassParent;
  end;

Name:=GetValidComponentName(Self);
end;

destructor TprBand.Destroy;
begin
{$IFDEF DEBUG}
WriteToLog('TprBand.Destroy : '+Name);
{$ENDIF}
if PDI<>nil then
  FreeMem(PDI);
while Objects.Count>0 do
  Objects[0].Free;

Objects.Free;
ResizeObjs.Free;

{$IFDEF DEBUG}
WriteToLog('TprBand.Destroy : '+Name+' destroyed');
{$ENDIF}
inherited;
end;

function TprBand.GetChildOwner;
begin
Result:=Report;
end;

procedure TprBand.SetParentComponent;
begin
Page:=Value as TprCustomPage;
end;

function TprBand.HasParent;
begin
Result:=true;
end;

function TprBand.GetParentComponent;
begin
Result:=Page;
end;

procedure TprBand.GetChildren;
var
  i : integer;
begin
for i:=0 to Objects.Count-1 do
  Proc(Objects[i]);
end;

procedure TprBand.Notification;
begin
{$IFDEF DEBUG}
if Owner=nil then
  WriteToLog('TprBand.Notification : '+Name+' '+GetEnumName(TypeInfo(TOperation),integer(Operation))+' AComponent : '+AComponent.Name+' Owner : nil')
else
  WriteToLog('TprBand.Notification : '+Name+' '+GetEnumName(TypeInfo(TOperation),integer(Operation))+' AComponent : '+AComponent.Name+' Owner : '+Owner.Name);
{$ENDIF}
if (Operation = opRemove) and (AComponent<>Self) and (AComponent is TprObj) then
  begin
    Objects.Remove(AComponent);
    ResizeObjs.Remove(AComponent);
  end;
end;

procedure TprBand.DefineProperties;
begin
inherited;
Filer.DefineProperty('ResizeObjs',ReadResizeObjs,WriteResizeObjs,ResizeObjs.Count>0);
end;

procedure TprBand.ReadResizeObjs;
begin
FResizeObjsNames:=TStringList.Create;
prReadStringList(Reader,FResizeObjsNames);
end;

procedure TprBand.WriteResizeObjs;
begin
prWriteCompListNames(Writer,ResizeObjs);
end;

procedure TprBand.GroupRemoved;
begin
end;

procedure TprBand.GroupAdded;
begin
end;

procedure TprBand.AfterReportLoaded;
var
  i,j : integer;
begin
if FResizeObjsNames<>nil then
  begin
    for i:=0 to FResizeObjsNames.Count-1 do
      begin
        j:=0;
        while (j<Objects.Count) and (CompText(Objects[j].Name,FResizeObjsNames[i])<>0) do Inc(j);
        if j<Objects.Count then
          ResizeObjs.Add(Objects[j]);
      end;
    FResizeObjsNames.Free;
  end;
end;

procedure TprBand.SetPage;
begin
if FPage=Value then exit;
if FPage<>nil then
  Page.Bands.Remove(Self);
FPage:=Value;
if (FPage<>nil) and (FPage.Bands.IndexOf(Self)=-1) then
  FPage.Bands.Add(Self);
end;

function TprBand.GetFullBandName;
begin
Result:=Name;
end;

function TprBand.Report;
begin
Result:=nil;
if Page<>nil then
  Result:=Page.Report;
end;

procedure TprBand.InitDataset;
begin
end;

//  , ,     
//   
function TprBand.GetDrawDesignCaption;
begin
Result:=BandTitles[BandType];
end;

//      ,
//   Exception    
procedure TprBand.OnInsertIntoPage;
begin
if p.Bands.IndexByBandType(BandType)<>-1 then
  raise Exception.CreateFmt(prLoadStr(sBandAlreadyExistsInPageType),[GetDrawDesignCaption]);
end;

//  PDI
procedure TprBand.UpdatePDI;
begin
if PDI=nil then
  begin
    GetMem(PDI,sizeof(rDesignInfo));
    with PDI^ do
      begin
        if BandType in HorizontalBands then
          AllowResizeTypes:=[ppTop,ppBottom]
        else
          AllowResizeTypes:=[ppLeft,ppRight];
        AllowDrag        :=false;
        AllowLinkTypes   :=[ltBottom];
        Parent           :=Self;
      end;
  end;
Report.ConvertToDesignerCoords(dPageRect,PDI.dPageRect);
end;

procedure TprBand.DesignerResize;
begin
end;

procedure TprBand.DesignerLink;
begin
end;

procedure TprBand.DesignerDelete;
begin
Free;
end;

function TprBand.GetStartNewPage;
begin
Result:=false;
end;

function TprBand.GetLinkToBand;
begin
Result:=nil;
end;

function TprBand.GetLinkType;
begin
Result:=prltAfter;
end;

//////////////////////////
//
// TprBands
//
//////////////////////////
function TprBands.GetItm;
begin
Result:=TprBand(inherited Items[index]);
end;

function TprBands.GetByBandType;
var
  i : integer;
begin
i:=IndexByBandType(BandType);
if i=-1 then
  raise Exception.CreateFmt(prLoadStr(sUnknownBandType),[GetEnumName(TypeInfo(TprBandType),integer(BandType))])
else
  Result:=Items[i];
end;

function TprBands.GetByName;
var
  i : integer;
begin
i:=IndexByName(Name);
if i=-1 then
  raise Exception.CreateFmt(prLoadStr(sUnknownBandName),[Name])
else
  Result:=Items[i];
end;

function TprBands.IndexByBandType;
begin
Result:=0;
while (Result<Count) and (Items[Result].BandType<>BandType) do Inc(Result);
if Result>=Count then
  Result:=-1;
end;

function TprBands.IndexByName;
begin
Result:=0;
while (Result<Count) and (CompText(Items[Result].Name,Name)<>0) do Inc(Result);
if Result>=Count then
  Result:=-1;
end;

/////////////////////////////
//
// TprCustomHBand
//
/////////////////////////////
constructor TprCustomHBand.Create;
begin
inherited;
end;

function TprCustomHBand.PrintOnStartPage;
begin
Result:=false;
end;

procedure TprCustomHBand.CalcdPageRect;
begin
if FCalced then exit;
dPageRect      :=Rect(CurPageRect.Left,CurPageRect.Top,CurPageRect.Right,CurPageRect.Top+Height);
CurPageRect.Top:=CurPageRect.Top+Height;
FCalced        :=true;
end;

procedure TprCustomHBand.DesignerResize;
begin
Height:=dPageRect.Bottom+oBottom-dPageRect.Top-oTop;
Report.ConvertToDesignerCoords(dPageRect,PDI.dPageRect);
end;

procedure TprCustomHBand.DesignerLink;
begin
if (Linked.Parent is TprObj) and
   (Self=TprObj(Linked.Parent).Band) then
  begin
    case LinkMode of
      ltBottom : if ResizeObjs.IndexOf(Linked.Parent)=-1 then
                   ResizeObjs.Add(Linked.Parent);
    end;
    LinkAccepted:=true;
  end;
end;

//
//
//
function TprCustomHBand.GenerateCell;
var
  i : integer;
  o : TprObjRec;
  MinObj,MaxObj : TprObj;
begin
Report.DoOnBandGenerateCell(Self);
Result:=nil;
if not Visible then exit;
if Page.HasVerticalBands and UseVerticalBands then
  begin
    //     
    for i:=0 to Page.Bands.Count-1 do
      if Page.Bands[i].BandType in VerticalBands then
        Page.Bands[i].InitDataSet;

    //      Detail  ,
    //  ,        
    //   Detail
    for i:=0 to Page.Bands.Count-1 do
      if (Page.Bands[i].BandType=btvDetail) and (TprCustomVDetailBand(Page.Bands[i]).ParentDetail=nil) then
        TprCustomVBand(Page.Bands[i]).GenerateCell(Self,GenerateCellCallback,false);

    //  
    i:=Page.Bands.IndexByBandType(btvSummary);
    if i<>-1 then
      Page.Bands[i].GenerateCell(nil,nil,false);
  end
else
  begin
    //     ,
    //   ,    
    if fSimpleCreateCell then
      Result:=TprGenCell.Create
    else
      Result:=Page.Grid.AddCell(Self,nil);

    //    
    for i:=0 to Objects.Count-1 do
      Objects[i].FirstPassProcessed:=false;

    //     aRec.X,Y,DX,DY
    for i:=0 to Objects.Count-1 do
      Objects[i].FirstPass;

    //   Band, ..  Cell
    Result.Height:=Height;
    case ResizeMode of
      prbrmMaxObj:
        begin
          //      Band
          if Objects.Count>0 then
            begin
              Result.Height:=Objects[0].aRec.Y+Objects[0].aRec.DY;
              for i:=1 to Objects.Count-1 do
                if (Objects[i].aRec.Y+Objects[i].aRec.DY)>Result.Height then
                  Result.Height:=Objects[i].aRec.Y+Objects[i].aRec.DY;
            end
        end;
      prbrmMaxResizeObj,prbrmMinResizeObj:
        begin
          if ResizeObjs.Count>0 then
            begin
              MinObj:=ResizeObjs[0];
              MaxObj:=ResizeObjs[0];
              for i:=1 to ResizeObjs.Count-1 do
                begin
                  if ResizeObjs[i].aRec.Y+ResizeObjs[i].aRec.DY>MaxObj.aRec.Y+MaxObj.aRec.DY then
                    MaxObj:=ResizeObjs[i];
                  if ResizeObjs[i].aRec.Y+ResizeObjs[i].aRec.DY<MaxObj.aRec.Y+MaxObj.aRec.DY then
                    MinObj:=ResizeObjs[i];
                end;

              if ResizeMode=prbrmMaxResizeObj then
                Result.Height:=Height+(MaxObj.aRec.Y+MaxObj.aRec.DY-MaxObj.dRec.pRect.Bottom)
              else
                Result.Height:=Height+(MinObj.aRec.Y+MinObj.aRec.DY-MinObj.dRec.pRect.Bottom);
            end;
        end;
    end;

    //    Cell
    for i:=0 to Objects.Count-1 do
      begin
        o             :=Objects[i].aRec.CreateCopy;
        o.Page        :=nil;
        o.Obj         :=Objects[i];
        o.pRect.Left  :=o.X;
        o.pRect.Right :=o.X+o.DX;
        o.pRect.Top   :=o.Y;
        o.pRect.Bottom:=o.Y+o.DY;
        Result.Add(o);
      end;
  end;
Page.Grid.EndOfLine;
end;

procedure TprCustomHBand.GenerateCellCallback;
var
  i : integer;
  Obj,MinObj,MaxObj : TprObj;
begin
//   Band, ..  Cell
Cell.Height:=Height;
case ResizeMode of
  prbrmMaxObj:
    begin
      //      Band
      if Objects.Count>0 then
        begin
          Cell.Height:=Objects[0].aRec.Y+Objects[0].aRec.DY;
          for i:=1 to Objects.Count-1 do
            if (Objects[i].aRec.Y+Objects[i].aRec.DY)>Cell.Height then
              Cell.Height:=Objects[i].aRec.Y+Objects[i].aRec.DY;
        end
    end;
  prbrmMaxResizeObj,prbrmMinResizeObj:
    begin
      MinObj:=nil;
      MaxObj:=nil;
      for i:=0 to ResizeObjs.Count-1 do
        begin
          Obj:=ResizeObjs[i];
          if Objects.IndexOf(Obj)<>-1 then
            begin
              if (MaxObj=nil) or (Obj.aRec.Y+Obj.aRec.DY>MaxObj.aRec.Y+MaxObj.aRec.DY) then
                MaxObj:=Obj;
              if (MinObj=nil) or (Obj.aRec.Y+Obj.aRec.DY<MaxObj.aRec.Y+MaxObj.aRec.DY) then
                MinObj:=Obj;
            end;
        end;
      if ResizeMode=prbrmMaxResizeObj then
        begin
          if (MaxObj<>nil) then
            Cell.Height:=Height+(MaxObj.aRec.Y+MaxObj.aRec.DY-MaxObj.dRec.pRect.Bottom)
        end
      else
        begin
          if (MinObj<>nil) then
            Cell.Height:=Height+(MinObj.aRec.Y+MinObj.aRec.DY-MinObj.dRec.pRect.Bottom);
        end;
    end;
end;
end;

function TprCustomHBand.GetUseColumns;
begin
Result:=false;
end;

function TprCustomHBand.GetColCount;
begin
Result:=0;
end;

function TprCustomHBand.GetColDirection;
begin
Result:=prcdTopBottomLeftRight;
end;















//////////////////////////
//
// TprCustomHTitleBand
//
//////////////////////////
function TprCustomHTitleBand.PrintOnStartPage;
begin
Result:=true;
end;

/////////////////////////
//
// TprCustomHSummaryBand
//
/////////////////////////
constructor TprCustomHSummaryBand.Create;
begin
inherited;
end;

function TprCustomHSummaryBand.GetLinkToBand;
begin
Result:=PrintWithBand;
end;

function TprCustomHSummaryBand.GetLinkType;
begin
Result:=prltBefore;
end;

//////////////////////////
//
// TprCustomHPageHeaderBand
//
//////////////////////////

/////////////////////////////
//
// TprCustomHPageFooterBand
//
/////////////////////////////
procedure TprCustomHPageFooterBand.CalcdPageRect;
begin
if FCalced then exit;
dPageRect         :=Rect(CurPageRect.Left,CurPageRect.Bottom-Height,CurPageRect.Right,CurPageRect.Bottom);
CurPageRect.Bottom:=CurPageRect.Bottom-Height;
FCalced           :=true;
end;

//////////////////////////
//
// TprCustomHDetailBand
//
//////////////////////////
constructor TprCustomHDetailBand.Create;
begin
inherited;
Dataset     :=TprDatasetLink.Create;
Bands       :=TprBands.Create;
Groups      :=TprGroups.Create;
end;

destructor TprCustomHDetailBand.Destroy;
begin
Bands.Free;
Bands:=nil;
Groups.Free;
Groups:=nil;
Dataset.Free;
inherited;
end;

procedure TprCustomHDetailBand.Notification;
begin
if (Operation=opRemove) and (AComponent<>Self) then
  begin
    if AComponent is TprGroup then
      Groups.Remove(AComponent);
    if AComponent is TprBand then
      Bands.Remove(AComponent);
    if AComponent=FParentDetail then
      FParentDetail:=nil;
  end;

inherited;
end;

procedure TprCustomHDetailBand.DefineProperties;
begin
inherited;
Filer.DefineProperty('Groups',ReadGroups,WriteGroups,Groups.Count>0);
Filer.DefineProperty('Bands',ReadBands,WriteBands,Bands.Count>0);
end;

procedure TprCustomHDetailBand.ReadGroups;
begin
FGroupsNames:=TStringList.Create;
prReadStringList(Reader,FGroupsNames);
end;

procedure TprCustomHDetailBand.WriteGroups;
begin
prWriteCompListNames(Writer,Groups);
end;

procedure TprCustomHDetailBand.ReadBands;
begin
FBandsNames:=TStringList.Create;
prReadStringList(Reader,FBandsNames);
end;

procedure TprCustomHDetailBand.WriteBands;
begin
prWriteCompListNames(Writer,Bands);
end;

procedure TprCustomHDetailBand.GroupRemoved;
begin
Groups.Remove(Group);
end;

procedure TprCustomHDetailBand.GroupAdded;
var
  i : integer;
begin
i:=Groups.IndexOf(Group);
if i=-1 then
  Groups.Add(Group);
end;

procedure TprCustomHDetailBand.AfterReportLoaded;
var
  i : integer;
begin
inherited;
if FGroupsNames<>nil then
  begin
    Groups.Clear;
    for i:=0 to FGroupsNames.Count-1 do
      Groups.Add(Report.Groups.ByName[FGroupsNames[i]]);
    FGroupsNames.Free;
  end;
if FBandsNames<>nil then
  begin
    Bands.Clear;
    for i:=0 to FBandsNames.Count-1 do
      Bands.Add(Page.Bands.ByName[FBandsNames[i]]);
    FBandsNames.Free;
  end;
end;

procedure TprCustomHDetailBand.SetParentDetail;
var
  i : integer;
begin
if Value=FParentDetail then exit;
if (FParentDetail<>nil) and (FParentDetail.Bands<>nil) then
  begin
    FParentDetail.Bands.Remove(Self);
  end;
FParentDetail:=Value;
if (FParentDetail<>nil) and (FParentDetail.Bands<>nil) then
  begin
    i:=FParentDetail.Bands.IndexOf(Self);
    if i=-1 then
      FParentDetail.Bands.Add(Self);
  end;
end;

function TprCustomHDetailBand.GetUseColumns;
begin
Result:=ColCount>1;
end;

function TprCustomHDetailBand.GetColCount;
begin
Result:=ColCount;
end;

function TprCustomHDetailBand.GetColDirection;
begin
Result:=ColDirection;
end;

function TprCustomHDetailBand.GenerateCell;
var
  i : integer;
  fGroupEnded : boolean;
begin
Result:=nil;
//     
i:=Bands.IndexByBandType(bthDetailHeader);
if i<>-1 then
  (Bands[i] as TprCustomHBand).GenerateCell(nil,nil,false);

//  
for i:=0 to Groups.Count-1 do
  Groups[i].Reset;

//     
while not DataSet.Eof do
  begin
    // 
    for i:=0 to Groups.Count-1 do
      Groups[i].CalcValue;

    for i:=Groups.Count-1 downto 0 do
      begin
        fGroupEnded:=Groups[i].GroupEnded;
        if fGroupEnded then
          DataSet.Prior;
        Groups[i].FootersGenerateCell(nil,nil);
        if fGroupEnded then
          DataSet.Next;
      end;
    for i:=0 to Groups.Count-1 do
      Groups[i].HeadersGenerateCell(nil,nil);

    // 
    Report.DoOnDataSetNext(DataSet,Self,nil);

    //  DetailBand
    inherited GenerateCell(nil,nil,false);

    //    Band
    for i:=0 to Bands.Count-1 do
      if Bands[i].BandType in [bthDetail] then
        (Bands[i] as TprCustomHBand).GenerateCell(nil,nil,false);

    DataSet.Next;
    Report.ADSRN_Next(DataSet);
  end;

//  FooterBands
for i:=Groups.Count-1 downto 0 do
  begin
    DataSet.Prior;
    Groups[i].FootersAlwaysGenerateCell(nil,nil);
    DataSet.Next;
  end;

//    
i:=Bands.IndexByBandType(bthDetailFooter);
if i<>-1 then
  (Bands[i] as TprCustomHBand).GenerateCell(nil,nil,false);

//     DataSet
Report.DoOnDataSetEof(DataSet);
end;


procedure TprCustomHDetailBand.InitDataSet;
begin
Dataset.Dataset:=Report.GetDataSetByName(DataSetName);
if Dataset.Dataset=nil then
  raise Exception.CreateFmt(prLoadStr(sDataSetNotFound),[DataSetName]);
if not Dataset.Active then
  Dataset.Open;
DataSet.First;
Report.ADSRN_First(DataSet);
end;

procedure TprCustomHDetailBand.CalcdPageRect;
var
  i,j : integer;
begin
if FCalced then exit;

i:=Bands.IndexByBandType(bthDetailHeader);
if i<>-1 then
  Bands[i].CalcdPageRect(CurPageRect,nil);

for i:=0 to Groups.Count-1 do
  for j:=0 to Groups[i].Headers.Count-1 do
    Groups[i].Headers[j].CalcdPageRect(CurPageRect,nil);

inherited;

for i:=0 to Bands.Count-1 do
  if Bands[i].BandType in [bthDetail] then
    Bands[i].CalcdPageRect(CurPageRect,nil);

for i:=Groups.Count-1 downto 0 do
  for j:=0 to Groups[i].Footers.Count-1 do
    Groups[i].Footers[j].CalcdPageRect(CurPageRect,nil);

i:=Bands.IndexByBandType(bthDetailFooter);
if i<>-1 then
  Bands[i].CalcdPageRect(CurPageRect,nil);

FCalced:=true;
end;

procedure TprCustomHDetailBand.OnInsertIntoPage;
begin
end;

////////////////////////////////
//
// TprCustomHDetailHeaderBand
//
////////////////////////////////
procedure TprCustomHDetailHeaderBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FDetailBand) then
  FDetailBand:=nil;

inherited;
end;

procedure TprCustomHDetailHeaderBand.SetDetailBand;
var
  i : integer;
begin
if Value=FDetailBand then exit;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    FDetailBand.Bands.Remove(Self);
  end;
FDetailBand:=Value;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    i:=FDetailBand.Bands.IndexOf(Self);
    if i=-1 then
      FDetailBand.Bands.Add(Self);
  end;
end;

procedure TprCustomHDetailHeaderBand.OnInsertIntoPage;
begin
end;

function TprCustomHDetailHeaderBand.GetUseColumns;
begin
Result:=ColCount>1;
end;

function TprCustomHDetailHeaderBand.GetColCount;
begin
Result:=ColCount;
end;

function TprCustomHDetailHeaderBand.GetColDirection;
begin
Result:=ColDirection;
end;

function TprCustomHDetailHeaderBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (DetailBand<>nil) then
  Result:=DetailBand;
end;

function TprCustomHDetailHeaderBand.GetLinkType;
begin
Result:=prltAfter;
end;

////////////////////////////////
//
// TprCustomHDetailFooterBand
//
////////////////////////////////
procedure TprCustomHDetailFooterBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FDetailBand) then
  FDetailBand:=nil;

inherited;
end;

procedure TprCustomHDetailFooterBand.SetDetailBand;
var
  i : integer;
begin
if Value=FDetailBand then exit;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    FDetailBand.Bands.Remove(Self);
  end;
FDetailBand:=Value;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    i:=FDetailBand.Bands.IndexOf(Self);
    if i=-1 then
      FDetailBand.Bands.Add(Self);
  end;
end;

procedure TprCustomHDetailFooterBand.OnInsertIntoPage;
begin
end;

function TprCustomHDetailFooterBand.GetUseColumns;
begin
Result:=ColCount>1;
end;

function TprCustomHDetailFooterBand.GetColCount;
begin
Result:=ColCount;
end;

function TprCustomHDetailFooterBand.GetColDirection;
begin
Result:=ColDirection;
end;

function TprCustomHDetailFooterBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (DetailBand<>nil) then
  Result:=DetailBand;
end;

function TprCustomHDetailFooterBand.GetLinkType;
begin
Result:=prltBefore;
end;

////////////////////////////////////
//
// TprCustomHGroupHeaderBand
//
////////////////////////////////////
procedure TprCustomHGroupHeaderBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FGroup) then
  FGroup:=nil;

inherited;
end;

procedure TprCustomHGroupHeaderBand.OnInsertIntoPage;
begin
end;

procedure TprCustomHGroupHeaderBand.SetGroup;
var
  i : integer;
begin
if Value=FGroup then exit;
if FGroup<>nil then
  begin
    FGroup.Headers.Remove(Self);
  end;
FGroup:=Value;
if FGroup<>nil then
  begin
    i:=FGroup.Headers.IndexOf(Self);
    if i=-1 then
      FGroup.Headers.Add(Self);
  end;
end;

function TprCustomHGroupHeaderBand.GetDrawDesignCaption;
begin
if (Group=nil) or (Group.DetailBand=nil) then
  Result:=inherited GetDrawDesignCaption
else
  Result:=Format('%s (%s)',[inherited GetDrawDesignCaption,Group.DetailBand.Name]);
end;

function TprCustomHGroupHeaderBand.GetUseColumns;
begin
Result:=ColCount>1;
end;

function TprCustomHGroupHeaderBand.GetColCount;
begin
Result:=ColCount;
end;

function TprCustomHGroupHeaderBand.GetColDirection;
begin
Result:=ColDirection;
end;

function TprCustomHGroupHeaderBand.GetStartNewPage;
begin
Result:=StartNewPage;
end;

function TprCustomHGroupHeaderBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (Group<>nil) then
  Result:=Group.DetailBand;
end;

function TprCustomHGroupHeaderBand.GetLinkType;
begin
Result:=prltAfter;
end;

////////////////////////////////////
//
// TprCustomHGroupFooterBand
//
////////////////////////////////////
procedure TprCustomHGroupFooterBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FGroup) then
  FGroup:=nil;

inherited;
end;

procedure TprCustomHGroupFooterBand.OnInsertIntoPage;
begin
end;

procedure TprCustomHGroupFooterBand.SetGroup;
var
  i : integer;
begin
if Value=FGroup then exit;
if FGroup<>nil then
  begin
    FGroup.Footers.Remove(Self);
  end;
FGroup:=Value;
if FGroup<>nil then
  begin
    i:=FGroup.Footers.IndexOf(Self);
    if i=-1 then
      FGroup.Footers.Add(Self);
  end;
end;

function TprCustomHGroupFooterBand.GetDrawDesignCaption;
begin
if (Group=nil) or (Group.DetailBand=nil) then
  Result:=inherited GetDrawDesignCaption
else
  Result:=Format('%s (%s)',[inherited GetDrawDesignCaption,Group.DetailBand.Name]);
end;

function TprCustomHGroupFooterBand.GetUseColumns;
begin
Result:=ColCount>1;
end;

function TprCustomHGroupFooterBand.GetColCount;
begin
Result:=ColCount;
end;

function TprCustomHGroupFooterBand.GetColDirection;
begin
Result:=ColDirection;
end;

function TprCustomHGroupFooterBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (Group<>nil) then
  Result:=Group.DetailBand;
end;

function TprCustomHGroupFooterBand.GetLinkType;
begin
Result:=prltBefore;
end;



















/////////////////////////////
//
// TprCustomVBand
//
/////////////////////////////
constructor TprCustomVBand.Create;
begin
inherited;
end;

function TprCustomVBand.PrintOnStartPage;
begin
Result:=false;
end;

//
//  Band  
//
procedure TprCustomVBand.CalcdPageRect;
begin
if FCalced then exit;
dPageRect       :=Rect(CurPageRect.Left,CurPageRect.Top,CurPageRect.Left+Width,CurPageRect.Bottom);
CurPageRect.Left:=CurPageRect.Left+Width;
FCalced         :=true;
end;

procedure TprCustomVBand.DesignerResize;
begin
Width:=dPageRect.Right+oRight-dPageRect.Left-oLeft;
Report.ConvertToDesignerCoords(dPageRect,PDI.dPageRect);
end;

procedure TprCustomVBand.DesignerLink;
begin
if (Linked.Parent is TprObj) and
   RectInRect(Linked.dPageRect,PDI.dPageRect) then
  begin
    case LinkMode of
      ltRight : if ResizeObjs.IndexOf(Linked.Parent)=-1 then
                  ResizeObjs.Add(Linked.Parent);
    end;
    LinkAccepted:=true;
  end;
end;

function TprCustomVBand.GenerateCell;
var
  i : integer;
  o : TprObjRec;
  HBand : TprCustomHBand;
  MinObj,MaxObj : TprObj;
begin
Result:=nil;
if (CallerBand=nil) and FCrossTabGenerated then exit;

Report.DoOnBandGenerateCell(Self);

if not Visible then exit;

if (CallerBand=nil) or not UseHorizontalBands then
  HBand:=nil
else
  HBand:=CallerBand as TprCustomHBand;

if fSimpleCreateCell then
  Result:=TprGenCell.Create
else
  Result:=Page.Grid.AddCell(HBand,Self);
  
try
  //   Objects   ,    
  //   Band    Band
  if (HBand<>nil) and (UseHorizontalBands) then
    for i:=0 to HBand.Objects.Count-1 do
      if (HBand.Objects[i].dRec.pRect.Left+HBand.dPageRect.Left>=dPageRect.Left) and
         (HBand.Objects[i].dRec.pRect.Right+HBand.dPageRect.Left<=dPageRect.Right) then
        //     Band
        Objects.Add(HBand.Objects[i]);

  for i:=0 to Objects.Count-1 do
    Objects[i].FirstPassProcessed:=false;
  
  //     aRec.X,Y,DX,DY
  for i:=0 to Objects.Count-1 do
    Objects[i].FirstPass;

  //   Band   
  Result.Width:=Width;
  case ResizeMode of
    prbrmMaxObj:
      begin
        if Objects.Count>0 then
          begin
            //    
            Result.Width:=Objects[0].aRec.X+Objects[0].aRec.DX;
            for i:=1 to Objects.Count-1 do
              if (Objects[i].aRec.X+Objects[i].aRec.DX)>Result.Width then
                Result.Width:=Objects[i].aRec.X+Objects[i].aRec.DX;

            if HBand<>nil then
              // ,    -
              //   , ,     X
              //  -    
              Result.Width:=Result.Width-dPageRect.Left+Objects[0].Band.dPageRect.Left;
          end;
      end;
    prbrmMaxResizeObj,prbrmMinResizeObj:
      begin
        if ResizeObjs.Count>0 then
          begin
            MinObj:=ResizeObjs[0];
            MaxObj:=ResizeObjs[0];
            for i:=1 to ResizeObjs.Count-1 do
              begin
                if ResizeObjs[i].aRec.X+ResizeObjs[i].aRec.DX>MaxObj.aRec.X+MaxObj.aRec.DX then
                  MaxObj:=ResizeObjs[i];
                if ResizeObjs[i].aRec.X+ResizeObjs[i].aRec.DX<MaxObj.aRec.X+MaxObj.aRec.DX then
                  MinObj:=ResizeObjs[i];
              end;
  
            if ResizeMode=prbrmMaxResizeObj then
              Result.Width:=Width+(MaxObj.aRec.DX-
                                   (MaxObj.dRec.pRect.Right-
                                    MaxObj.dRec.pRect.Left))
            else
              Result.Width:=Width+(MinObj.aRec.DX-
                                   (MinObj.dRec.pRect.Right-
                                    MinObj.dRec.pRect.Left))
          end;
      end;
  end;

  if Assigned(CallBackProc) then
    CallBackProc(Self,Result,Objects);

  for i:=0 to Objects.Count-1 do
    begin
      o             :=Objects[i].aRec.CreateCopy;
      o.Page        :=nil;
      o.Obj         :=Objects[i];
      o.pRect.Top   :=o.Y;
      o.pRect.Bottom:=o.Y+o.DY;
      o.pRect.Left  :=o.X;
      o.pRect.Right :=o.X+o.DX;
      Result.Add(o);
    end;

  FCrossTabGenerated:=true;
finally
  if (HBand<>nil) and (UseHorizontalBands) then
    Objects.Clear
end;
end;











//////////////////////////
//
// TprCustomVTitleBand
//
//////////////////////////
function TprCustomVTitleBand.PrintOnStartPage;
begin
Result:=true;
end;
/////////////////////////
//
// TprCustomVSummaryBand
//
/////////////////////////
constructor TprCustomVSummaryBand.Create;
begin
inherited;
end;

function TprCustomVSummaryBand.GetLinkToBand;
begin
Result:=PrintWithBand;
end;

function TprCustomVSummaryBand.GetLinkType;
begin
Result:=prltBefore;
end;

//////////////////////////
//
// TprCustomVPageHeaderBand
//
//////////////////////////

/////////////////////////////
//
// TprCustomVPageFooterBand
//
/////////////////////////////
procedure TprCustomVPageFooterBand.CalcdPageRect;
begin
if FCalced then exit;
dPageRect         :=Rect(CurPageRect.Right-Width,CurPageRect.Top,CurPageRect.Right,CurPageRect.Bottom);
CurPageRect.Right :=CurPageRect.Right-Width;
FCalced           :=true;
end;

//////////////////////////
//
// TprCustomVDetailBand
//
//////////////////////////
constructor TprCustomVDetailBand.Create;
begin
inherited;
Dataset:=TprDatasetLink.Create;
Bands  :=TprBands.Create;
Groups :=TprGroups.Create;
end;

destructor TprCustomVDetailBand.Destroy;
begin
Bands.Free;
Bands:=nil;
Groups.Free;
Groups:=nil;
Dataset.Free;
inherited;
end;

procedure TprCustomVDetailBand.Notification;
begin
if (Operation=opRemove) and (AComponent<>Self) then
  begin
    if AComponent is TprGroup then
      Groups.Remove(AComponent);
    if AComponent is TprBand then
      Bands.Remove(AComponent);
    if AComponent=FParentDetail then
      FParentDetail:=nil;
  end;

inherited;
end;

procedure TprCustomVDetailBand.DefineProperties;
begin
inherited;
Filer.DefineProperty('Groups',ReadGroups,WriteGroups,Groups.Count>0);
Filer.DefineProperty('Bands',ReadBands,WriteBands,Bands.Count>0);
end;

procedure TprCustomVDetailBand.ReadGroups;
begin
FGroupsNames:=TStringList.Create;
prReadStringList(Reader,FGroupsNames);
end;

procedure TprCustomVDetailBand.WriteGroups;
begin
prWriteCompListNames(Writer,Groups);
end;

procedure TprCustomVDetailBand.ReadBands;
begin
FBandsNames:=TStringList.Create;
prReadStringList(Reader,FBandsNames);
end;

procedure TprCustomVDetailBand.WriteBands;
begin
prWriteCompListNames(Writer,Bands);
end;

procedure TprCustomVDetailBand.GroupRemoved;
begin
Groups.Remove(Group);
end;

procedure TprCustomVDetailBand.GroupAdded;
var
  i : integer;
begin
i:=Groups.IndexOf(Group);
if i=-1 then
  Groups.Add(Group);
end;

procedure TprCustomVDetailBand.AfterReportLoaded;
var
  i : integer;
begin
inherited;
if FGroupsNames<>nil then
  begin
    Groups.Clear;
    for i:=0 to FGroupsNames.Count-1 do
      Groups.Add(Report.Groups.ByName[FGroupsNames[i]]);
    FGroupsNames.Free;
  end;
if FBandsNames<>nil then
  begin
    Bands.Clear;
    for i:=0 to FBandsNames.Count-1 do
      Bands.Add(Page.Bands.ByName[FBandsNames[i]]);
    FBandsNames.Free;
  end;
end;

procedure TprCustomVDetailBand.SetParentDetail;
var
  i : integer;
begin
if Value=FParentDetail then exit;
if (FParentDetail<>nil) and (FParentDetail.Bands<>nil) then
  begin
    FParentDetail.Bands.Remove(Self);
  end;
FParentDetail:=Value;
if (FParentDetail<>nil) and (FParentDetail.Bands<>nil) then
  begin
    i:=FParentDetail.Bands.IndexOf(Self);
    if i=-1 then
      FParentDetail.Bands.Add(Self);
  end;
end;

procedure TprCustomVDetailBand.InitDataSet;
begin
Dataset.Dataset:=Report.GetDataSetByName(DataSetName);
if Dataset.Dataset=nil then
  raise Exception.CreateFmt(prLoadStr(sDataSetNotFound),[DataSetName]);
if not Dataset.Active then
  Dataset.Open;
DataSet.First;
Report.ADSRN_First(DataSet);
end;

function TprCustomVDetailBand.GenerateCell;
var
  i : integer;
  fGroupEnded : boolean;
begin
Result:=nil;

// 
i:=Bands.IndexByBandType(btvDetailHeader);
if i<>-1 then
  Bands[i].GenerateCell(CallerBand,CallbackProc,false);

//  
for i:=0 to Groups.Count-1 do
  Groups[i].Reset;

while not DataSet.Eof do
  begin
    // 
    for i:=0 to Groups.Count-1 do
      Groups[i].CalcValue;

    for i:=Groups.Count-1 downto 0 do
      begin
        fGroupEnded:=Groups[i].GroupEnded;
        if fGroupEnded then
          DataSet.Prior;
        Groups[i].FootersGenerateCell(CallerBand,CallbackProc);
        if fGroupEnded then
          DataSet.Next;
      end;
    for i:=0 to Groups.Count-1 do
      Groups[i].HeadersGenerateCell(CallerBand,CallbackProc);

    // 
    if (CallerBand<>nil) and (CallerBand is TprCustomHBand) then
      Report.DoOnDataSetNext(DataSet,TprCustomHBand(CallerBand),Self)
    else
      Report.DoOnDataSetNext(DataSet,nil,Self);

    inherited GenerateCell(CallerBand,CallbackProc,false);

    //    Band
    for i:=0 to Bands.Count-1 do
      if Bands[i].BandType in [bthDetail] then
        Bands[i].GenerateCell(CallerBand,CallbackProc,false);

    DataSet.Next;
    Report.ADSRN_Next(DataSet);
  end;

//  FooterBands
for i:=Groups.Count-1 downto 0 do
  begin
    DataSet.Prior;
    Groups[i].FootersAlwaysGenerateCell(CallerBand,CallbackProc);
    DataSet.Next;
  end;

//    
i:=Bands.IndexByBandType(btvDetailFooter);
if i<>-1 then
  Bands[i].GenerateCell(CallerBand,CallbackProc,false);

//     DataSet
Report.DoOnDataSetEof(DataSet);
end;


procedure TprCustomVDetailBand.CalcdPageRect;
var
  i,j : integer;
begin
if FCalced then exit;

i:=Bands.IndexByBandType(btvDetailHeader);
if i<>-1 then
  Bands[i].CalcdPageRect(CurPageRect,nil);

for i:=0 to Groups.Count-1 do
  for j:=0 to Groups[i].Headers.Count-1 do
    Groups[i].Headers[j].CalcdPageRect(CurPageRect,nil);

inherited;

for i:=0 to Bands.Count-1 do
  if Bands[i].BandType in [btvDetail] then
    Bands[i].CalcdPageRect(CurPageRect,nil);

for i:=0 to Groups.Count-1 do
  for j:=0 to Groups[i].Footers.Count-1 do
    Groups[i].Footers[j].CalcdPageRect(CurPageRect,nil);

i:=Bands.IndexByBandType(btvDetailFooter);
if i<>-1 then
  Bands[i].CalcdPageRect(CurPageRect,nil);

FCalced:=true;
end;

procedure TprCustomVDetailBand.OnInsertIntoPage;
begin
end;

////////////////////////////////
//
// TprCustomVDetailHeaderBand
//
////////////////////////////////
procedure TprCustomVDetailHeaderBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FDetailBand) then
  FDetailBand:=nil;

inherited;
end;

procedure TprCustomVDetailHeaderBand.SetDetailBand;
var
  i : integer;
begin
if Value=FDetailBand then exit;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    FDetailBand.Bands.Remove(Self);
  end;
FDetailBand:=Value;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    i:=FDetailBand.Bands.IndexOf(Self);
    if i=-1 then
      FDetailBand.Bands.Add(Self);
  end;
end;

procedure TprCustomVDetailHeaderBand.OnInsertIntoPage;
begin
end;

function TprCustomVDetailHeaderBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (DetailBand<>nil) then
  Result:=DetailBand;
end;

function TprCustomVDetailHeaderBand.GetLinkType;
begin
Result:=prltAfter;
end;

////////////////////////////////
//
// TprCustomVDetailFooterBand
//
////////////////////////////////
procedure TprCustomVDetailFooterBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FDetailBand) then
  FDetailBand:=nil;

inherited;
end;

procedure TprCustomVDetailFooterBand.SetDetailBand;
var
  i : integer;
begin
if Value=FDetailBand then exit;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    FDetailBand.Bands.Remove(Self);
  end;
FDetailBand:=Value;
if (FDetailBand<>nil) and (FDetailBand.Bands<>nil) then
  begin
    i:=FDetailBand.Bands.IndexOf(Self);
    if i=-1 then
      FDetailBand.Bands.Add(Self);
  end;
end;

procedure TprCustomVDetailFooterBand.OnInsertIntoPage;
begin
end;

function TprCustomVDetailFooterBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (DetailBand<>nil) then
  Result:=DetailBand;
end;

function TprCustomVDetailFooterBand.GetLinkType;
begin
Result:=prltBefore;
end;

////////////////////////////////////
//
// TprCustomVGroupHeaderBand
//
////////////////////////////////////
procedure TprCustomVGroupHeaderBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FGroup) then
  FGroup:=nil;

inherited;
end;

procedure TprCustomVGroupHeaderBand.OnInsertIntoPage;
begin
end;

function TprCustomVGroupHeaderBand.GetStartNewPage;
begin
Result:=StartNewPage;
end;

function TprCustomVGroupHeaderBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (Group<>nil) then
  Result:=Group.DetailBand;
end;

function TprCustomVGroupHeaderBand.GetLinkType;
begin
Result:=prltAfter;
end;

procedure TprCustomVGroupHeaderBand.SetGroup;
var
  i : integer;
begin
if Value=FGroup then exit;
if FGroup<>nil then
  begin
    FGroup.Headers.Remove(Self);
  end;
FGroup:=Value;
if FGroup<>nil then
  begin
    i:=FGroup.Headers.IndexOf(Self);
    if i=-1 then
      FGroup.Headers.Add(Self);
  end;
end;

function TprCustomVGroupHeaderBand.GetDrawDesignCaption;
begin
if (Group=nil) or (Group.DetailBand=nil) then
  Result:=inherited GetDrawDesignCaption
else
  Result:=Format('%s (%s)',[inherited GetDrawDesignCaption,Group.DetailBand.Name]);
end;

////////////////////////////////////
//
// TprCustomVGroupFooterBand
//
////////////////////////////////////
procedure TprCustomVGroupFooterBand.Notification;
begin
if (Operation=opRemove) and (AComponent=FGroup) then
  FGroup:=nil;

inherited;
end;

procedure TprCustomVGroupFooterBand.OnInsertIntoPage;
begin
end;

procedure TprCustomVGroupFooterBand.SetGroup;
var
  i : integer;
begin
if Value=FGroup then exit;
if FGroup<>nil then
  begin
    FGroup.Footers.Remove(Self);
  end;
FGroup:=Value;
if FGroup<>nil then
  begin
    i:=FGroup.Footers.IndexOf(Self);
    if i=-1 then
      FGroup.Footers.Add(Self);
  end;
end;

function TprCustomVGroupFooterBand.GetLinkToBand;
begin
Result:=nil;
if LinkToDetail and (Group<>nil) then
  Result:=Group.DetailBand;
end;

function TprCustomVGroupFooterBand.GetLinkType;
begin
Result:=prltBefore;
end;

function TprCustomVGroupFooterBand.GetDrawDesignCaption;
begin
if (Group=nil) or (Group.DetailBand=nil) then
  Result:=inherited GetDrawDesignCaption
else
  Result:=Format('%s (%s)',[inherited GetDrawDesignCaption,Group.DetailBand.Name]);
end;



















//////////////////////////
//
// TprCustomEndPage
//
//////////////////////////
constructor TprCustomEndPage.CreateEmpty;
begin
inherited Create;
Report:=_Report;
FoRecs:=TList.Create;
end;

constructor TprCustomEndPage.Create;
begin
CreateEmpty(_Page.Report);
end;

destructor TprCustomEndPage.Destroy;
var
  i : integer;
begin
for i:=0 to FoRecs.Count-1 do
  oRec[i].Free;
FoRecs.Free;
inherited;
end;

function TprCustomEndPage.oRecsCount;
begin
Result:=FoRecs.Count;
end;

function TprCustomEndPage.GetoRec;
begin
Result:=TprObjRec(FoRecs[index]);
end;

procedure TprCustomEndPage.FreeGeneratedObjects;
var
  i : integer;
begin
for i:=0 to oRecsCount-1 do
  oRec[i].Free;
FoRecs.Clear;
end;

function TprCustomEndPage.CurrentWidth;
begin
Result:=Rect.Right-Rect.Left;
end;

function TprCustomEndPage.CurrentHeight;
begin
Result:=Rect.Bottom-Rect.Top;
end;

//////////////////////////
//
// TprCustomPage
//
//////////////////////////
constructor TprCustomPage.Create;
begin
{$IFDEF DEBUG}
if AOwner=nil then
  WriteToLog('TprCustomPage.Create : AOwner : nil')
else
  WriteToLog('TprCustomPage.Create : AOwner : '+AOwner.Name);
{$ENDIF}
inherited;
dPageRect:=nil;
Bands    :=TprBands.Create;
Grid     :=TprGenGrid.Create(Self);

Initialize(FOldEndPagesRects);

Name:=GetValidComponentName(Self);
end;

destructor TprCustomPage.Destroy;
begin
{$IFDEF DEBUG}
WriteToLog('TprCustomPage.Destroy : '+Name);
{$ENDIF}
Grid.Free;

while Bands.Count>0 do
  Bands[0].Free;

Bands.Free;
Finalize(FOldEndPagesRects);

{$IFDEF DEBUG}
WriteToLog('TprCustomPage.Destroy : '+Name+' destroyed');
{$ENDIF}

inherited;
end;

procedure TprCustomPage.SetParentComponent;
begin
Report:=Value as TprCustomReport;
{$IFDEF DEBUG}
if Report=nil then
  WriteToLog('TprCustomPage.SetParentComponent : nil')
else
  WriteToLog('TprCustomPage.SetParentComponent : '+Report.Name);
{$ENDIF}
end;

function TprCustomPage.HasParent;
begin
Result:=true;
end;

function TprCustomPage.GetParentComponent;
begin
Result:=Report;
end;

function TprCustomPage.GetChildOwner;
begin
Result:=Report;
end;

procedure TprCustomPage.GetChildren;
var
  i : integer;
begin
for i:=0 to Bands.Count-1 do
  Proc(Bands[i]);
end;

procedure TprCustomPage.Notification;
begin
{$IFDEF DEBUG}
if Owner=nil then
  WriteToLog('TprCustomPage.Notification : '+Name+' '+GetEnumName(TypeInfo(TOperation),integer(Operation))+' AComponent : '+AComponent.Name+' Owner : nil')
else
  WriteToLog('TprCustomPage.Notification : '+Name+' '+GetEnumName(TypeInfo(TOperation),integer(Operation))+' AComponent : '+AComponent.Name+' Owner : '+Owner.Name);
{$ENDIF}
if (Operation=opRemove) and (AComponent<>Self) then
  begin
    if AComponent is TprBand then
      Bands.Remove(AComponent);
  end;
end;

procedure TprCustomPage.SetReport;
begin
if FReport=Value then exit;
if FReport<>nil then
  FReport.FPages.Remove(Self);

FReport:=Value;

if (FReport<>nil) and (FReport.FPages.IndexOf(Self)=-1) then
  begin
    FReport.FPages.Add(Self);
    ReportSetted;
  end;
{$IFDEF DEBUG}
if Report=nil then
  WriteToLog('TprCustomPage.SetReport : nil')
else
  WriteToLog('TprCustomPage.SetReport : '+Report.Name);
{$ENDIF}
end;

procedure TprCustomPage.ReportSetted;
begin
end;

function TprCustomPage.IndexInReport;
begin
Result:=Report.FPages.IndexOf(Self);
end;

function TprCustomPage.HasVerticalBands;
var
  i : integer;
begin
Result:=false;
for i:=0 to Bands.Count-1 do
  if Bands[i].BandType in VerticalBands then
    begin
      Result:=true;
      exit;
    end;
end;

procedure TprCustomPage.FirstPassGenerateGrid;
var
  i : integer;
begin
//  Grid,   - 
Grid.Clear;

//
for i:=0 to Bands.Count-1 do
  if Bands[i].BandType in VerticalBands then
    TprCustomVBand(Bands[i]).FCrossTabGenerated:=false;

//   
for i:=0 to Bands.Count-1 do
  if Bands[i].BandType in HorizontalBands then
    Bands[i].InitDataSet;

//   
i:=Bands.IndexByBandType(btvTitle);
if i<>-1 then
  Grid.vTitleCell:=Bands[i].GenerateCell(nil,nil,true);

//   
i:=Bands.IndexByBandType(bthTitle);
if i<>-1 then
  Grid.hTitleCell:=Bands[i].GenerateCell(nil,nil,true);

//         , 
//        ,
//       
for i:=0 to Bands.Count-1 do
  if (Bands[i].BandType=bthDetail) and
     (TprCustomHDetailBand(Bands[i]).ParentDetail=nil) then
    Bands[i].GenerateCell(nil,nil,false);

//   
i:=Bands.IndexByBandType(bthSummary);
if i<>-1 then
  Bands[i].GenerateCell(nil,nil,false);
end;

type

TLinkInfo = class(TObject)
  Vector : TprGenVector;
  Band   : TprBand;

  constructor Create(_Vector : TprGenVector; _Band : TprBand);
end;

//////////////////////////////////
//
// TLinkInfo
//
//////////////////////////////////
constructor TLinkInfo.Create;
begin
inherited Create;
Vector:=_Vector;
Band  :=_Band;
end;

type

TLinksList = class(TList)
private
  function GetItm(i : integer) : TLinkInfo;
public
  property Items[i : integer] : TLinkInfo read GetItm; default;

  procedure FreeItem(i : integer);
  procedure FreeAll;

  destructor Destroy; override;
end;

///////////////////////
//
// TLinksList
//
///////////////////////
destructor TLinksList.Destroy;
begin
FreeAll;
inherited;
end;

function TLinksList.GetItm;
begin
Result:=TLinkInfo(inherited Items[i]);
end;

procedure TLinksList.FreeItem;
begin
Items[i].Free;
Delete(i);
end;

procedure TLinksList.FreeAll;
begin
while Count>0 do
  FreeItem(0);
end;

//
//   
//
procedure TprCustomPage.SecondPassPlaceGridOnEndPage;
var
  Cell : TprGenCell;
  Band : TprBand;
  fPlace : boolean;
  LinksList : TLinksList;
  PriorEndPage : TprCustomEndPage;
  MaxCell,MinCell : TprGenCell;
  PriorColDirection : TprColDirectionType;
  V,Vector,MinVector : TprGenVector;
  i,CurVectorIndex,j,PageX,PageY,StartEndPageIndex,MaxTop,StartPageTop,ColWidth : integer;

  procedure DeleteBandFromLinksList(Vector : TprGenVector; Band : TprBand);
  var
    i : integer;
  begin
  i:=0;
  while i<LinksList.Count do
    if LinksList[i].Band=Band then
      begin
        //      Vector
        //     - LinksList[j].Vector
        LinksList[i].Vector.FLinkToVector:=Vector;
        LinksList.FreeItem(i);
      end
    else
      Inc(i);
  end;

  function FindEndPage(PageX,PageY : integer) : TprCustomEndPage;
  var
    i : integer;
  begin
  i:=Report.EndPagesCount-1;
  while (i>=StartEndPageIndex) and ((Report.EndPages[i].PageX<>PageX) or (Report.EndPages[i].PageY<>PageY)) do Dec(i);
  if i>=StartEndPageIndex then
    Result:=Report.EndPages[i]
  else
    Result:=nil;
  end;

  procedure InsertVerticalDetailHeader(Band : TprBand);
  var
    j : integer;
    VDetailBand : TprCustomVDetailBand;
  begin
  if Band.BandType=btvGroupHeader then
    VDetailBand:=TprCustomVDetailBand(TprCustomVGroupHeaderBand(Band).Group.DetailBand)
  else
    if Band.BandType=btvGroupFooter then
      VDetailBand:=TprCustomVDetailBand(TprCustomVGroupFooterBand(Band).Group.DetailBand)
    else
      if Band.BandType=btvDetail then
        VDetailBand:=TprCustomVDetailBand(Band)
      else
        if Band.BandType=btvDetailFooter then
          VDetailBand:=TprCustomVDetailBand(TprCustomVDetailFooterBand(Band).DetailBand)
        else
          exit;

  //         ,  
  //   , (    )
  //  
  for j:=0 to VDetailBand.Bands.Count-1 do
    if (VDetailBand.Bands[j].BandType=btvDetailHeader) and TprCustomVDetailHeaderBand(VDetailBand.Bands[j]).ReprintOnEachPage then
      begin
        //        
        V:=Grid.FindPriorVVector(CurVectorIndex,VDetailBand.Bands[j],StartEndPageIndex);
        if V<>nil then
          begin
            V      :=Grid.CopyVVector(V,CurVectorIndex);
            V.Page :=PageX;
            V.YOffs:=-1;
            V.XOffs:=Report.CurEndPage.Rect.Left;
            Report.CurEndPage.Rect.Left:=Report.CurEndPage.Rect.Left+V.Width;
          end;
      end;
  end;

  procedure InsertHorizontalDetailHeader(Band : TprBand);
  var
    j : integer;
    HDetailBand : TprCustomHDetailBand;
  begin
  if Band.BandType=bthGroupHeader then
    HDetailBand:=TprCustomHDetailBand(TprCustomHGroupHeaderBand(Band).Group.DetailBand)
  else
    if Band.BandType=bthGroupFooter then
      HDetailBand:=TprCustomHDetailBand(TprCustomHGroupFooterBand(Band).Group.DetailBand)
    else
      if Band.BandType=bthDetail then
        HDetailBand:=TprCustomHDetailBand(Band)
      else
        if Band.BandType=bthDetailFooter then
          HDetailBand:=TprCustomHDetailBand(TprCustomHDetailFooterBand(Band).DetailBand)
        else
          exit;

  //         ,  
  //   , (    )
  //  
  for j:=0 to HDetailBand.Bands.Count-1 do
    if (HDetailBand.Bands[j].BandType=bthDetailHeader) and TprCustomHDetailHeaderBand(HDetailBand.Bands[j]).ReprintOnEachPage then
      begin
        //        
        V:=Grid.FindPriorHVector(CurVectorIndex,HDetailBand.Bands[j],StartEndPageIndex);
        if V<>nil then
          begin
            V      :=Grid.CopyHVector(V,CurVectorIndex);
            V.Page :=PageY;
            V.XOffs:=Report.CurEndPage.Rect.Left;
            V.YOffs:=Report.CurEndPage.Rect.Top;
            Report.CurEndPage.Rect.Top:=Report.CurEndPage.Rect.Top+V.Height;
          end;
      end;
  end;

  procedure MoveObjects(Cell : TprGenCell);
  var
    i : integer;
  begin
  for i:=0 to Cell.ObjRecsCount-1 do
    begin
      Cell.ObjRecs[i].pRect.Left  :=Cell.ObjRecs[i].pRect.Left+Cell.XOffs;
      Cell.ObjRecs[i].pRect.Top   :=Cell.ObjRecs[i].pRect.Top+Cell.YOffs;
      Cell.ObjRecs[i].pRect.Right :=Cell.ObjRecs[i].pRect.Right+Cell.XOffs;
      Cell.ObjRecs[i].pRect.Bottom:=Cell.ObjRecs[i].pRect.Bottom+Cell.YOffs;

      if (Cell.VVector<>nil) and (Cell.HVector<>nil) then
        begin
          Cell.ObjRecs[i].pRect.Left :=Cell.ObjRecs[i].pRect.Left-Cell.VVector.Band.dPageRect.Left;
          Cell.ObjRecs[i].pRect.Right:=Cell.ObjRecs[i].pRect.Right-Cell.VVector.Band.dPageRect.Left;
          if dPageRect<>nil then
            begin
              Cell.ObjRecs[i].pRect.Left :=Cell.ObjRecs[i].pRect.Left+dPageRect.Left;
              Cell.ObjRecs[i].pRect.Right:=Cell.ObjRecs[i].pRect.Right+dPageRect.Left;
            end;
        end;

      Cell.ObjRecs[i].Page:=Cell.Page;
      Cell.Page.FoRecs.Add(Cell.ObjRecs[i]);
    end;
  end;

  //  :
  //      
  //  :
  //    
  //      
  //      
  //    /    
  //
  procedure SelectEndPage;
  var
    Cell : TprGenCell;
    i,j,Left,Top : integer;
  begin
  i:=StartEndPageIndex;
  while (i<Report.EndPagesCount) and
        ((Report.EndPages[i].PageX<>PageX) or
         (Report.EndPages[i].PageY<>PageY)) do Inc(i);
  if i>=Report.EndPagesCount then
    begin
      //  
      if (Report.CurEndPage<>nil) and (StartEndPageIndex<Report.EndPagesCount) then
        Report.DoOnPageEnd(Report.CurEndPage);

      //   
      Report.AddEndPage(Self);
      Report.CurEndPage.PageX:=PageX;
      Report.CurEndPage.PageY:=PageY;

      //  
      Report.DoOnPageStart(Report.CurEndPage);

      //  ProgressForm
      Report.UpdateProgressForm(Format(prLoadStr(sSecondPassCaption),[Report.EndPagesCount]));

      Left:=Report.CurEndPage.Rect.Left;
      Top :=Report.CurEndPage.Rect.Top;

      if (PageX=0) and (Grid.vTitleCell<>nil) then
        begin
          if PageY=0 then
            begin
              //    s
              Grid.vTitleCell.Page  :=Report.CurEndPage;
              Grid.vTitleCell.XOffs :=Left;
              Grid.vTitleCell.YOffs :=Top;
              MoveObjects(Grid.vTitleCell);
            end;
          Report.CurEndPage.Rect.Left:=Report.CurEndPage.Rect.Left+Grid.vTitleCell.Width;
        end;

      if (PageY=0) and (Grid.hTitleCell<>nil) then
        begin
          if PageX=0 then
            begin
              //    
              Grid.hTitleCell.Page  :=Report.CurEndPage;
              Grid.hTitleCell.XOffs :=Left;
              Grid.hTitleCell.YOffs :=Top;
              MoveObjects(Grid.hTitleCell);
            end;
          Report.CurEndPage.Rect.Top:=Report.CurEndPage.Rect.Top+Grid.hTitleCell.Height;
        end;

      //  / 
      for i:=0 to High(PageHeadersFootersBands) do
        for j:=0 to Bands.Count-1 do
          if Bands[j].BandType=PageHeadersFootersBands[i] then
            begin
              fPlace:=false;
              case Bands[j].BandType of
                btvPageHeader: fPlace:=TprCustomVPageHeaderBand(Bands[j]).PrintOnFirstPage or (PageX>0);
                btvPageFooter: fPlace:=TprCustomVPageFooterBand(Bands[j]).PrintOnFirstPage or (PageX>0);
                bthPageHeader: fPlace:=TprCustomHPageHeaderBand(Bands[j]).PrintOnFirstPage or (PageY>0);
                bthPageFooter: fPlace:=TprCustomHPageFooterBand(Bands[j]).PrintOnFirstPage or (PageY>0);
              end;
              if fPlace then
                begin
                  if Bands[j].BandType in [btvPageHeader,btvPageFooter] then
                    TprCustomVBand(Bands[j]).FCrossTabGenerated:=false;
                  Cell:=Bands[j].GenerateCell(nil,nil,true);
                  if Cell<>nil then
                    try
                      Cell.Page :=Report.CurEndPage;

                      case Bands[j].BandType of
                        btvPageHeader:
                          begin
                            Cell.YOffs                 :=Top;
                            Cell.XOffs                 :=Report.CurEndPage.Rect.Left;
                            Report.CurEndPage.Rect.Left:=Report.CurEndPage.Rect.Left+Cell.Width;
                          end;
                        btvPageFooter:
                          begin
                            Cell.YOffs                  :=Top;
                            Report.CurEndPage.Rect.Right:=Report.CurEndPage.Rect.Right-Cell.Width;
                            Cell.XOffs                  :=Report.CurEndPage.Rect.Right;
                          end;
                        bthPageHeader:
                          begin
                            Cell.XOffs                :=Left;
                            Cell.YOffs                :=Report.CurEndPage.Rect.Top;
                            Report.CurEndPage.Rect.Top:=Report.CurEndPage.Rect.Top+Cell.Height;
                          end;
                        bthPageFooter:
                          begin
                            Cell.XOffs                   :=Left;
                            Report.CurEndPage.Rect.Bottom:=Report.CurEndPage.Rect.Bottom-Cell.Height;
                            Cell.YOffs                   :=Report.CurEndPage.Rect.Bottom;
                          end;
                      end;
                      MoveObjects(Cell);
                    finally
                      Cell.Free;
                    end;
                end;
            end;

      Report.CurEndPage.AvailableRect:=Report.CurEndPage.Rect;
    end
  else
    begin
      Report.IndexCurEndPage:=i;
    end;
  end;


  function CheckHLinksList : boolean;
  var
    i : integer;
    MinVector : TprGenVector;
  begin
  Result:=false;
  if LinksList.Count>0 then
    begin
      MinVector:=LinksList[0].Vector;
      for i:=1 to LinksList.Count-1 do
        if MinVector.YOffs>LinksList[i].Vector.YOffs then
          MinVector:=LinksList[i].Vector;

      //    MinVector   
      //         
      //        :((.

      //  
      LinksList.FreeAll;

      CurVectorIndex:=Grid.IndexOfHVector(MinVector);
      Result:=true;
    end;
  end;

begin
LinksList:=TLinksList.Create;
try
//          
//   Band     
// ,      
for i:=0 to Grid.HVectorsCount-1 do
  begin
    Vector :=Grid.HVectors[i];
    MaxCell:=Vector[0];
    MinCell:=Vector[0];
    for j:=1 to Vector.Count-1 do
      begin
        if Vector[j].Height>MaxCell.Height then
          MaxCell:=Vector[j];
        if Vector[j].Height<MinCell.Height then
          MinCell:=Vector[j];
      end;
    case Vector.Band.FResizeMode of
      prbrmNone,prbrmMaxObj,prbrmMaxResizeObj:
        Vector.Height:=MaxCell.Height;
      prbrmMinResizeObj:
        Vector.Height:=MinCell.Height;
    end
  end;

for i:=0 to Grid.VVectorsCount-1 do
  begin
    Vector :=Grid.VVectors[i];
    MaxCell:=Vector[0];
    MinCell:=Vector[0];
    for j:=1 to Vector.Count-1 do
      begin
        if Vector[j].Width>MaxCell.Width then
          MaxCell:=Vector[j];
        if Vector[j].Width<MinCell.Width then
          MinCell:=Vector[j];
      end;
    case Vector.Band.FResizeMode of
      prbrmNone,prbrmMaxObj,prbrmMaxResizeObj:
        Vector.Width:=MaxCell.Width;
      prbrmMinResizeObj:
        Vector.Width:=MinCell.Width;
    end
  end;

//  ,     
StartEndPageIndex:=Report.EndPagesCount;

//   
LinksList.FreeAll;

PageX:=0;
PageY:=0;
SelectEndPage;
CurVectorIndex:=0;
while CurVectorIndex<Grid.VVectorsCount do
  begin
    Vector:=Grid.VVectors[CurVectorIndex];
    Band  :=Vector.Band;

    if Band.GetStartNewPage then
      begin
        //     
        //     ,       
        if Grid.FindPriorVVector(CurVectorIndex,Band,StartEndPageIndex)<>nil then
          begin
            Inc(PageX);
            SelectEndPage;

            //  ( )   
            //      
            InsertVerticalDetailHeader(Band);

            LinksList.FreeAll;
          end
      end
    else
      begin
        if Report.CurEndPage.CurrentWidth<Vector.Width then
          begin
            Inc(PageX);
            SelectEndPage;

            //  ( )   
            InsertVerticalDetailHeader(Band);

            //     ,
            //   LinksList
            //        
            //   
            if LinksList.Count>0 then
              begin
                MinVector:=LinksList[0].Vector;
                for j:=1 to LinksList.Count-1 do
                  if MinVector.XOffs>LinksList[j].Vector.XOffs then
                    MinVector:=LinksList[j].Vector;

                //    MinVector   
                //         
                //        :((.

                //  
                LinksList.FreeAll;


                CurVectorIndex:=Grid.IndexOfVVector(MinVector);
                continue;
              end;
          end;
      end;

    if (Band.GetLinkToBand<>nil) and (Band.GetLinkType=prltAfter) then
      //  Band      
      //  Band  
      LinksList.Add(TLinkInfo.Create(Vector,Band.GetLinkToBand));

    if (Band.GetLinkToBand<>nil) and (Band.GetLinkType=prltBefore) then
      begin
        //         GetLinkToBand
        V:=Grid.FindPriorVVectorOnPage(CurVectorIndex,Band.GetLinkToBand,PageX);
        if V=nil then
          begin
            //       !!!
            //      
            V:=Grid.FindPriorVVectorOnPage(CurVectorIndex,Band.GetLinkToBand,PageX-1);
            if V<>nil then
              begin
                //  V         
                //    CurVectorIndex,      V
                //     ,    
                //    ,     
                //   Vector      
                //    :   
                //   ,   
                //  -  ,    
                j:=Grid.IndexOfVVector(V);
                MinVector:=nil;
                while (j>=0) and (Grid.VVectors[j].Page=PageX-1) do
                  begin
                    if Grid.VVectors[j].FLinkToVector=V then
                      MinVector:=Grid.VVectors[j];
                    Dec(j);
                  end;
                if MinVector<>nil then
                  V:=MinVector;

                PriorEndPage:=FindEndPage(PageX-1,0);
                if PriorEndPage.CurrentWidth+(V.XOffs-PriorEndPage.AvailableRect.Left)>=V.Width then
                  begin
                    //  
                    CurVectorIndex:=Grid.IndexOfVVector(V);
                    continue;
                  end;
              end;
          end;
      end;

    //    LinksList , 
    //   Band     Vector
    DeleteBandFromLinksList(Vector,Band);

    Vector.Page                :=PageX;
    Vector.YOffs               :=-1;
    Vector.XOffs               :=Report.CurEndPage.Rect.Left;
    Report.CurEndPage.Rect.Left:=Report.CurEndPage.Rect.Left+Vector.Width;

    Inc(CurVectorIndex);
  end;






//   ,   ,  
// - 
LinksList.FreeAll;
PageX:=0;
PageY:=0;
SelectEndPage;
CurVectorIndex:=0;
while CurVectorIndex<Grid.HVectorsCount do
  begin
    Vector:=Grid.HVectors[CurVectorIndex];
    Band  :=Vector.Band;

    if TprCustomHBand(Vector.Band).GetUseColumns then
      begin
        //
        //  
        //
        //     
        StartPageTop     :=Report.CurEndPage.Rect.Top;
        MaxTop           :=Report.CurEndPage.Rect.Top;
        PriorColDirection:=TprCustomHBand(Vector.Band).GetColDirection;
        while (CurVectorIndex<Grid.HVectorsCount) and
              (TprCustomHBand(Grid.HVectors[CurVectorIndex].Band).GetUseColumns) and
              (TprCustomHBand(Grid.HVectors[CurVectorIndex].Band).GetColDirection=PriorColDirection) do
          begin
            Vector  :=Grid.HVectors[CurVectorIndex];
            ColWidth:=(Report.CurEndPage.AvailableRect.Right-Report.CurEndPage.AvailableRect.Left) div TprCustomHBand(Vector.Band).GetColCount;

            case PriorColDirection of
              prcdTopBottomLeftRight:
                begin
                  //    
                  if Report.CurEndPage.CurrentHeight<Vector.Height then
                    begin
                      Report.CurEndPage.Rect.Left:=Report.CurEndPage.Rect.Left+ColWidth;
                      if Report.CurEndPage.CurrentWidth<ColWidth then
                        begin
                          //  
                          Inc(PageY);
                          SelectEndPage;
                          MaxTop      :=Report.CurEndPage.Rect.Top;
                          StartPageTop:=Report.CurEndPage.Rect.Top;

                          if CheckHLinksList then
                            continue;
                        end
                      else
                        //  
                        Report.CurEndPage.Rect.Top:=StartPageTop;
                    end;
                  Vector.Page               :=PageY;
                  Vector.XOffs              :=Report.CurEndPage.Rect.Left;
                  Vector.YOffs              :=Report.CurEndPage.Rect.Top;
                  Report.CurEndPage.Rect.Top:=Report.CurEndPage.Rect.Top+Vector.Height;

                  if Report.CurEndPage.Rect.Top>MaxTop then
                    MaxTop:=Report.CurEndPage.Rect.Top;
                end;
              prcdLeftRightTopBottom:
                begin
                  if Report.CurEndPage.CurrentWidth<ColWidth then
                    begin
                      //    
                      Report.CurEndPage.Rect.Top :=MaxTop;
                      Report.CurEndPage.Rect.Left:=Report.CurEndPage.AvailableRect.Left;
                    end;

                  if Report.CurEndPage.CurrentHeight<Vector.Height then
                    begin
                      //    
                      Inc(PageY);
                      SelectEndPage;
                      MaxTop:=Report.CurEndPage.Rect.Top;

                      if CheckHLinksList then
                        continue;
                    end;

                  Vector.Page                :=PageY;
                  Vector.XOffs               :=Report.CurEndPage.Rect.Left;
                  Vector.YOffs               :=Report.CurEndPage.Rect.Top;
                  Report.CurEndPage.Rect.Left:=Report.CurEndPage.Rect.Left+ColWidth;

                  if Report.CurEndPage.Rect.Top+Vector.Height>MaxTop then
                    MaxTop:=Report.CurEndPage.Rect.Top+Vector.Height;
                end;
            end;
            DeleteBandFromLinksList(Vector,Vector.Band);
            //
            Inc(CurVectorIndex);
          end;

        //   
        Report.CurEndPage.Rect.Top :=MaxTop;
        Report.CurEndPage.Rect.Left:=Report.CurEndPage.AvailableRect.Left;
      end
    else
      begin
        //
        //   
        //
        if Band.GetStartNewPage then
          begin
            //     
            //     ,       
            if Grid.FindPriorHVector(CurVectorIndex,Band,StartEndPageIndex)<>nil then
              begin
                Inc(PageY);
                SelectEndPage;

                //  ( )   
                //      
                InsertHorizontalDetailHeader(Band);

                LinksList.FreeAll;
              end;
          end
        else
          begin
            if Report.CurEndPage.CurrentHeight<Vector.Height then
              begin
                Inc(PageY);
                SelectEndPage;

                //  ( )   
                InsertHorizontalDetailHeader(Band);

                //     ,
                //   LinksList
                //        
                //   
                if CheckHLinksList then
                  continue;
              end;
          end;
    
        if (Band.GetLinkToBand<>nil) and (Band.GetLinkType=prltAfter) then
          //  Band      
          //  Band  
          LinksList.Add(TLinkInfo.Create(Vector,Band.GetLinkToBand));
    
        if (Band.GetLinkToBand<>nil) and (Band.GetLinkType=prltBefore) then
          begin
            //         GetLinkToBand
            V:=Grid.FindPriorHVectorOnPage(CurVectorIndex,Band.GetLinkToBand,PageY);
            if V=nil then
              begin
                //       !!!
                //      
                V:=Grid.FindPriorHVectorOnPage(CurVectorIndex,Band.GetLinkToBand,PageY-1);
                if V<>nil then
                  begin
                    //  V         
                    //    CurVectorIndex,      V
                    //     ,    
                    //    ,     
                    //   Vector      
                    //    :   
                    //   ,   
                    //  -  ,    
                    j:=Grid.IndexOfHVector(V);
                    MinVector:=nil;
                    while (j>=0) and (Grid.HVectors[j].Page=PageY-1) do
                      begin
                        if Grid.HVectors[j].FLinkToVector=V then
                          MinVector:=Grid.HVectors[j];
                        Dec(j);
                      end;
                    if MinVector<>nil then
                      V:=MinVector;
    
                    PriorEndPage:=FindEndPage(0,PageY-1);
                    if PriorEndPage.CurrentHeight+(V.YOffs-PriorEndPage.AvailableRect.Top)>=V.Height then
                      begin
                        //  
                        CurVectorIndex:=Grid.IndexOfHVector(V);
                        continue;
                      end;
                  end;
              end;
          end;
    
        //    LinksList , 
        //   Band     Vector
        DeleteBandFromLinksList(Vector,Band);

        Vector.Page               :=PageY;
        Vector.XOffs              :=Report.CurEndPage.AvailableRect.Left;
        Vector.YOffs              :=Report.CurEndPage.Rect.Top;
        Report.CurEndPage.Rect.Top:=Report.CurEndPage.Rect.Top+Vector.Height;
        //    
        Inc(CurVectorIndex);
      end;
  end;

//       
for i:=0 to Grid.CellsCount-1 do
  begin
    Cell:=Grid.Cells[i];
    if Cell.VVector=nil then
      begin
        //     -   
        //   
        //    ,    
        //    CrossTab ( )
        PageX:=0;
        PageY:=Cell.HVector.Page;
        SelectEndPage;
        Cell.Page :=Report.CurEndPage;
        Cell.YOffs:=Cell.HVector.YOffs;
        Cell.XOffs:=Cell.HVector.XOffs
      end
    else
      if Cell.HVector=nil then
        begin
          PageX:=Cell.VVector.Page;
          PageY:=0;
          SelectEndPage;
          Cell.Page :=Report.CurEndPage;
          Cell.XOffs:=Cell.VVector.XOffs;
          Cell.YOffs:=Report.CurEndPage.AvailableRect.Top;
        end
      else
        begin
          PageX:=Cell.VVector.Page;
          PageY:=Cell.HVector.Page;
          SelectEndPage;
          Cell.Page :=Report.CurEndPage;
          Cell.XOffs:=Cell.VVector.XOffs;
          Cell.YOffs:=Cell.HVector.YOffs;
        end;
  end;

//   ,   
//    
//  - ,  - ,      ,
//         (     )
if Grid.VVectorsCount>0 then
  begin
    for i:=0 to Report.EndPagesCount-1 do
      for j:=i+1 to Report.EndPagesCount-1 do
        if (Report.EndPages[i].PageY>Report.EndPages[j].PageY) or
           ((Report.EndPages[i].PageY=Report.EndPages[j].PageY) and (Report.EndPages[i].PageX>Report.EndPages[j].PageX)) then
          Report.FEndPages.Exchange(i,j);
  end;

//   ,  -     
// EndPage
for i:=0 to Grid.CellsCount-1 do
  MoveObjects(Grid.Cells[i]);

finally
  LinksList.Free;
end;
end;


procedure TprCustomPage.UpdateBandsPageRect;
var
  r : TRect;
  i,j : integer;
begin
for i:=0 to Bands.Count-1 do
  Bands[i].FCalced:=false;

r:=GetPageRect;
for i:=0 to MAX_PAGEBANDSORDER do
  for j:=0 to Bands.Count-1 do
    if Bands[j].BandType=PageBandsOrder[i] then
      Bands[j].CalcdPageRect(r,nil);

r:=GetPageRect;
for i:=0 to MAX_PAGEBANDSORDERVERT do
  for j:=0 to Bands.Count-1 do
    if Bands[j].BandType=PageBandsOrderVert[i] then
      Bands[j].CalcdPageRect(r,nil);
end;

procedure TprCustomPage.DrawDesign;
begin
end;

///////////////////////////
//
// TprCustomReport
//
///////////////////////////
constructor TprCustomReport.Create;
begin
inherited;
ObjectCalcSizesDevice:=nil;

Designer             :=nil;

FGACList             :=TStringList.Create;
FReportPrepared      :=false;
                     
FCollate             :=false;
FCopies              :=1;
FFromPage            :=-1;
FToPage              :=-1;
                     
FEndPages            :=TList.Create;
FPages               :=TList.Create;
FPrintPagesList      :=TList.Create;
Groups               :=TprGroups.Create;
FValues              :=TprValues.Create(TprValue);
FValues.Report       :=Self;
SystemValues         :=TprValues.Create(TprValue);
SystemValues.Report  :=Self;
FParser              :=TprParser.Create(Self);

with TprValue(SystemValues.Add) do
  begin
    Name    :='PagesCount';
    ResetOn :=rvtReport;
    CalcOn  :=cvtEventOnReset;
    OnCalc  :=OnCalcPagesCount;
  end;
with TprValue(SystemValues.Add) do
  begin
    Name                   :='Page';
    ResetOn                :=rvtReport;
    CalcOn                 :=cvtEventOnReset;
    OnGetVersionByVersionID:=OnCalcPageNo;
    OnCalc                 :=OnCalcPageNo2;
  end;
end;

destructor TprCustomReport.Destroy;
begin
{$IFDEF DEBUG}
WriteToLog('TprCustomReport.Destroy');
{$ENDIF}
if (Application<>nil) and (not Application.Terminated) then
  begin
    if FProgressForm<>nil then
      FProgressForm.Release;
    {$IFDEF DEBUG}
    WriteToLog('FProgressForm.Release');
    {$ENDIF}
    if FPreviewForm<>nil then
      FPreviewForm.Release;
    {$IFDEF DEBUG}
    WriteToLog('FPreviewForm.Release');
    {$ENDIF}
    if FDesignerForm<>nil then
      FDesignerForm.Release;
    {$IFDEF DEBUG}
    WriteToLog('FDesignerForm.Release');
    {$ENDIF}
  end;

while PagesCount>0 do
  Pages[0].Free;

while Groups.Count>0 do
  Groups[0].Free;

FEndPages.Free;
FPages.Free;
FParser.Free;
FValues.Free;
SystemValues.Free;
Groups.Free;
FGACList.Free;

inherited;
end;

procedure TprCustomReport.Loaded;
var
  i,j,k : integer;
begin
{$IFDEF DEBUG}
WriteToLog('TprCustomReport.Loaded');
{$ENDIF}
inherited;
for i:=0 to PagesCount-1 do
  for j:=0 to Pages[i].Bands.Count-1 do
    for k:=0 to Pages[i].Bands[j].Objects.Count-1 do
      Pages[i].Bands[j].Objects[k].AfterReportLoaded;
for i:=0 to PagesCount-1 do
  for j:=0 to Pages[i].Bands.Count-1 do
    Pages[i].Bands[j].AfterReportLoaded;
end;

function TprCustomReport.GetChildOwner;
begin
Result:=Self;
end;

procedure TprCustomReport.GetChildren;
var
  i : integer;
begin
for i:=0 to PagesCount-1 do
  Proc(Pages[i]);
for i:=0 to Groups.Count-1 do
  Proc(Groups[i]);
end;

procedure TprCustomReport.Notification;
var
  i : integer;
begin
{$IFDEF DEBUG}
WriteToLog('TprCustomReport.Notification : '+Name+' '+GetEnumName(TypeInfo(TOperation),integer(Operation))+' AComponent : '+AComponent.Name);
{$ENDIF}
if (Operation=opRemove) and (AComponent<>Self) then
  begin
    if AComponent is TprGroup then
      begin
        Groups.Remove(AComponent);
        for i:=0 to AllValuesCount-1 do
          if AllValues[i].Group=AComponent then
            AllValues[i].Group:=nil;
      end;

    if AComponent is TprCustomPage then
      FPages.Remove(AComponent);
  end;

inherited;
end;

function TprCustomReport.GetprOwner;
begin
Result:=Self;
{
if Owner=nil then
  Result:=Self
else
//  Result:=Owner;
  Result:=Self;
}
end;

function TprCustomReport.GetAllValuesCount;
begin
Result:=FValues.Count+SystemValues.Count;
end;

function TprCustomReport.GetAllValue;
begin
if i<SystemValues.Count then
  Result:=SystemValues[i]
else
  Result:=FValues[i-SystemValues.Count];
end;

function TprCustomReport.ADSRN_GetIndex(Dataset : TprDatasetLink) : integer;
begin
Result:=0;
while (Result<Length(ADSRN)) and (ADSRN[Result].DataSet<>DataSet) do Inc(Result);
if Result>=Length(ADSRN) then
  Result:=-1;
end;

function TprCustomReport.ADSRN_GetIndex(Dataset : TObject) : integer;
begin
Result:=0;
while (Result<Length(ADSRN)) and (ADSRN[Result].DataSet.Dataset<>DataSet) do Inc(Result);
if Result>=Length(ADSRN) then
  Result:=-1;
end;

procedure TprCustomReport.ADSRN_First;
var
  i : integer;
begin
i:=ADSRN_GetIndex(Dataset);
if i<>-1 then
  ADSRN[i].RecNo:=1
else
  begin
    SetLength(ADSRN,Length(ADSRN)+1);
    ADSRN[High(ADSRN)].DataSet:=DataSet;
    ADSRN[High(ADSRN)].RecNo  :=1;
  end;
end;

procedure TprCustomReport.ADSRN_Next;
var
  i : integer;
begin
i:=ADSRN_GetIndex(Dataset);
if (i<>-1) and (not ADSRN[i].DataSet.Eof) then
  ADSRN[i].RecNo:=ADSRN[i].RecNo+1
else
  begin
    SetLength(ADSRN,Length(ADSRN)+1);
    ADSRN[High(ADSRN)].DataSet:=DataSet;
    ADSRN[High(ADSRN)].RecNo  :=1;
  end;
end;

function  TprCustomReport.GetDataSetRecNo(Dataset : TprDatasetLink) : integer;
var
  i : integer;
begin
i:=ADSRN_GetIndex(Dataset);
if i<>-1 then
  Result:=ADSRN[i].RecNo
else
  Result:=-1;
end;

function  TprCustomReport.GetDataSetRecNo(Dataset : TObject) : integer;
var
  i : integer;
begin
i:=ADSRN_GetIndex(Dataset);
if i<>-1 then
  Result:=ADSRN[i].RecNo
else
  Result:=-1;
end;

procedure TprCustomReport.OnCalcPagesCount;
begin
Value.CurrentValue:=EndPagesCount;
end;

procedure TprCustomReport.OnCalcPageNo;
begin
ValueVersion.V:=IndexCurEndPage+1;
end;

procedure TprCustomReport.OnCalcPageNo2;
begin
Value.CurrentValue:=0;
end;

function TprCustomReport.GetPage;
begin
Result:=TprCustomPage(FPages[index]);
end;

function TprCustomReport.GetEndPage;
begin
Result:=TprCustomEndPage(FEndPages[index]);
end;

function TprCustomReport.PagesCount;
begin
Result:=FPages.Count;
end;

function TprCustomReport.EndPagesCount;
begin
Result:=FEndPages.Count;
end;

function TprCustomReport.CurEndPage;
begin
if (IndexCurEndPage<0) or (IndexCurEndPage>=FEndPages.Count) then
  Result:=nil
else
  Result:=TprCustomEndPage(FEndPages[IndexCurEndPage]);
end;

procedure TprCustomReport.AddEndPage;
begin
FEndPages.Add(CreateEndPage(Page));

IndexCurEndPage:=FEndPages.Count-1;
end;

procedure TprCustomReport.AddEndPageToList;
begin
FEndPages.Add(EndPage);
end;

procedure TprCustomReport.DeleteEndPage;
begin
EndPages[index].Free;
FEndPages.Delete(index);
end;

function TprCustomReport.Calc;
var
  v : TprVarValue;
begin
if TprParser(FParser).Calc(Expression,v,Value) then
  Result:=_vAsVariant(v)
else
  raise Exception.Create(prLoadStr(sErrorCalcExpressionInOnePass));
end;

function TprCustomReport.FormatTemplate;
begin
Result:=TprParser(FParser).FormatTemplate(Template,Res);
end;

function TprCustomReport.FormatStrings;
var
  i : integer;
  s,Buf : string;
begin
Buf   :='';
Result:=true;
for i:=0 to lSource.Count-1 do
  begin
    Result:=FormatTemplate(lSource[i],s) and Result;
    //   #13,#10
    if (Trim(s)<>'') or (not DeleteEmptyLines) then
      begin
        if Buf='' then
          Buf:=s
        else
          Buf:=Buf+#13#10+s;
      end;
  end;

lDest.Text:=Buf;

if not DeleteEmptyLines and DeleteEmptyLinesAtEnd then
  begin
    while lDest.Count>0 do
      begin
        if Trim(lDest[lDest.Count-1])<>'' then
          break;
        lDest.Delete(lDest.Count-1);
      end;
  end;
end;

procedure TprCustomReport.DoOnBandGenerateCell;
begin
if Assigned(OnBandGenerateCell) then
  OnBandGenerateCell(Self,Band);
end;

procedure TprCustomReport.DoOnDataSetNext;
var
  i : integer;
begin
for i:=0 to AllValuesCount-1 do
  begin
    if (AllValues[i].CalcOn=cvtDataSetNext) and (AllValues[i].Dataset.Dataset=DataSet.Dataset) then
      Calc(AllValues[i].FFormula,AllValues[i]);

    if (AllValues[i].CalcOn=cvtCrossTab) and (AllValues[i].CrossTabHorzDataSet.Dataset=DataSet.Dataset) then
      begin
        //    
        //        
        //  DetailBand
        if (HorizontalBand<>nil) and (HorizontalBand.BandType in [bthDetail]) then
          Calc(AllValues[i].FFormula,AllValues[i]);
      end;
  end;
end;

procedure TprCustomReport.DoOnGroupEnd;
var
  i : integer;
begin
for i:=0 to AllValuesCount-1 do
  begin
    if (AllValues[i].ResetOn=rvtGroup) and (AllValues[i].Group=Group) then
      begin
        AllValues[i].Reset;
      end;
  end;
end;

procedure TprCustomReport.DoOnDataSetEof;
var
  i : integer;
begin
for i:=0 to AllValuesCount-1 do
  begin
    if (AllValues[i].ResetOn=rvtDataSetEof) and (AllValues[i].ResetDataSet.Dataset=DataSet.Dataset) then
      begin
        AllValues[i].Reset;
      end;

    if AllValues[i].CalcOn=cvtCrossTab then
      begin
        if AllValues[i].CrossTabHorzDataSet.Dataset=DataSet.Dataset then
          begin
            //  Eof   
          end;
      end;
  end;
end;

procedure TprCustomReport.DoOnPageEnd;
begin
if Assigned(FOnPageEnd) then
  FOnPageEnd(Self,EndPage);
end;

procedure TprCustomReport.DoOnPageStart;
begin
if Assigned(FOnPageStart) then
  FOnPageStart(Self,EndPage);
end;

procedure TprCustomReport.DoOnFirstPassObject;
begin
ManuallyProcessed:=false;
if Assigned(FOnFirstPassObject) then
  begin
    FOnFirstPassObject(Self,Sender,ManuallyProcessed);
  end;
end;

procedure TprCustomReport.ClearEndPages;
begin
//  EndPages
while EndPagesCount>0 do
  DeleteEndPage(0);
end;

procedure TprCustomReport.ConvertToDesignerCoords;
begin
if DesignerForm<>nil then
  DesignerForm.ConvertToDesignerCoords(rSource,rDest)
else
  rDest:=rSource;
end;

//
//      
//
function TprCustomReport.PrepareReport;
var
  i : integer;
  ep : TprCustomEndPage;
begin
FActionCanceled:=false;
Result         :=false;
FReportPrepared:=false;

if PreviewForm<>nil then
  PreviewForm.Release;

StartDateTime  :=Now;
IndexCurEndPage:=-1;
FProgressForm  :=nil;

if ShowProgress then
  begin
    FProgressForm:=TprProgressForm.Create(Application);
    FProgressForm.InitPF(prLoadStr(sCreateReportCaption),0,nil);
  end;

try
  try
    SetLength(ADSRN,0);

    //    
    ClearEndPages;

    //   
    for i:=0 to AllValuesCount-1 do
      begin
        AllValues[i].Clear;
        AllValues[i].Init;
      end;

    //  ,  Grid
    i:=0;
    while i<PagesCount do
      begin
        Pages[i].FirstPassGenerateGrid;
        Inc(i);
      end;

    //    Grid  
    i:=0;
    while i<PagesCount do
      begin
        Pages[i].SecondPassPlaceGridOnEndPage;
        Inc(i);
      end;

    //   ,     
    for i:=0 to AllValuesCount-1 do
      AllValues[i].Reset;

    //      
    i:=0;
    while i<EndPagesCount do
      begin
        UpdateProgressForm(Format(prLoadStr(sThirdPassCaption),[i+1,EndPagesCount]));

        IndexCurEndPage:=i;
        ep:=EndPages[i];

        ep.ThirdPass;

        ep.FreeGeneratedObjects;

        Inc(i);
      end;

    FReportPrepared:=true;
    Result         :=true;
  except
    on E : Exception do
      begin
        ClearEndPages;
        if E is EActionCanceled then
          FActionCanceled:=true
        else
          raise;
      end;
  end;

finally
  if FProgressForm<>nil then
    begin
      FProgressForm.Release;
      FProgressForm:=nil;
    end;

  Finalize(ADSRN);
end;
end;

function TprCustomReport.GetDataSetByName;
var
  i : integer;
  L : TStringList;
begin
L:=TStringList.Create;
try
  GetAvailableDataSets(L);
  i:=L.IndexOf(DataSetName);
  if i=-1 then
    Result:=nil
  else
    Result:=TDataSet(L.Objects[i]);
finally
  L.Free;
end;
end;

procedure TprCustomReport.GetAvailableDataSets;
var
  i : integer;
begin
if Assigned(OnGetAvailableDatasets) then
  begin
    L.Clear;
    FOnGetAvailableDatasets(Self,L);
  end
else
  begin
    GetAvailableComponents(L);
    //   TDataSet
    i:=0;
    while i<L.Count do
      if (TComponent(L.Objects[i]) is TDataSet) or (TComponent(L.Objects[i]) is TprDataSet) then
        Inc(i)
      else
        L.Delete(i);
  end;
end;

procedure TprCustomReport.GetAvailableComponents;
var
  i : integer;

  procedure AddComponents(C : TComponent);
  var
    i,j : integer;
  begin
  for i:=0 to C.ComponentCount-1 do
    begin
      //     , 
      j:=1;
      while (j<=prMAX_OBJFUNCTIONS) and not (C.Components[i] is ObjFuncInfo[j].ObjClassRef) do Inc(j);
      if j>prMAX_OBJFUNCTIONS then
        begin
          j:=1;
          while (j<=prMAX_OBJPROPS) and not (C.Components[i] is ObjPropInfo[j].ObjClassRef) do Inc(j);
          if j>prMAX_OBJPROPS then continue;
        end;
      //   
      if L.IndexOfObject(C.Components[i])=-1 then
        begin
          if (Owner=C) or (C=Self) then
            L.AddObject(C.Components[i].Name,C.Components[i])
          else
            L.AddObject(C.Name+'.'+C.Components[i].Name,C.Components[i])
        end;
    end;
  end;


begin
L.Clear;
L.AddObject('',Self);

if Assigned(OnGetAvailableComponents) then
  begin
    FOnGetAvailableComponents(Self,L);
  end
else
  begin
    AddComponents(Self);
    if Owner<>nil then
      AddComponents(Owner);

    for i:=0 to Screen.DataModuleCount-1 do
      AddComponents(Screen.DataModules[i]);

    for i:=0 to Screen.FormCount-1 do
      AddComponents(Screen.Forms[i]);
  end;
end;

procedure TprCustomReport.TranslateObjectName;
var
  j,l,p : integer;
  ModuleName : string;


begin
l        :=Length(Ident);
Component:=nil;
LastName :='';

// -   
p:=l;
while (p>=1) and (Ident[p]<>'.') do Dec(p);
if p<1 then exit;

ModuleName:=Copy(Ident,1,p-1);
Component :=Self.FindComponent(ModuleName);
if Component<>nil then
  begin
    LastName:=Copy(Ident,p+1,l);
    exit;
  end;

// p          -     Ident
// ModuleName -    
if Assigned(OnGetAvailableComponents) then
  begin
    FGACList.Clear;
    OnGetAvailableComponents(Self,FGACList);
    
    j:=0;
    while (j<FGACList.Count) and
          (CompText(FGACList[j],ModuleName)<>0) do Inc(j);
    if p<FGACList.Count then
      begin
        Component:=TComponent(FGACList.Objects[j]);
        LastName :=Copy(Ident,p+1,l);
      end;
  end
else
  begin
    //   ( )  :
    //   -   Owner    
    //   - DataModule  
    //   -   
    p        :=1;
    LastName :=ExtractSubStr(Ident,p,['.']);
    Component:=Owner.FindComponent(LastName);
    if Component=nil then
      begin
        j:=0;
        while (j<Screen.DataModuleCount) and
              (CompText(Screen.DataModules[j].Name,ModuleName)<>0) do Inc(j);
        if j>=Screen.DatamoduleCount then
          begin
            j:=0;
            while (j<Screen.FormCount) and
                  (CompText(Screen.Forms[j].Name,ModuleName)<>0) do Inc(j);
            if j<Screen.FormCount then
              Component:=Screen.Forms[j];
          end
        else
          Component:=Screen.DataModules[j];
      end;

    if Component<>nil then
      begin
        LastName:=ExtractSubStr(Ident,p,['.']);
        while (p<=l) and (Component<>nil) do
          begin
            Component:=Component.FindComponent(LastName);
            LastName :=ExtractSubStr(Ident,p,['.']);
          end;
      end;
  end;
end;

procedure TprCustomReport.UpdateProgressForm;
begin
if FProgressForm<>nil then
  FProgressForm.UpdatePF(Text);
end;

procedure TprCustomReport.DesignReport;
var
  FormClass : TClass;
begin
FReportPrepared:=false;
if DesignerForm<>nil then
  begin
    DesignerForm.Show;
  end
else
  begin
    if (DesignerFormMode=fmMDIChild) and Modal then
      raise Exception.Create(prLoadStr(sErrorNoModalForMDIChild));

    FormClass:=GetClass(GetDesignerFormClass);
    if FormClass=nil then
      raise Exception.Create(prLoadStr(sErrorDesignerClassNotFound));

    TprDesignerClass(FormClass).CreateDesigner(nil,Self);
    if Assigned(OnCreateDesigner) then
      OnCreateDesigner(Self);

    if Modal then
      DesignerForm.ShowModal
    else
      DesignerForm.Show;
  end;

if DesignerForm.WindowState=wsMinimized then
  DesignerForm.WindowState:=wsNormal;
end;

procedure TprCustomReport.PreviewPreparedReport;
var
  FormClass : TClass;
begin
if EndPagesCount<=0 then
  raise Exception.Create(prLoadStr(sReportEmptyInPreview));

if PreviewForm<>nil then
  begin
    PreviewForm.Show;
  end
else
  begin
    if (DesignerFormMode=fmMDIChild) and Modal then
      raise Exception.Create(prLoadStr(sErrorNoModalForMDIChild));

    FormClass:=GetClass(GetPreviewFormClass);
    if FormClass=nil then
      raise Exception.Create(prLoadStr(sErrorPreviewClassNotFound));

    TprPreviewClass(FormClass).CreatePreview(nil,Self);
    if Assigned(OnCreatePreview) then
      OnCreatePreview(Self);

    if Modal then
      PreviewForm.ShowModal
    else
      PreviewForm.Show;
  end;

if PreviewForm.WindowState=wsMinimized then
  PreviewForm.WindowState:=wsNormal;
end;

procedure TprCustomReport.ClearTemplate;
begin
while PagesCount>0 do
  Pages[0].Free;

while Groups.Count>0 do
  Groups[0].Free;

Values.Clear;
end;

procedure TprCustomReport.SetTemplateChanged;
begin
FTemplateChanged:=Value;
if Designer<>nil then
  Designer.Modified;
end;

//
//    
//
procedure TprCustomReport.SaveTemplate;
var
  s : string;
  ms : TMemoryStream;
begin
s:=Name;
try
  Name:='';
  if InBinaryFormat then
    begin
      Stream.WriteComponent(Self);
    end
  else
    begin
      ms:=TMemoryStream.Create;
      try
        ms.WriteComponent(Self);
        ms.Seek(0,soFromBeginning);
        ObjectBinaryToText(ms,Stream);
      finally
        ms.Free;
      end;
    end;
  FTemplateChanged:=false;
finally
  Name:=s
end;
end;

procedure TprCustomReport.LoadTemplate;
var
  s : string;
  ms : TMemoryStream;
begin
ClearTemplate;
s:=Name;
try
  if InBinaryFormat then
    begin
      Stream.ReadComponent(Self);
    end
  else
    begin
      ms:=TMemoryStream.Create;
      try
        ObjectTextToBinary(Stream,ms);
        ms.Seek(0,soFromBeginning);
        ms.ReadComponent(Self);
      finally
        ms.Free;
      end;
    end;
  FTemplateChanged:=false;
finally
  Name:=s
end;
end;

procedure TprCustomReport.SaveTemplateToFile;
var
  fs : TFileStream;
begin
fs:=TFileStream.Create(FileName,fmCreate);
try
  SaveTemplate(fs,InBinaryFormat);
finally
  fs.Free;
end;
end;

procedure TprCustomReport.SaveTemplateToStrings;
var
  ms : TMemoryStream;
begin
ms:=TMemoryStream.Create;
try
  SaveTemplate(ms,false);
  ms.Seek(0,soFromBeginning);
  Dest.LoadFromStream(ms);
finally
  ms.Free;
end;
end;

procedure TprCustomReport.LoadTemplateFromStrings;
var
  ms : TMemoryStream;
begin
if Source.Count=0 then exit;
ms:=TMemoryStream.Create;
try
  Source.SaveToStream(ms);
  ms.Seek(0,soFromBeginning);
  LoadTemplate(ms,false);
finally
  ms.Free;
end;
end;

procedure TprCustomReport.LoadTemplateFromFile;
var
  fs : TFileStream;
begin
fs:=TFileStream.Create(FileName,fmOpenRead);
try
  ClearTemplate;
  LoadTemplate(fs,InBinaryFormat);
finally
  fs.Free;
end;
end;

procedure TprCustomReport.LoadPreparedReport;
begin
end;

procedure TprCustomReport.SavePreparedReport;
begin
end;

procedure TprCustomReport.LoadPreparedReportFromFile;
var
  fs : TFileStream;
begin
fs:=TFileStream.Create(FileName,fmOpenRead);
try
  LoadPreparedReport(fs);
finally
  fs.Free;
end;
end;

procedure TprCustomReport.SavePreparedReportToFile;
var
  fs : TFileStream;
begin
fs:=TFileStream.Create(FileName,fmCreate);
try
  SavePreparedReport(fs);
finally
  fs.Free;
end;
end;

function TprCustomReport.GetBandClass;
begin
Result:=nil;
end;

var
  i : TprBandType;

initialization

prGetComponentsReportDesc:=prLoadStr(sGetComponentsReportDesc);

for i:=Low(TprBandType) to High(TprBandType) do
  BandTitles[i]:=prLoadStr(integer(sBandTitlesOffset-integer(i)));

prIniFileName := GetFindFileName(_sPrIniFileName);

finalization

Finalize(prObjRegInfos);

end.


