unit pr_DesignerFunctions;

interface

uses
  Windows, Menus, SysUtils, Classes, StdCtrls, comctrls,
  typinfo, graphics,

  pr_Common, pr_MultiLang;

type
  TprElementMode = (premNotExists,premExists,premExistsOrNot);
  TCardinalSet = set of 0..SizeOf(Cardinal) * 8 - 1;

const
  DesignerEmptyString = #12;

procedure InitprObjToolbar(Designer : TprDesigner;
                           Toolbar  : TToolbar;
                           _OnClick : TNotifyEvent);

procedure InitBandsMenu(Designer  : TprDesigner;
                        PopupMenu : TPopupMenu;
                        _OnClick  : TNotifyEvent);

function ATG(Coord,Step : integer) : integer;
function NormalizeRect(X1,Y1,X2,Y2 : integer) : TRect;

procedure MakedRecDefVersionList(LSource,LDest : TList);

function prGetListItemString(LB : TComboBox) : string;
function prGetListItemObject(LB : TComboBox) : TObject; 

function prGetProp(L : TList; PropName : string; var PropValue : Variant) : boolean;
function prGetPropDef(L : TList; PropName : string; DefPropValue : Variant) : Variant;
function prGetPropDefBool(L : TList; PropName : string) : TCheckBoxState;
function prGetPropDefSet(L : TList; PropName : string; SetElement : integer) : TprElementMode;
procedure prSetProp(L : TList; PropName : string; const PropValue : Variant; IsNull : boolean);
procedure prSetPropSet(L : TList; PropName : string; SetElement : integer; IncludeFlag : boolean; IsNull : boolean);

implementation

uses
  pr_Strings;

procedure MakedRecDefVersionList;
var
  i : integer;
begin
LDest.Clear;
for i:=0 to LSource.Count-1 do
  if TObject(LSource[i]) is TprObj then
    LDest.Add(TprObj(LSource[i]).dRec.Versions[TprObj(LSource[i]).dRec.DefVersion]);
end;

function prGetListItemString;
begin
if LB.ItemIndex=-1 then
  Result:=''
else
  Result:=LB.Items[LB.ItemIndex];
end;

function prGetListItemObject;
begin
if LB.ItemIndex=-1 then
  Result:=nil
else
  Result:=LB.Items.Objects[LB.ItemIndex];
end;

procedure DecompileProp(o : TObject; var obj : TObject; var lpi : PPropInfo; HasSubProp : boolean; PropName,ClassPropName,SubPropName : string);
var
  td : PTypeData;
begin
if HasSubProp then
  begin
    lpi:=GetPropInfo(o.ClassInfo,ClassPropName);
    if (lpi<>nil) and (lpi^.PropType^.Kind=tkClass) then
      begin
        Obj:=TObject(GetOrdProp(o,lpi));
        td :=GetTypeData(lpi^.PropType^);
        lpi:=GetPropInfo(td.ClassType.ClassInfo,SubPropName);
      end;
  end
else
  begin
    Obj:=o;
    lpi:=GetPropInfo(o.ClassInfo,PropName);
  end;
end;

procedure prSetProp;
var
  i,p,PropValueInt : integer;
  Obj : TObject;
  lpi : PPropInfo;
  HasSubProp : boolean;
  PropValueString,ClassPropName,SubPropName : string;
  PropValueExt : extended;
begin
if IsNull then exit;

PropValueInt   :=0;
PropValueExt   :=0.0;
PropValueString:='';
case VarType(PropValue) of
  varSmallint,varInteger,varByte:
    PropValueInt:=PropValue;
  varBoolean:
    if PropValue then
      PropValueInt:=1
    else
      PropValueInt:=0;
  varSingle,varDouble,varCurrency,varDate:
    PropValueExt:=PropValue;
  varOleStr,varString:
    PropValueString:=VarToStr(PropValue);
  else
    exit;
end;

p         :=pos('.',PropName);
HasSubProp:=p<>0;
if HasSubProp then
  begin
    ClassPropName:=Copy(PropName,1,p-1);
    SubPropName  :=Copy(PropName,p+1,Length(PropName));
  end;

for i:=0 to L.Count-1 do
  begin
    DecompileProp(L[i],Obj,lpi,HasSubProp,PropName,ClassPropName,SubPropName);
    if (Obj<>nil) and (lpi<>nil) then
      begin
        case lpi^.PropType^.Kind of
          tkInteger,tkSet,tkEnumeration,tkChar,tkClass:
            SetOrdProp(Obj,lpi,PropValueInt);
          tkString,tkLString:
            SetStrProp(Obj,lpi,PropValueString);
          tkFloat:
            SetFloatProp(Obj,lpi,PropValueExt);
        end;
      end;
  end;
end;

procedure prSetPropSet;
var
//  ti : PTypeInfo;
  i,p : integer;
  Obj : TObject;
  lpi : PPropInfo;
  cs : TCardinalSet;
  HasSubProp : boolean;
  ClassPropName,SubPropName : string;
begin
if IsNull then exit;

p         :=pos('.',PropName);
HasSubProp:=p<>0;
if HasSubProp then
  begin
    ClassPropName:=Copy(PropName,1,p-1);
    SubPropName  :=Copy(PropName,p+1,Length(PropName));
  end;

for i:=0 to L.Count-1 do
  begin
    DecompileProp(L[i],Obj,lpi,HasSubProp,PropName,ClassPropName,SubPropName);
    if (Obj<>nil) and (lpi<>nil) then
      begin
        if lpi^.PropType^.Kind=tkSet then
          begin
            cs:=TCardinalSet(cardinal(GetOrdProp(Obj,lpi)));
            if IncludeFlag then
              Include(cs,SetElement)
            else
              Exclude(cs,SetElement);
            SetOrdProp(Obj,lpi,cardinal(cs));
          end;
      end;
  end;
end;

function prGetProp;
var
//  ti : PTypeInfo;
  f : boolean;
  i,p : integer;
  Obj : TObject;
  Lastlpi,lpi : PPropInfo;
  HasSubProp : boolean;
  FirstOrdProp : longint;
  FirstFloatProp : extended;
  FirstStringProp : string;
  ClassPropName,SubPropName : string;

begin
Result   :=false;
PropValue:=UnAssigned;
if L.Count<=0 then exit;

p         :=pos('.',PropName);
HasSubProp:=p<>0;
if HasSubProp then
  begin
    ClassPropName:=Copy(PropName,1,p-1);
    SubPropName  :=Copy(PropName,p+1,Length(PropName));
  end;

FirstOrdProp   :=0;
FirstStringProp:='';
FirstFloatProp :=0;
Lastlpi        :=nil;

i     :=0;
f     :=true;
Result:=true;
while i<L.Count do
  begin
    DecompileProp(L[i],Obj,lpi,HasSubProp,PropName,ClassPropName,SubPropName);
    if (Obj<>nil) and (lpi<>nil) then
      begin
        Lastlpi:=lpi;
        if f then
          begin
            case lpi^.PropType^.Kind of
              tkInteger,tkSet,tkEnumeration,tkChar,tkClass:
                FirstOrdProp:=GetOrdProp(Obj,lpi);
              tkString,tkLString:
                FirstStringProp:=GetStrProp(Obj,lpi);
              tkFloat:
                FirstFloatProp:=GetFloatProp(Obj,lpi);
            end;
            f:=false;
          end
        else
          begin
            case lpi^.PropType^.Kind of
              tkInteger,tkSet,tkEnumeration,tkChar,tkClass:
                Result:=Result and (FirstOrdProp=GetOrdProp(Obj,lpi));
              tkString,tkLString:
                Result:=Result and (FirstStringProp=GetStrProp(Obj,lpi));
              tkFloat:
                Result:=Result and (FirstFloatProp=GetFloatProp(Obj,lpi));
            end;
          end;
      end;
    Inc(i);
  end;

Result:=Result and (not f);
if Result then
  case Lastlpi^.PropType^.Kind of
    tkInteger,tkSet,tkEnumeration,tkChar,tkClass:
      PropValue:=FirstOrdProp;
    tkString,tkLString:
      PropValue:=FirstStringProp;
    tkFloat:
      PropValue:=FirstFloatProp;
  end
end;

function prGetPropDefSet;
var
//  ti : PTypeInfo;
  i,p : integer;
  Obj : TObject;
  lpi : PPropInfo;
  cs : TCardinalSet;
  HasSubProp : boolean;
  ClassPropName,SubPropName : string;
begin
Result   :=premExistsOrNot;
if L.Count<=0 then exit;

p         :=pos('.',PropName);
HasSubProp:=p<>0;
if HasSubProp then
  begin
    ClassPropName:=Copy(PropName,1,p-1);
    SubPropName  :=Copy(PropName,p+1,Length(PropName));
  end;

i     :=0;
Result:=premExistsOrNot;
while i<L.Count do
  begin
    DecompileProp(L[i],Obj,lpi,HasSubProp,PropName,ClassPropName,SubPropName);
    if (Obj<>nil) and (lpi<>nil) then
      begin
        if lpi^.PropType^.Kind=tkSet then
          begin
            cs:=TCardinalSet(cardinal(GetOrdProp(Obj,lpi)));
            if Result=premExistsOrNot then
              begin
                if SetElement in cs then
                  Result:=premExists
                else
                  Result:=premNotExists;
              end
            else
              begin
                if SetElement in cs then
                  begin
                    if Result=premNotExists then
                      Result:=premExistsOrNot;
                  end
                else
                  begin
                    if Result=premExists then
                      Result:=premExistsOrNot;
                  end;
                if Result=premExistsOrNot then
                  break;
              end;
          end;
      end;
    Inc(i);
  end;
end;

function prGetPropDef;
begin
if not prGetProp(L,PropName,Result) then
  Result:=DefPropValue;
end;

function prGetPropDefBool;
var
  v : Variant;
begin
if not prGetProp(L,PropName,v) then
  Result:=cbGrayed
else
  begin
    if v then
      Result:=cbChecked
    else
      Result:=cbUnchecked;
  end;
end;

function NormalizeRect;
begin
if X1>X2 then
  begin
    Result.Left :=X2;
    Result.Right:=X1;
  end
else
  begin
    Result.Left :=X1;
    Result.Right:=X2;
  end;
if Y1>Y2 then
  begin
    Result.Top   :=Y2;
    Result.Bottom:=Y1;
  end
else
  begin
    Result.Top   :=Y1;
    Result.Bottom:=Y2;
  end;
end;

function ATG;
begin
Result:=(Coord div Step) * Step
end;

procedure InitBandsMenu;
var
  m : TMenuItem;
  bt : TprBandType;
begin
for bt:=Low(TprBandType) to High(TprBandType) do
  begin
    m        :=TMenuItem.Create(Designer);
    m.Caption:=BandTitles[bt];
    m.Tag    :=integer(bt);
    m.Onclick:=_OnClick;
    m.Enabled:=Designer.Report.GetBandClass(bt)<>nil;
    PopupMenu.Items.Add(m);
  end;
end;

procedure InitprObjToolbar;
var
  i : integer;
  b : TBitmap;

begin
b:=TBitmap.Create;
try
  for i:=0 to High(prObjRegInfos) do
    if Designer.Report is prObjRegInfos[i].ReportRef then //// Report.ClassName=prObjRegInfos[i].ReportRef.ClassName then
      begin
        Toolbar.Width:=Toolbar.Width+23;

        with TToolButton.Create(Designer) do
          begin
            Parent     :=Toolbar;
            Left       :=Toolbar.ClientWidth;
            Caption    :=prLoadStr(prObjRegInfos[i].CaptionResID);
            Hint       :=prLoadStr(prObjRegInfos[i].HintResID);
            AllowAllUp :=true;
            Grouped    :=true;
            ShowHint   :=true;
            Tag        :=i;
            Style      :=tbsCheck;
            OnClick    :=_OnClick;

            try
              b.LoadFromResourceName(hInstance,AnsiUpperCase(prObjRegInfos[i].ClassRef.ClassName));
            except
              b.LoadFromResourceName(HInstance,AnsiUpperCase(sDefaultObjResName));
            end;
            ImageIndex:=Toolbar.Images.AddMasked(b,b.TransparentColor);
          end;
      end;
finally
  b.Free;
end;
end;

end.
