unit pr_Preview;

interface     

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ImgList, ActnList, StdCtrls, ToolWin, ComCtrls,
  flatsb,

  pr_Common, pr_Classes, Pr_Utils, pr_MultiLang;

const
  PagesStepX = 5;
  PagesStepY = 5;

  LeftOffs   = 10;
  TopOffs    = 10;
  RightOffs  = 10;
  BottomOffs = 10;

type
  TprScaleMode = (smPageWidth,smOnePagePercent,smPages);
  ////////////////////////
  //
  // TprPrevewBox
  //
  ////////////////////////
  TprPreviewBox = class(TScrollBox)
  private
    FNormalCursor       : HCURSOR;
    FDragCursor         : HCURSOR;
    FNormalCursorLoaded : boolean;
    FDragCursorLoaded   : boolean;
    FLastX              : integer;
    FLastY              : integer;

    FMaxWidth       : integer;
    FMaxHeight      : integer;
    FScaleKoef      : Double;
    FFindPos        : integer;

    FPagesPerHeight : integer;
    FPagesPerWidth  : integer;
    FScaleMode      : TprScaleMode;
    FScalePercent   : integer;
    OldScaleKoef    : double;

    DrawWidth       : integer;   //     
    DrawHeight      : integer;   //     

    FOnScalePercentChanged : TNotifyEvent;

    procedure SetPagesPerHeight(Value : integer);
    procedure SetPagesPerWidth(value : integer);
    procedure SetScaleMode(Value : TprScaleMode);
    procedure SetScalePercent(Value : integer);
    function  GetScaleKoef : integer;
    function  GetFindedCount : integer;
    function  GetIsFindMode : boolean;

    procedure CalcMaxSizes;
    procedure ClearFindList;

    procedure SetRPDI;
    procedure ResetRPDI;

    procedure InternalCalcPagesSizes;
    function  InternalFind : boolean;
  protected
    procedure Resize; override;
    procedure mPaint(var message : TWMPaint); message WM_PAINT;
    procedure mVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure mHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure mKeyDown(var Message : TWMKeyDown); message WM_KEYDOWN;
    procedure mGetDlgCode(var Message : TWMGetDlgCode); message WM_GETDLGCODE;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  public
    Report : TprReport;
    rpdi   : rPrPreviewDrawInfo;

    property PagesPerHeight : integer read FPagesPerHeight write SetPagesPerHeight; //   
    property PagesPerWidth  : integer read FPagesPerWidth write SetPagesPerWidth;  //   
    property ScaleMode      : TprScaleMode read FScaleMode write SetScaleMode;
    property ScalePercent   : integer read FScalePercent write SetScalePercent;    //  (   )
    property ScaleKoef      : integer read GetScaleKoef;
    property FindedCount    : integer read GetFindedCount;
    property IsFindMode     : boolean read GetIsFindMode;

    property OnScalePercentChanged : TNotifyEvent read FOnScalePercentChanged write FOnScalePercentChanged;

    procedure Find;
    procedure CancelFind;
    procedure FindNext;
    procedure FindPrior;

    procedure UpdateSettings;

    procedure SetToNewReport;
    
    procedure CreateParams(var Params : TCreateParams); override;
    constructor CreatePreview(AOwner : TComponent; _Report : TprReport);
    destructor Destroy; override;
  end;

  //////////////////////////
  //
  // TprPreviewForm
  //
  //////////////////////////
  TprPreviewForm = class(TprPreview)
    ActionList: TActionList;
    ImageList: TImageList;
    aPrint: TAction;
    aWholePage: TAction;
    aPages: TAction;
    aClose: TAction;
    aPageWidth: TAction;
    aTwoPages: TAction;
    aFind: TAction;
    aFindNext: TAction;
    aFindPrev: TAction;
    aFindCancel: TAction;
    aSave: TAction;
    aOpen: TAction;
    SaveDialog: TSaveDialog;
    OpenDialog: TOpenDialog;
    SB: TStatusBar;
    FindDialog: TFindDialog;
    aCustomAction: TAction;
    prMLRes1: TprMLRes;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton16: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    ToolButton8: TToolButton;
    ToolButton9: TToolButton;
    EDScale: TEdit;
    ToolButton10: TToolButton;
    ToolButton11: TToolButton;
    ToolButton12: TToolButton;
    ToolButton13: TToolButton;
    ToolButton14: TToolButton;
    ToolButton15: TToolButton;
    procedure FormCreate(Sender: TObject);
    procedure aWholePageExecute(Sender: TObject);
    procedure aCloseExecute(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure aPageWidthExecute(Sender: TObject);
    procedure aTwoPagesExecute(Sender: TObject);
    procedure aPrintExecute(Sender: TObject);
    procedure EDScaleKeyPress(Sender: TObject; var Key: Char);
    procedure aFindNextExecute(Sender: TObject);
    procedure aFindPrevExecute(Sender: TObject);
    procedure aFindNextUpdate(Sender: TObject);
    procedure aFindPrevUpdate(Sender: TObject);
    procedure aFindCancelUpdate(Sender: TObject);
    procedure aFindCancelExecute(Sender: TObject);
    procedure aPagesExecute(Sender: TObject);
    procedure aSaveExecute(Sender: TObject);
    procedure aOpenExecute(Sender: TObject);
    procedure FindDialogFind(Sender: TObject);
    procedure aFindExecute(Sender: TObject);
    procedure SBDrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel;
      const Rect: TRect);
    procedure aCustomActionUpdate(Sender: TObject);
    procedure aCustomActionExecute(Sender: TObject);
  private
    { Private declarations }
    Preview : TprPreviewBox;

    procedure ScaleChanged(Sender : TObject);
    procedure WMGetDlgCode(var Message : TWMGetDlgCode); message WM_GETDLGCODE;

    procedure UpdateStatusBar;
  public
    { Public declarations }

    Dx,dy   : integer;
  end;

implementation

uses pr_Strings;

{$R *.DFM}

////////////////////////
//
// TprPreviewBox
//
////////////////////////
constructor TprPreviewBox.CreatePreview;
begin
Report         :=_Report;

rpdi.FindList  :=TList.Create;

FScaleMode     :=smPageWidth;
FScalePercent  :=100;
FPagesPerHeight:=1;
FPagesPerWidth :=1;

CalcMaxSizes;
SetRPDI;

FNormalCursor:=LoadImage(hInstance,'SCROLL',IMAGE_CURSOR,0,0,LR_DEFAULTSIZE or LR_DEFAULTCOLOR);
FNormalCursorLoaded:=FNormalCursor<>0;
if not FNormalCursorLoaded then
  FNormalCursor:=LoadCursor(0,IDC_ARROW);
FDragCursor:=LoadImage(hInstance,'SCROLLDRAG',IMAGE_CURSOR,0,0,LR_DEFAULTSIZE or LR_DEFAULTCOLOR);
FDragCursorLoaded:=FDragCursor<>0;
if not FDragCursorLoaded then
  FDragCursor:=LoadCursor(0,IDC_ARROW);

inherited Create(AOwner);

TabStop:=true;
end;

destructor TprPreviewBox.Destroy;
begin
if FNormalCursorLoaded then
  DestroyCursor(FNormalCursor);
if FDragCursorLoaded then
  DestroyCursor(FDragCursor);
ResetRPDI;
ClearFindList;
rpdi.FindList.Free;
inherited;
end;

procedure TprPreviewBox.CreateParams;
begin
inherited;
Params.WindowClass.hCursor:=FNormalCursor;
end;

procedure TprPreviewBox.SetRPDI;
var
  i : integer;
begin
for i:=0 to Report.EndPagesCount-1 do
  TprEndPage(Report.EndPages[i]).rdi.ppdi:=@rpdi;
end;

procedure TprPreviewBox.ResetRPDI;
var
  i : integer;
begin
for i:=0 to Report.EndPagesCount-1 do
  TprEndPage(Report.EndPages[i]).rdi.ppdi:=nil;
end;

procedure TprPreviewBox.mGetDlgCode;
begin
inherited;
Message.Result:=Message.Result or DLGC_WANTARROWS;
end;

procedure TprPreviewBox.MouseDown;
begin
inherited;
if [ssLeft] = Shift then
  begin
    SetCursor(FDragCursor);
    FLastX:=X;
    FLastY:=Y;
  end;
end;

procedure TprPreviewBox.MouseUp;
begin
inherited;
SetCursor(FNormalCursor);
end;

procedure TprPreviewBox.MouseMove;
begin
inherited;
if ([ssLeft] = Shift) and ((FLastX<>X) or (FLastY<>Y)) then
  begin
    HorzScrollBar.Position:=HorzScrollBar.Position+FLastX-X;
    VertScrollBar.Position:=VertScrollBar.Position+FLastY-Y;
    FLastX:=X;
    FLastY:=Y;
  end;
end;

procedure TprPreviewBox.mKeyDown;
var
  si : tagSCROLLINFO;
  ss : TShiftState;

begin
//inherited;
ss:=KeyDataToShiftState(Message.KeyData);

if ss=[] then
  case Message.CharCode of
    VK_DOWN:
      VertScrollBar.Position:=VertScrollBar.Position+VertScrollBar.Increment;
    VK_UP:
      VertScrollBar.Position:=VertScrollBar.Position-VertScrollBar.Increment;
    VK_RIGHT:
      HorzScrollBar.Position:=HorzScrollBar.Position+HorzScrollBar.Increment;
    VK_LEFT:
      HorzScrollBar.Position:=HorzScrollBar.Position-HorzScrollBar.Increment;
    VK_NEXT,VK_PRIOR:
      begin
        ZeroMemory(@si,sizeof(si));
        si.cbSize:=sizeof(si);
        si.fMask :=SIF_PAGE;
        if GetScrollInfo(Handle,SB_VERT,si) then
          begin
            if Message.CharCode=VK_NEXT then
              VertScrollBar.Position:=VertScrollBar.Position+integer(si.nPage)
            else
              VertScrollBar.Position:=VertScrollBar.Position-integer(si.nPage);
          end;
      end;
    else
      inherited;
  end
else
  if ss=[ssCtrl] then
    case Message.CharCode of
      VK_NEXT:
        VertScrollBar.Position:=VertScrollBar.Range;
      VK_PRIOR:
        VertScrollBar.Position:=0;
    end;
end;

procedure TprPreviewBox.CalcMaxSizes;
var
  i : integer;
begin
FMaxWidth :=TprEndPage(Report.EndPages[0]).pInfo.sWidth;
FMaxHeight:=TprEndPage(Report.EndPages[0]).pInfo.sHeight;
for i:=1 to Report.EndPagesCount-1 do
  begin
    if FMaxWidth<TprEndPage(Report.EndPages[i]).pInfo.sWidth then
      FMaxWidth:=TprEndPage(Report.EndPages[i]).pInfo.sWidth;
    if FMaxHeight<TprEndPage(Report.EndPages[i]).pInfo.sHeight then
      FMaxHeight:=TprEndPage(Report.EndPages[i]).pInfo.sHeight;
  end;
end;

function TprPreviewBox.GetScaleKoef;
begin
Result:=Round(FScaleKoef*100);
end;

procedure TprPreviewBox.SetPagesPerHeight;
begin
if (FPagesPerHeight=Value) or (Value<1) then exit;

FPagesPerHeight:=Value;
FScaleMode     :=smPages;
UpdateSettings;
end;

procedure TprPreviewBox.SetPagesPerWidth;
begin
if (FPagesPerWidth=Value) or (Value<1) then exit;

FPagesPerWidth:=Value;
FScaleMode    :=smPages;
UpdateSettings;
end;

procedure TprPreviewBox.SetScaleMode;
begin
if FScaleMode=Value then exit;

FScaleMode:=Value;
UpdateSettings;
end;

procedure TprPreviewBox.SetScalePercent;
begin
if (FScalePercent=Value) or (Value<1) then exit;

FScalePercent:=Value;
FScaleMode   :=smOnePagePercent;
UpdateSettings;
end;

procedure TprPreviewBox.UpdateSettings;
begin
Resize;
end;

procedure TprPreviewBox.InternalCalcPagesSizes;
var
  ep : TprEndPage;
  kx,ky : double;
  i,j,sbw,sbh,mmw,mmh,n,l,t,w,h : integer;

  procedure CachePage;
  var
    r : TRect;
  begin
  r.Left  :=l;
  r.Top   :=t;
  r.Right :=l+w;
  r.Bottom:=t+h;

  ep.Cache(r,false);
  end;

begin
//   ScaleKoef    
OldScaleKoef:=FScaleKoef;
sbw         :=ClientWidth;
sbh         :=ClientHeight;
case ScaleMode of
  smPages:
    begin
      mmw:=FMaxWidth*PagesPerWidth+(PagesPerWidth-1)*PagesStepX;
      mmh:=FMaxHeight*PagesPerHeight+(PagesPerHeight-1)*PagesStepY;
      kx :=(sbw-LeftOffs-RightOffs) / mmw;
      ky :=(sbh-TopOffs-BottomOffs) / mmh;
      if kx<ky then FScaleKoef:=kx
               else FScaleKoef:=kY;

      n:=Report.EndPagesCount div PagesPerWidth;
      n:=n+(Report.EndPagesCount mod PagesPerWidth);

      DrawWidth :=LeftOffs+RightOffs+Round(FMaxWidth*FScaleKoef)*PagesPerWidth+(PagesPerWidth-1)*PagesStepX;
      DrawHeight:=TopOffs+BottomOffs+Round(FMaxHeight*FScaleKoef)*n+(n-1)*PagesStepY;
    end;

  smPageWidth:
    begin
      FScaleKoef :=(sbw-LeftOffs-RightOffs) / FMaxWidth;
      DrawWidth  :=sbw;
      DrawHeight :=TopOffs+BottomOffs+Round(FMaxHeight*FScaleKoef)*(Report.EndPagesCount)+PagesStepY*(Report.EndPagesCount-1);
    end;

  smOnePagePercent:
    begin
      FScaleKoef :=ScalePercent / 100;
      DrawWidth  :=LeftOffs+RightOffs+Round(FMaxWidth*FScaleKoef);
      DrawHeight :=TopOffs+BottomOffs+Round(FMaxHeight*FScaleKoef)*(Report.EndPagesCount)+PagesStepY*(Report.EndPagesCount-1);
    end;
end;

//   ScrollBars
HorzScrollBar.Range:=DrawWidth;
VertScrollBar.Range:=DrawHeight;

//  
case ScaleMode of
  smPages:
    begin
      i:=0;
      while i*PagesPerWidth<Report.EndPagesCount do
        begin
          j:=0;
          while (j<PagesPerWidth) and (i*PagesPerWidth+j<Report.EndPagesCount) do
            begin
              ep:=TprEndPage(Report.EndPages[i*PagesPerWidth+j]);
              w :=Round(ep.pInfo.sWidth*FScaleKoef);
              h :=Round(ep.pInfo.sHeight*FScaleKoef);
              l :=LeftOffs+(Round((FMaxWidth*FScaleKoef))+PagesStepX)*j;
              t :=TopOffs+(Round((FMaxHeight*FScaleKoef))+PagesStepY)*i;

              CachePage;
              Inc(j);
            end;
          Inc(i);
        end;
    end;
  smPageWidth,smOnePagePercent:
    begin
      for i:=0 to Report.EndPagesCount-1 do
        begin
          ep:=TprEndPage(Report.EndPages[i]);
          w :=Round(ep.pInfo.sWidth*FScaleKoef);
          h :=Round(ep.pInfo.sHeight*FScaleKoef);
          l :=LeftOffs;
          t :=TopOffs+(Round((FMaxHeight*FScaleKoef))+PagesStepY)*i;

          CachePage;
        end;
    end;
end;
end;

procedure TprPreviewBox.ClearFindList;
var
  i : integer;
begin
for i:=0 to rpdi.FindList.Count-1 do
  TprFindText(rpdi.FindList[i]).Free;
rpdi.FindList.Clear;
end;

function TprPreviewBox.InternalFind;
var
  ep : TprEndPage;
  DC,TempDC : HDC;
  i,j,k,n,Top,TopIndex : integer;
begin
Result:=false;
DC    :=0;

ClearFindList;
rpdi.DrawMode:=dmFindFirst;

if rpdi.ProgressBar<>nil then
  begin
    rpdi.ProgressBar.Min     :=0;
    rpdi.ProgressBar.Position:=0;
    rpdi.ProgressBar.Max     :=Report.EndPagesCount;
  end;
if rpdi.StatusBar<>nil then
  begin
    rpdi.StatusBar.Panels[1].Text:=prLoadStr(sFind);
    rpdi.StatusBar.Repaint;
  end;
try
  TempDC       :=GetDC(0);
  DC           :=CreateCompatibleDC(TempDC);
  ReleaseDC(0,TempDC);

  for i:=0 to Report.EndPagesCount-1 do
    begin
      if rpdi.ProgressBar<>nil then
        begin
          rpdi.ProgressBar.Position:=rpdi.ProgressBar.Position+1;
          Application.ProcessMessages;
        end;

      //  
      n :=rpdi.FindList.Count;
      ep:=TprEndPage(Report.EndPages[i]);
      for j:=0 to ep.vl.Count-1 do
        TprExObjRecVersion(ep.vl[j]).Draw(DC,@ep.rdi);

      //     TextRect.Top
      j:=n;
      while j<rpdi.FindList.Count do
        begin
          Top:=TprFindText(rpdi.FindList[j]).TextRect.Top;
          k  :=j+1;
          while k<rpdi.FindList.Count do
            if Top=TprFindText(rpdi.FindList[k]).TextRect.Top then
              begin
                TprFindText(rpdi.FindList[k]).Free;
                rpdi.FindList.Delete(k)
              end
            else
              Inc(k);
          with TprFindText(rpdi.FindList[j]).TextRect do
            begin
              Left  :=Left+ep.rdi.bRect.Left;
              Top   :=Top+ep.rdi.bRect.Top;
              Right :=Right+ep.rdi.bRect.Left;
              Bottom:=Bottom+ep.rdi.bRect.Top;
            end;
          Inc(j);
        end;

      //   TextRect.Top
      for j:=n to rpdi.FindList.Count-1 do
        begin
          Top     :=TprFindText(rpdi.FindList[j]).TextRect.Top;
          TopIndex:=j;
          for k:=j+1 to rpdi.FindList.Count-1 do
            if TprFindText(rpdi.FindList[k]).TextRect.Top<Top then
              begin
                Top     :=TprFindText(rpdi.FindList[k]).TextRect.Top;
                TopIndex:=k;
              end;
          if j<>TopIndex then
            rpdi.FindList.Exchange(j,TopIndex);
        end;
    end;

  Result:=rpdi.FindList.Count>0;
finally
  if rpdi.StatusBar<>nil then
    rpdi.StatusBar.Panels[1].Text:='';
  if rpdi.ProgressBar<>nil then
    rpdi.ProgressBar.Position:=0;
  if DC<>0 then
    DeleteDC(DC);
  if Result then
    rpdi.DrawMode:=dmFind
  else
    rpdi.DrawMode:=dmDraw;
end;
end;

procedure TprPreviewBox.Resize;
begin
inherited;
InternalCalcPagesSizes;

if rpdi.DrawMode=dmFind then
  InternalFind;
Repaint;

if (Abs(OldScaleKoef-FScaleKoef)>0.001) and Assigned(OnScalePercentChanged) then
  OnScalePercentChanged(Self);
end;

procedure TprPreviewBox.mPaint;
var
  v : TprExObjRecVersion;
  hsb,vsb,i,j : integer;
  DC : hDC;
  ep : TprEndPage;
  PS : TPaintStruct;
  r1,r2 : TRect;
  nbr,obr : HBRUSH;
  npn,opn : HPEN;
  rgn : HRGN;
  OldOrgEx : TPoint;

begin
Message.Result:=1;
DC :=BeginPaint(Handle,PS);

hsb:=HorzScrollBar.Position;
vsb:=VertScrollBar.Position;
r1 :=Rect(ps.rcPaint.Left+hsb,
          ps.rcPaint.Top+vsb,
          ps.rcPaint.Right+hsb,
          ps.rcPaint.Bottom+vsb);

rgn:=CreateRectRgnIndirect(ps.rcPaint);
nbr:=CreateSolidBrush(clWhite);
npn:=CreatePen(PS_SOLID,1,clBlack);
obr:=SelectObject(DC,nbr);
opn:=SelectObject(DC,npn);
SelectClipRgn(DC,rgn);
try
  for i:=0 to Report.EndPagesCount-1 do
    begin
      ep:=TprEndPage(Report.EndPages[i]);
      if RectOverRect(ep.rdi.bRect,r1) then
        begin
          //   
          r2 :=Rect(PS.rcPaint.Left-ep.rdi.bRect.Left+hsb,
                    PS.rcPaint.Top-ep.rdi.bRect.Top+vsb,
                    PS.rcPaint.Right-ep.rdi.bRect.Left+hsb,
                    PS.rcPaint.Bottom-ep.rdi.bRect.Top+vsb);
          SetViewportOrgEx(DC,
                           ep.rdi.bRect.Left-hsb,
                           ep.rdi.bRect.Top-vsb,
                           @OldOrgEx);

          //  
          Rectangle(DC,0,0,ep.rdi.bRect.Right-ep.rdi.bRect.Left,ep.rdi.bRect.Bottom-ep.rdi.bRect.Top);
          for j:=0 to ep.vl.Count-1 do
            begin
              v:=TprExObjRecVersion(ep.vl[j]);
              if RectOverRect(r2,v.RealRect) then
                v.Draw(DC,@ep.rdi);
            end;

          SetViewportOrgEx(DC,
                           OldOrgEx.X,
                           OldOrgEx.Y,
                           nil);
        end;
    end;
finally
  SelectObject(DC,obr);
  SelectObject(DC,opn);
  SelectClipRgn(DC,0);
  DeleteObject(nbr);
  DeleteObject(npn);
  DeleteObject(rgn);
  EndPaint(Handle,PS);
end;
end;

procedure TprPreviewBox.mVScroll;
begin
inherited;
end;

procedure TprPreviewBox.mHScroll;
begin
inherited;
end;

function TprPreviewBox.GetFindedCount;
begin
Result:=rpdi.FindList.Count;
end;

function TprPreviewBox.GetIsFindMode;
begin
Result:=rpdi.DrawMode=dmFind;
end;

procedure TprPreviewBox.Find;
begin
if InternalFind then
  begin
    FFindPos:=-1;
    FindNext;
    Repaint;
  end
else
  begin
    MBError(Format(prLoadStr(sTextNotFound),[rpdi.FindText]));
  end;
end;

procedure TprPreviewBox.CancelFind;
begin
rpdi.DrawMode:=dmDraw;
ClearFindList;

Repaint;
end;

procedure TprPreviewBox.FindNext;
begin
if FFindPos<rpdi.FindList.Count-1 then
  begin
    Inc(FFindPos)
  end
else
  begin
    if MBox(prLoadStr(sEndOfFind),prLoadStr(sAttention),MB_YESNO+MB_ICONQUESTION)=IDNO then
      exit;
    FFindPos:=0;
  end;

HorzScrollBar.Position:=TprFindText(rpdi.FindList[FFindPos]).TextRect.Left;
VertScrollBar.Position:=TprFindText(rpdi.FindList[FFindPos]).TextRect.Top;
end;

procedure TprPreviewBox.FindPrior;
begin
if FFindPos>0 then
  begin
    Dec(FFindPos);
  end
else
  begin
    if MBox(prLoadStr(sStartOfFind),prLoadStr(sAttention),MB_YESNO+MB_ICONQUESTION)=IDNO then
      exit;
    FFindPos:=rpdi.FindList.Count-1;
  end;

HorzScrollBar.Position:=TprFindText(rpdi.FindList[FFindPos]).TextRect.Left;
VertScrollBar.Position:=TprFindText(rpdi.FindList[FFindPos]).TextRect.Top;
end;

procedure TprPreviewBox.SetToNewReport;
begin
SetRPDI;
CalcMaxSizes;

VertScrollBar.Position:=0;
HorzScrollBar.Position:=0;

Resize;
end;

///////////////////////////
//
// TprPreviewForm
//
///////////////////////////
procedure TprPreviewForm.FormCreate(Sender: TObject);
begin
prLoadResImages(Self,ImageList);

if Report.Title='' then
  Caption:=Format(prLoadStr(sPreviewCaption),[prLoadStr(sNoReportName)])
else
  Caption:=Format(prLoadStr(sPreviewCaption),[Report.Title]);

Preview                       :=TprPreviewBox.CreatePreview(Self,TprReport(Report));
Preview.Parent                :=Self;
Preview.Align                 :=alClient;
Preview.OnScalePercentChanged :=ScaleChanged;
Preview.HorzScrollBar.Tracking:=true;
Preview.VertScrollBar.Tracking:=true;

Preview.rpdi.StatusBar       :=SB;
Preview.rpdi.ProgressBar     :=TProgressBar.Create(Self);
with Preview.rpdi.ProgressBar do
  begin
    Parent:=SB;
  end;

UpdateStatusBar;

aCustomAction.Hint:=Report.CustomActionInPreviewCaption;
ActiveControl     :=Preview;
end;

procedure TprPreviewForm.ScaleChanged;
begin
EDScale.Text:=Format('%d %%',[Preview.ScaleKoef]);
end;

procedure TprPreviewForm.aWholePageExecute(Sender: TObject);
begin
Preview.FPagesPerHeight:=1;
Preview.FPagesPerWidth :=1;
Preview.FScaleMode     :=smPages;
Preview.UpdateSettings;
end;

procedure TprPreviewForm.aPageWidthExecute(Sender: TObject);
begin
Preview.FScaleMode:=smPageWidth;
Preview.UpdateSettings;
end;

procedure TprPreviewForm.aTwoPagesExecute(Sender: TObject);
begin
Preview.FPagesPerHeight:=1;
Preview.FPagesPerWidth :=2;
Preview.FScaleMode     :=smPages;
Preview.UpdateSettings;
end;

procedure TprPreviewForm.aPagesExecute(Sender: TObject);
begin
Preview.FPagesPerHeight:=2;
Preview.FPagesPerWidth :=2;
Preview.FScaleMode     :=smPages;
Preview.UpdateSettings;
end;

procedure TprPreviewForm.aPrintExecute(Sender: TObject);
begin
if Report.SetupPrintParams then
  begin
    Report.PrintPreparedReport;
    Preview.Resize;
  end;
end;

procedure TprPreviewForm.aCloseExecute(Sender: TObject);
begin
Self.Close;
end;

procedure TprPreviewForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
Preview.rpdi.StatusBar  :=nil;
Preview.rpdi.ProgressBar:=nil;
Action:=caFree;
end;

procedure TprPreviewForm.EDScaleKeyPress(Sender: TObject; var Key: Char);
begin
if (Key=#13) and (Trim(EDScale.Text)<>'') then
  begin
    Preview.ScalePercent:=StrToIntDef(Trim(EDScale.Text),100);
    Preview.ScaleMode   :=smOnePagePercent;
  end;
end;

procedure TprPreviewForm.aFindNextExecute(Sender: TObject);
begin
Preview.FindNext;
end;

procedure TprPreviewForm.aFindPrevExecute(Sender: TObject);
begin
Preview.FindPrior;
end;

procedure TprPreviewForm.aFindNextUpdate(Sender: TObject);
begin
TAction(Sender).Enabled:=Preview.IsFindMode;
end;

procedure TprPreviewForm.aFindPrevUpdate(Sender: TObject);
begin
TAction(Sender).Enabled:=Preview.IsFindMode;
end;

procedure TprPreviewForm.aFindCancelUpdate(Sender: TObject);
begin
TAction(Sender).Enabled:=Preview.IsFindMode;
end;

procedure TprPreviewForm.aFindCancelExecute(Sender: TObject);
begin
Preview.CancelFind;
end;

procedure TprPreviewForm.aSaveExecute(Sender: TObject);
begin
if SaveDialog.Execute then
  Report.SavePreparedReportToFile(SaveDialog.FileName);
end;

procedure TprPreviewForm.aOpenExecute(Sender: TObject);
begin
if OpenDialog.Execute then
  begin
    Report.LoadPreparedReportFromFile(OpenDialog.FileName);

    Preview.SetToNewReport;
    UpdateStatusBar;
  end;
end;

procedure TprPreviewForm.WMGetDlgCode;
begin
inherited;

Message.Result:=Message.Result and not DLGC_WANTARROWS;
end;

procedure TprPreviewForm.UpdateStatusBar;
begin
SB.Panels[0].Text:=Format(prLoadStr(sTotalPages),[Report.EndPagesCount]);
end;

procedure TprPreviewForm.FindDialogFind(Sender: TObject);
begin
if not Preview.IsFindMode then
  begin
    Preview.rpdi.CaseSensitive:=frMatchCase in FindDialog.Options;
    if Preview.rpdi.CaseSensitive then
      Preview.rpdi.FindText:=FindDialog.FindText
    else
      Preview.rpdi.FindText:=AnsiUpperCase(FindDialog.FindText);
    Preview.Find;
  end
else
  begin
    if frDown in FindDialog.Options then
      Preview.FindNext
    else
      Preview.FindPrior;
  end;
end;

procedure TprPreviewForm.aFindExecute(Sender: TObject);
begin
FindDialog.Execute;
end;

procedure TprPreviewForm.SBDrawPanel(StatusBar: TStatusBar;
  Panel: TStatusPanel; const Rect: TRect);
begin
if Panel.Index=2 then
  begin
//    Preview.rpdi.ProgressBar.SetBounds(Rect.Left+2,Rect.Top+1,Rect.Right-Rect.Left-2,Rect.Bottom-Rect.Top-2);
    Preview.rpdi.ProgressBar.SetBounds(Rect.Left-1,Rect.Top-1,Rect.Right-Rect.Left+2,Rect.Bottom-Rect.Top+2);
  end;
end;

procedure TprPreviewForm.aCustomActionUpdate(Sender: TObject);
begin
TACtion(Sender).Enabled:=Assigned(Report.OnCustomActionInPreview);
end;          

procedure TprPreviewForm.aCustomActionExecute(Sender: TObject);
begin
DoOnCustomAction;
end;

initialization

RegisterClass(TprPreviewForm);

end.

    